/* -------------------------------------------------------------------------- 
 * Copyright (c) 2018-2019 Kneron Inc. All rights reserved.
 *
 *      Name:    main.c
 *      Purpose: Kneron NCPU
 *
 *---------------------------------------------------------------------------*/


#include <string.h>
#include "kneron_mozart.h"
#include "kneron_mozart_ext.h"
#include "cmsis_os2.h"
#include "os_tick.h"
#include "DrvUART010.h"
#include "ipc.h"
#include "io.h"
#include "base.h"
#include "kdpio.h"
#include "scu_ipc.h"
#include "version.h"
#include "power_mgr.h"
#include "dbg.h"
#include "cpu_ops_ex.h"
#include "post_processing_ex.h"
#include "pre_processing_ex.h"
#include "kdrv_gdma.h"
#include "version.h"

#define FLAG_SCPU_COMM_ISR				BIT0
#define FLAG_SCPU_RESET     			BIT1
#define FLAG_SCPU_ERROR                 BIT2

#define FLAG_SCPU_ACTIVE_0 				BIT10
#define FLAG_SCPU_ACTIVE_1 				BIT11

#define FLAG_SCPU_ALL                   (FLAG_SCPU_ACTIVE_0 | FLAG_SCPU_ACTIVE_1)

#define FLAG_NPU_POST_PROC_0            BIT0
#define FLAG_NPU_POST_PROC_1            BIT1

#define FLAG_NPU_POST_PROC_ALL          (FLAG_NPU_POST_PROC_0 | FLAG_NPU_POST_PROC_1)

#define FLAG_NPU_CPU_NODE_0             BIT0
#define FLAG_NPU_CPU_NODE_1             BIT1

#define FLAG_NPU_CPU_NODE_ALL          (FLAG_NPU_CPU_NODE_0 | FLAG_NPU_CPU_NODE_1)

#define SCPU_IRQ                SYSC_SGI_S_STATUS_IRQ

osThreadId_t tid_scpu_comm;
osThreadId_t tid_cpu_node;
osThreadId_t tid_post_proc;

struct scpu_to_ncpu_s *in_comm_p;
struct ncpu_to_scpu_s *out_comm_p;

#define CMD_FLAGS(i)                (in_comm_p->cmd_flags[i])
#define CMD_FLAGS_IS_ACTIVE(i)      (in_comm_p->cmd_flags[i] == IMAGE_STATE_ACTIVE)
#define CMD_STATUS(i)               (out_comm_p->cmd_status[i])
#define CMD_STATUS_IS_NPU_BUSY(i)   (out_comm_p->cmd_status[i] == IMAGE_STATE_NPU_BUSY)
#define CMD_STATUS_IS_DONE_IDLE(i)  (out_comm_p->cmd_status[i] == IMAGE_STATE_IDLE || out_comm_p->cmd_status[i] == IMAGE_STATE_DONE)

/* IPC_IMAGE_ACTIVE_MAX=2 for 2 parallel input images:
       one for npu, another for post-process */
struct kdp_image_s image_s[IPC_IMAGE_ACTIVE_MAX];


extern void kdp_ddr_dma_copy(unsigned int *src_ptr, unsigned int *dst_ptr, int size);
extern void kdrv_mpu_config(void);

extern void cpu_processing_add(void);
extern void pre_processing_add(void);
extern void post_processing_add(void);

static void pre_processing_main(uint32_t active_idx)
{
    struct kdp_img_raw_s *raw_img_p;
    struct kdp_model_s *model_p;
    int model_slot_index, image_index;
    struct kdp_image_s *image_p;

    model_slot_index = in_comm_p->model_slot_index[active_idx];
    model_p = &in_comm_p->models[model_slot_index];

    dbg_msg("NPU[%d]: model idx %d, output_mem_addr=0x%x, len=%d\n", active_idx, model_slot_index, (uint32_t)model_p->output_mem_addr, model_p->output_mem_len);
    dbg_msg("      : input_mem_addr = 0x%x, len=0x%x, cmd_mem_addr=0x%x, len=0x%x, wt_addr=0x%x, wt_len=0x%x, setup_addr=0x%x, setup_len=0x%x\n",
               (uint32_t)model_p->input_mem_addr, (uint32_t)model_p->input_mem_len,
               (uint32_t)model_p->cmd_mem_addr, (uint32_t)model_p->cmd_mem_len,
               (uint32_t)model_p->weight_mem_addr, (uint32_t)model_p->weight_mem_len,
               (uint32_t)model_p->setup_mem_addr, (uint32_t)model_p->setup_mem_len);

    image_index = in_comm_p->active_img_index[active_idx];
    raw_img_p = &in_comm_p->raw_images[image_index];
    image_p = &image_s[active_idx];

    image_p->raw_img_p = raw_img_p;
    image_p->model_id = in_comm_p->models_type[model_slot_index];
    PREPROC_INPUT_MEM_ADDR(image_p) = model_p->input_mem_addr;
    PREPROC_INPUT_MEM_LEN(image_p) = model_p->input_mem_len;
    POSTPROC_RESULT_MEM_ADDR(image_p) = raw_img_p->results[model_slot_index].result_mem_addr;
    POSTPROC_RESULT_MEM_LEN(image_p) = raw_img_p->results[model_slot_index].result_mem_len;
    POSTPROC_PARAMS_P(image_p) =  raw_img_p->params_s.params;
    
    dbg_msg("NPU: model_type = %d\n", image_p->model_id);

    out_comm_p->img_results[image_index].seq_num = raw_img_p->seq_num;
    out_comm_p->img_results[image_index].status = IMAGE_STATE_ACTIVE;

    PREPROC_INPROC_MEM_ADDR(image_p) = in_comm_p->inproc_mem_addr;
    PREPROC_INPUT_MEM_ADDR2(image_p) = in_comm_p->input_mem_addr2;
    PREPROC_INPUT_MEM_LEN2(image_p) = in_comm_p->input_mem_len2;
    POSTPROC_OUTPUT_MEM_ADDR2(image_p) = in_comm_p->output_mem_addr2;
    POSTPROC_OUTPUT_MEM_LEN2(image_p) = in_comm_p->output_mem_len2;
    POSTPROC_OUTPUT_MEM_ADDR3(image_p) = in_comm_p->output_mem_addr3;
    POSTPROC_OUTPUT_MEM_ADDR4(image_p) = in_comm_p->output_mem_addr4;

    kdpio_set_model(image_p, model_p);

    kdpio_run_preprocess(image_p);
    kdpio_run_npu_op(image_p);
}

/*----------------------------------------------------------------------------
 *      SCPU Communication and Main Thread
 *---------------------------------------------------------------------------*/
void scpu_comm_thread(void *argument)
{
    int32_t flags;
    uint32_t active_idx;
    struct kdp_image_s *image_p;

    /* Init kdpio lib and interrupt/status */
    kdpio_init();

    /* Register pre_processing function dynamically */
    pre_processing_add();

    /* An example to register preprocess function in pre_processing_ex.c */
    //kdpio_pre_processing_register(TINY_YOLO_V3, preprocess_rgba_right_shift_yolo);
    /* Register cpu_op function dynamically */
    cpu_processing_add();

    /* Register post_processing function dynamically */
    post_processing_add();

    scu_ipc_enable_to_scpu_int();

    NVIC_EnableIRQ((IRQn_Type)SCPU_IRQ);
    NVIC_EnableIRQ((IRQn_Type)NPU_NPU_IRQ);

    out_comm_p->id = NCPU2SCPU_ID;
    out_comm_p->status = STATUS_OK;

    DSG("ncpu: Ready!\n");

    /* Let's run tickless */
    OS_Tick_Disable();

    /* Waiting for all events */
    for (;;) {
        flags = osThreadFlagsWait(FLAG_SCPU_ALL, osFlagsWaitAny, osWaitForever);
        dbg_msg("ncpu: got IPC flags 0x%x\n", flags);

        // Enable os tick for profiling or disable for power saving
        if (log_get_level_ncpu() >= LOG_PROFILE)
            OS_Tick_Enable();
        else
            OS_Tick_Disable();

        if (flags & FLAG_SCPU_ACTIVE_0) {
            flags &= ~FLAG_SCPU_ACTIVE_0;
            active_idx = 0;
            pre_processing_main(active_idx);
        } else if (flags & FLAG_SCPU_ACTIVE_1) {
            flags &= ~FLAG_SCPU_ACTIVE_1;
            active_idx = 1;
            pre_processing_main(active_idx);
        }

        if (flags) {
            err_msg("ncpu: unhandled flags 0x%x (%d)\n", flags, active_idx);
            continue;
        }

        /* Return to SCPU for any model without NPU OP */
        image_p = &image_s[active_idx];

        if (RAW_FORMAT(image_p) & IMAGE_FORMAT_BYPASS_NPU_OP) {
            int i = image_p->raw_img_p->ref_idx;
            out_comm_p->img_results[i].status = IMAGE_STATE_DONE;
            out_comm_p->img_index_done[active_idx] = i;
            CMD_STATUS(active_idx) = IMAGE_STATE_DONE;
            out_comm_p->status = STATUS_OK + active_idx;

            scu_ipc_trigger_to_scpu_int();
        }
    }
}

void scpu_isr(void)
{
    int handled = 0;

    dbg_msg("NCPU: cmd %d for states: [0] %d/%d, [1] %d/%d\n", in_comm_p->cmd,
            CMD_FLAGS(0), CMD_STATUS(0),
            CMD_FLAGS(1), CMD_STATUS(1));

    if (in_comm_p->cmd == CMD_RUN_NPU) {
        if (CMD_FLAGS_IS_ACTIVE(0) && CMD_STATUS_IS_DONE_IDLE(0)) {
            CMD_STATUS(0) = IMAGE_STATE_NPU_BUSY;
            osThreadFlagsSet(tid_scpu_comm, FLAG_SCPU_ACTIVE_0);
            handled = 1;
        }
    }

    if (in_comm_p->cmd == CMD_RUN_NPU_1) {
        if (CMD_FLAGS_IS_ACTIVE(1) && CMD_STATUS_IS_DONE_IDLE(1)) {
            CMD_STATUS(1) = IMAGE_STATE_NPU_BUSY;
            osThreadFlagsSet(tid_scpu_comm, FLAG_SCPU_ACTIVE_1);
            handled = 1;
        }
    }

    if (!handled) {
        err_msg("NCPU: unknown cmd %d, states: [0] %d/%d, [1] %d/%d\n", in_comm_p->cmd,
                CMD_FLAGS(0), CMD_STATUS(0),
                CMD_FLAGS(1), CMD_STATUS(1));
    }

    scu_ipc_clear_from_scpu_int();
    NVIC_ClearPendingIRQ((IRQn_Type)SCPU_IRQ);
}

static void npu_out_data_move(void *dst_p, void *src_p, int size)
{
    /* Move data out from npu output buffer for parallel handling */
    kdrv_gdma_memcpy((uint32_t)dst_p,(uint32_t)src_p, size); 

    /* Notify scpu now */
    scu_ipc_trigger_to_scpu_int();
}

/*----------------------------------------------------------------------------
 *      NPU Post Processing Thread
 *---------------------------------------------------------------------------*/
void post_proc_thread(void *argument)
{
    int32_t flags;
    int rc, i;
    uint32_t    active_idx;
    struct kdp_image_s *image_p;

    for (;;) {
        flags = osThreadFlagsWait(FLAG_NPU_POST_PROC_ALL, osFlagsWaitAny, osWaitForever);

        if (flags & FLAG_NPU_POST_PROC_0) {
            flags &= ~FLAG_NPU_POST_PROC_0;
            active_idx = 0;
        } else if (flags & FLAG_NPU_POST_PROC_1) {
            flags &= ~FLAG_NPU_POST_PROC_1;
            active_idx = 1;
        }

        if (flags) {
            err_msg("post_proc_thread: unhandled flags 0x%x (%d)\n", flags, active_idx);
            continue;
        }

        image_p = &image_s[active_idx];
        i = image_p->raw_img_p->ref_idx;

        dbg_msg("post_proc_thread: [%d] img[%d] state: %d -> %d\n", active_idx, i, out_comm_p->img_results[i].status, IMAGE_STATE_NPU_DONE);

        /* Report status for NPU & CPU processing */
        out_comm_p->img_index_done[active_idx] = i;
        out_comm_p->status = STATUS_OK + active_idx;

        if (RAW_FORMAT(image_p) & IMAGE_FORMAT_PARALLEL_PROC) {
            out_comm_p->irq_queue[out_comm_p->irq_enqueued] = 
                    IRQ_BASE + IRQ_PER_IMG * active_idx + IRQ_NPU_DONE;
            out_comm_p->irq_enqueued++;
            out_comm_p->irq_enqueued %= IRQ_QUEUE_SIZE;
        }

        dbg_msg("  <- IPC status %d: [%d] %d/%d\n", out_comm_p->status, active_idx,
                CMD_FLAGS(active_idx), CMD_STATUS(active_idx));

        rc = kdpio_run_postprocess(image_p, npu_out_data_move);

        if (rc == RET_NO_ERROR) {
            dbg_msg("npu out: [%d] img[%d] state: %d -> %d\n", active_idx, i, out_comm_p->img_results[i].status, IMAGE_STATE_DONE);
            out_comm_p->img_results[i].status = IMAGE_STATE_DONE;
            CMD_STATUS(active_idx) = IMAGE_STATE_DONE;
            out_comm_p->status = STATUS_OK + active_idx;
        }

        if (rc == RET_ERROR) {
            /* report error status */
            out_comm_p->status = STATUS_ERR;
        }

        dbg_msg("  <- IPC status %d: [%d] %d/%d\n", out_comm_p->status, active_idx,
                CMD_FLAGS(active_idx), CMD_STATUS(active_idx));

        out_comm_p->irq_queue[out_comm_p->irq_enqueued] = IRQ_BASE + IRQ_PER_IMG * active_idx + IRQ_POST_PROC_DONE;
        out_comm_p->irq_enqueued++;
        out_comm_p->irq_enqueued %= IRQ_QUEUE_SIZE;

        /* notify scpu */
        scu_ipc_trigger_to_scpu_int();
    }
}

/*----------------------------------------------------------------------------
 *      CPU Node Processing Thread
 *---------------------------------------------------------------------------*/
void cpu_node_thread(void *argument)
{
    int32_t flags;
    int rc;
    uint32_t    active_idx;
    struct kdp_image_s *image_p;

    for (;;) {
        flags = osThreadFlagsWait(FLAG_NPU_CPU_NODE_ALL, osFlagsWaitAny, osWaitForever);

        if (flags & FLAG_NPU_CPU_NODE_0) {
            flags &= ~FLAG_NPU_CPU_NODE_0;
            active_idx = 0;
        } else if (flags & FLAG_NPU_CPU_NODE_1) {
            flags &= ~FLAG_NPU_CPU_NODE_1;
            active_idx = 1;
        }

        if (flags) {
            err_msg("cpu_node_thread: unhandled flags 0x%x (%d)\n", flags, active_idx);
            continue;
        }

        image_p = &image_s[active_idx];

        rc = kdpio_run_cpu_op(image_p);
        if (rc < 0)
            err_msg("kdpio_run_cpu_op: [%d] rc %d @ %p\n", active_idx, rc, image_p);
    }
}

void npu_isr(void)
{
    uint32_t    active_idx;
    struct kdp_image_s *image_p;
    uint32_t    *setup_mem_p;
    uint32_t    node_id;

    if (!kdpio_handle_npu_int()) {

        dbg_msg("npu_isr: states: [0] %d/%d, [1] %d/%d\n",
                CMD_FLAGS(0), CMD_STATUS(0),
                CMD_FLAGS(1), CMD_STATUS(1));

        if (CMD_FLAGS_IS_ACTIVE(0) && CMD_STATUS_IS_NPU_BUSY(0)) {
            active_idx = 0;
        } else if (CMD_FLAGS_IS_ACTIVE(1) && CMD_STATUS_IS_NPU_BUSY(1)) {
            active_idx = 1;
        }

        image_p = &image_s[active_idx];
        setup_mem_p = (uint32_t *)MODEL_SETUP_MEM_P(image_p);
        node_id = *setup_mem_p;

        if (node_id == NODE_TYPE_OUT) {
            CMD_STATUS(active_idx) = IMAGE_STATE_POST_PROCESSING;
            osThreadFlagsSet(tid_post_proc, FLAG_NPU_POST_PROC_0 << active_idx);
        } else if (node_id == NODE_TYPE_CPU) {
            osThreadFlagsSet(tid_cpu_node, FLAG_NPU_CPU_NODE_0 << active_idx);
        } else {
            err_msg("npu_isr: [%d] wrong node_id %d @ %p\n", active_idx, node_id, image_p);
        }
    } else {
        out_comm_p->status = STATUS_ERR;
        scu_ipc_trigger_to_scpu_int();
    }

    NVIC_ClearPendingIRQ((IRQn_Type)NPU_NPU_IRQ);
}

/*----------------------------------------------------------------------------
 *      Init
 *---------------------------------------------------------------------------*/
static void ncpu_init()
{
    osThreadAttr_t attr;

    memset(&attr, 0, sizeof(attr));
    attr.priority = osPriorityNormal;
    tid_scpu_comm = osThreadNew(scpu_comm_thread, NULL, &attr);
    attr.priority = osPriorityHigh;
    tid_cpu_node = osThreadNew(cpu_node_thread, NULL, &attr);
    attr.priority = osPriorityLow;
    tid_post_proc = osThreadNew(post_proc_thread, NULL, &attr);

    NVIC_SetVector((IRQn_Type)SCPU_IRQ, (uint32_t)scpu_isr);
    NVIC_SetVector((IRQn_Type)NPU_NPU_IRQ, (uint32_t)npu_isr);

    in_comm_p = (struct scpu_to_ncpu_s *)IPC_MEM_ADDR;
    out_comm_p = (struct ncpu_to_scpu_s *)IPC_MEM_ADDR2;

    power_mgr_init();
}

#define NCPU_VERSION_LEN    8   // length from end
#define NCPU_VERSION_ADDR   (SdRAM_MEM_BASE + SdRAM_MEM_SIZE - NCPU_VERSION_LEN)

static void ncpu_set_version()
{
    memcpy((void *)NCPU_VERSION_ADDR, (const void *)&g_fw_misc, NCPU_VERSION_LEN);
}

void main_init(void)
{
    ncpu_init();
    kdrv_gdma_initialize();
    kdrv_mpu_config();

    ncpu_set_version();
}

