/********************************************************************
 * Copyright (c) 2020 Kneron, Inc. All Rights Reserved.
 *
 * The information contained herein is property of Kneron, Inc.
 * Terms and conditions of usage are described in detail in Kneron
 * STANDARD SOFTWARE LICENSE AGREEMENT.
 *
 * Licensees are granted free, non-transferable use of the information.
 * NO WARRANTY of ANY KIND is provided. This heading must NOT be removed
 * from the file.
 ********************************************************************/

/**@addtogroup  KDRV_ADC  KDRV_ADC
 * @{
 * @brief       Kneron generic adc driver
 * @copyright   Copyright (C) 2020 Kneron, Inc. All rights reserved.
 */
#ifndef __KDRV_ADC_H
#define __KDRV_ADC_H

#include "cmsis_os2.h"
#include "kdrv_status.h"

/** ADC Resource Configuration */
typedef struct {
  int               io_base;        // I2C register interface
  int               irq;            // I2C Event IRQ Number
} kdrv_adc_resource_t;

/** ADC Register Configuration */
struct kdrv_adc_regs_t {
    /* DATA: offset 0x000~0x01C */
    uint32_t        data[8];
    /* reserve */
    uint32_t        reserve[24];
    /* THRHOLD: offset 0x080~0x09C */
    uint32_t        thrhold[8];
#define HTHR_EN             (1<<31)
#define HTHR(x)             (((x)&0xFFF)<<16)
#define LTHR_EN             (1<<15)
#define LTHR(x)             (((x)&0xFFF)<<0)
    /* reserve */
    uint32_t        reserve1[24];
    /* CTRL: offset 0x100 */
    uint32_t        ctrl;
#define SCAN_NUM(x)         (x<<16)
#define SCANMODE_CONT       (1<<9)
#define SCANMODE_SGL        (1<<8)
#define SWSTART             (1<<4)
#define ADC_EN              (1<<0)
    /* TRIM: offset 0x104 */
    uint32_t        trim;
    /* INTEN: offset 0x108 */
    uint32_t        inten;
#define CHDONE_INTEN(x)     (1<<((x)+8))
#define TS_OVR_INTREN       (1<<3)
#define TS_UDR_INTREN       (1<<2)
#define STOP_INTEN          (1<<1)
#define DONE_INTEN          (1<<0)
    /* INTST: offset 0x10C */
    uint32_t        intst;
#define CH_INTRSTS(x)       (1<<((x)+8))
#define TS_THDOD_INTRSTS    (1<<3)
#define TS_THDUD_INTRSTS    (1<<2)
#define ADC_STOP_INTRSTS    (1<<1)
#define ADC_DONE_INTSTS     (1<<0)
    /* TPARAM: offset 0x110 */
    uint32_t        tparam;
    /* TPARAM1: offset 0x114 */
    uint32_t        smpr;
    /* reserve */
    uint32_t        reserve2;
    /* PRESCAL: offset 0x11C */
    uint32_t        prescal;
    /* SQR:         offset 0x120 */
    uint32_t        sqr;

};

/**
 * @brief           ADC driver initialization
 *
 * @return          kdrv_status_t         see @ref kdrv_status_t
 */
kdrv_status_t kdrv_adc_initialize(void);
/**
 * @brief           ADC driver uninitialization
 *
 * @param[in]       *res                  a handle of a ADC resource
 * @return          kdrv_status_t         see @ref kdrv_status_t
 */
kdrv_status_t kdrv_adc_uninitialize(kdrv_adc_resource_t *res);
/**
 * @brief           ADC reset control
 *
 * @param[in]       *res                  a handle of a ADC resource
 * @return          kdrv_status_t         see @ref kdrv_status_t
 */
kdrv_status_t kdrv_adc_rest(kdrv_adc_resource_t *res);
/**
 * @brief           ADC enable control
 *
 * @param[in]       *res                  a handle of a ADC resource
 *                  mode                  which mode to be enabled
 * @return          kdrv_status_t         see @ref kdrv_status_t
 */
kdrv_status_t kdrv_adc_enable(kdrv_adc_resource_t *res, int mode);
/**
 * @brief           ADC data read control
 *
 * @param[in]       id                    which adc channel to be enabled
 * @return          int                   return ADC data
 */
int kdrv_adc_read(int id);
#endif //__KDRV_ADC_H
