/******************************************************************************
*  Copyright 2020, Kneron, Inc.
*  ---------------------------------------------------------------------------
*  Statement:
*  ----------
*  This software is protected by Copyright and the information contained
*  herein is confidential. The software may not be copied and the information
*  contained herein may not be used or disclosed except with the written
*  permission of Kneron, Inc. (C) 2020
******************************************************************************/

/******************************************************************************
*  Filename:
*  ---------
*  kdrv_fb_mgr.h
*
*  Project:
*  --------
*  KL520
*
*  Description:
*  ------------
*  Configuration for KL520
*
*  Author:
*  -------
*  XXX
*
*===========================================================================
*
******************************************************************************/
/**@addtogroup  KDRV_FB_MGR  KDRV_FB_MGR
 * @{
 * @brief       Kneron mipicsirx driver
 * @copyright   Copyright (C) 2020 Kneron, Inc. All rights reserved.
 */

#ifndef _KDRV_FB_MGR_H_
#define _KDRV_FB_MGR_H_
/******************************************************************************
Head Block of The File
******************************************************************************/
#include <stdint.h>

#define MAX_FRAME_BUFFER        7

/* Buffer sources for display */
#define DISPLAY_SRC_READ     0   // use read-done buffer : default
#define DISPLAY_SRC_INF      1   // use inf-done buffer
#define DISPLAY_SRC_WRITE    2   // use write-done buffer

typedef int (*fb_write_done_notify)(int cam_idx, int write_idx);

/**
 * @brief To initialize frame buffer pool for camera
 *
 * @param [in] cam_idx camera index
 * @param [in] buf_size buffer size
 * @param [in] buf_num buffer number
 * @param [in] frame_info_size frame info size (e.g. tile average)
 * @return 0 for success, or -1 for error
 */
int kdrv_fb_mgr_init(int cam_idx, uint32_t buf_size, int buf_num, int frame_info_size);

/**
 * @brief To get next available buffer for camera to write
 *
 * @param [in] cam_idx camera index
 * @param [in] write_idx an address/pointer to get the buffer index
 * @return the address of available buffer
 */
uint32_t kdrv_fb_mgr_next_write(int cam_idx, int *write_idx);

/**
 * @brief To notify the finish of a buffer written by camera
 *
 * @param [in] cam_idx camera index
 * @param [in] write_idx the buffer index of write-done
 * @return 0 for success, or -1 for error
 */
int kdrv_fb_mgr_write_done(int cam_idx, int write_idx);

/**
 * @brief To free the use of frame buffers for camera
 *
 * @param [in] cam_idx camera index
 * @return 0 for success, or -1 for error
 */
void kdrv_fb_mgr_free_write_buf(int cam_idx);

/**
 * @brief To get a buffer of an index
 *
 * @param [in] cam_idx camera index
 * @param [in] idx a buffer index
 * @return the buffer address of the idx
 */
uint32_t kdrv_fb_mgr_get_buf(int cam_idx, int idx);

/**
 * @brief To get a buffer of an index
 *
 * @param [in] cam_idx camera index
 * @param [in] idx a buffer index
 * @return the frame info buffer address of the idx
 */
uint32_t kdrv_fb_mgr_get_frame_info_buf(int cam_idx, int idx);

/**
 * @brief To get the (camera write) sequence number of a buffer
 *
 * @param [in] cam_idx camera index
 * @param [in] idx a buffer index
 * @return the sequence number starting from 0
 */
uint32_t kdrv_fb_mgr_get_buf_seq_num(int cam_idx, int idx);

/**
 * @brief To get current buffer sequence number
 *
 * @param [in] cam_idx camera index
 * @return the sequence number starting from 0
 */
uint32_t kdrv_fb_mgr_get_current_buf_seq_num(int cam_idx);

/**
 * @brief To get the next buffer for drawing
 *
 * @param [in] cam_idx camera index
 * @param [in] read_idx an address/pointer to get the buffer index
 * @return the buffer address
 */
uint32_t kdrv_fb_mgr_next_read(int cam_idx, int *read_idx);

/**
 * @brief To notify the finish of a buffer done for drawing
 *
 * @param [in] cam_idx camera index
 * @param [in] read_idx the buffer index done for drawing
 * @return 0 for success, or -1 for error
 */
int kdrv_fb_mgr_read_done(int cam_idx, int read_idx);

/**
 * @brief To free the use of frame buffers for drawing
 *
 * @param [in] cam_idx camera index
 * @return 0 for success, or -1 for error
 */
void kdrv_fb_mgr_free_read_buf(int cam_idx);

/**
 * @brief To get the next buffer for inference
 *
 * @param [in] cam_idx camera index
 * @param [in] inf_idx an address/pointer to get the buffer index
 * @return the buffer address
 */
uint32_t kdrv_fb_mgr_next_inf(int cam_idx, int *inf_idx);

/**
 * @brief To notify the finish of a buffer done for inference
 *
 * @param [in] cam_idx camera index
 * @param [in] inf_idx the buffer index done for inference
 * @return 0 for success, or -1 for error
 */
int kdrv_fb_mgr_inf_done(int cam_idx, int inf_idx);

/**
 * @brief To free the use of frame buffers for inference
 *
 * @param [in] cam_idx camera index
 * @return 0 for success, or -1 for error
 */
void kdrv_fb_mgr_free_inf_buf(int cam_idx);

/**
 * @brief To get the buffer for display
 *
 * @param [in] cam_idx camera index
 * @param [in] disp_idx an address/pointer to get the buffer index
 * @return the buffer address
 */
uint32_t kdrv_fb_mgr_next_display(int cam_idx, int *disp_idx);

/**
 * @brief To set display buffer source
 *
 * @param [in] cam_idx camera index
 * @param [in] disp_src one of display sources from read(drawing)/inference/write
 * @return 0 for success, or -1 for error
 */
int kdrv_fb_mgr_display_set_src(int cam_idx, int disp_src);

/**
 * @brief To register a callback for a finished frame by camera
 *
 * @param [in] cam_idx camera index
 * @param [in] callback the callback function when a frame buffer is done written
 * @return 0 for success, or -1 for error
 */
int kdrv_fb_mgr_notifier_register(int cam_idx, fb_write_done_notify callback);

#endif // _KDRV_FB_MGR_H_
