#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include "kmdw_console.h"

#define BACKSP_KEY 0x08
#define RETURN_KEY 0x0D
#define DELETE_KEY 0x7F
#define BELL 0x07

#define MAX_LOG_LENGTH 256
#define DDR_MAX_LOG_COUNT 1000 // if using DDR
#define DDR_LOG_BUFFER_SIZE (1 * 1024 * 1024)

#define HEAP_MAX_LOG_COUNT 20 // if using SRAM Heap

static uint32_t scpu_debug_flags = 0;

static kdrv_uart_handle_t handle0;
static osThreadId_t logger_tid = NULL;
static osMessageQueueId_t log_msgq = NULL;

static void _print_string(const char *str)
{
    kdrv_uart_write(handle0, (uint8_t *)str, strlen(str));
}

kmdw_status_t kmdw_printf(const char *fmt, ...)
{
    va_list arg_ptr;
    static char *buffer_s = NULL;

    if (NULL == buffer_s)
        buffer_s = (char*)malloc(sizeof(char)*MAX_LOG_LENGTH);

    va_start(arg_ptr, fmt);
    //vsprintf(buffer_s, fmt, arg_ptr);
    vsnprintf(buffer_s, MAX_LOG_LENGTH-1, fmt, arg_ptr);
    va_end(arg_ptr);

    buffer_s[MAX_LOG_LENGTH - 1] = 0; // just in case

    if (osThreadGetId() == NULL)
    {
        _print_string(buffer_s);
    }
    else
    {
        osStatus_t oss = osMessageQueuePut(log_msgq, buffer_s, NULL, 0);
        if (oss != osOK)
        {
            //_print_string("[logger] enqueue log1 failed\n");
            return KMDW_STATUS_ERROR;
        }
    }

    return KMDW_STATUS_OK;
}

kmdw_status_t kmdw_level_printf(int level, const char *fmt, ...)
{
    static char *buffer_s = NULL;
    uint32_t lvl = kmdw_console_get_log_level_scpu();
    lvl >>= 16;

    if ((level == LOG_PROFILE && level == lvl) || (level > 0 && level <= lvl))
    {
        va_list arg_ptr;

        if (NULL == buffer_s)
            buffer_s = (char*)malloc(sizeof(char)*MAX_LOG_LENGTH);

        va_start(arg_ptr, fmt);
        //vsprintf(buffer_s, fmt, arg_ptr);
        vsnprintf(buffer_s, MAX_LOG_LENGTH-1, fmt, arg_ptr);
        va_end(arg_ptr);

        buffer_s[MAX_LOG_LENGTH - 1] = 0; // just in case

        if (osThreadGetId() == NULL)
        {
            _print_string(buffer_s);
        }
        else
        {
            osStatus_t oss = osMessageQueuePut(log_msgq, buffer_s, NULL, 0);
            if (oss != osOK)
            {
                //_print_string("[logger] enqueue log failed\n");
                return KMDW_STATUS_ERROR;
            }
        }
    }

    return KMDW_STATUS_OK;
}

static void logger_thread(void *arg)
{
    uint8_t log[MAX_LOG_LENGTH];
    osThreadSetPriority(osThreadGetId(), osPriorityBelowNormal);

    while (1)
    {
        osStatus_t oss = osMessageQueueGet(log_msgq, &log[0], NULL, osWaitForever);
        if (oss != osOK)
        {
            _print_string("[logger] dequeue log failed\n");
            continue;
        }

        _print_string((const char *)log);
    }
}

__weak uint32_t kmdw_ddr_reserve(uint32_t numbyte)
{
    return 0;
}

kmdw_status_t kmdw_console_init(kdrv_uart_dev_id_t uart_dev)
{
    kdrv_status_t sts = kdrv_uart_open(&handle0, uart_dev,
                                       UART_MODE_SYNC_RX | UART_MODE_SYNC_TX, NULL);
    if (sts != KDRV_STATUS_OK)
        return KMDW_STATUS_ERROR;

    kdrv_uart_config_t cfg;
    cfg.baudrate = BAUD_115200;
    cfg.data_bits = 8;
    cfg.frame_length = 0;
    cfg.stop_bits = 1;
    cfg.parity_mode = PARITY_NONE;
    cfg.fifo_en = false;
#ifdef UART_RX_IRQ
    cfg.irq_en = false;
#endif

    sts = kdrv_uart_configure(handle0, UART_CTRL_CONFIG, (void *)&cfg);
    if (sts != KDRV_STATUS_OK)
        return KMDW_STATUS_ERROR;

    osThreadAttr_t thread_attr;
    memset(&thread_attr, 0, sizeof(thread_attr));
    thread_attr.stack_size = 512;
    thread_attr.priority = osPriorityBelowNormal;

    logger_tid = osThreadNew(logger_thread, NULL, &thread_attr);
    if (logger_tid == NULL)
    {
        //_print_string("[logger] osThreadNew failed\n");
        return KMDW_STATUS_ERROR;
    }

    uint32_t log_count = 0;
    osMessageQueueAttr_t msgq_attr;

    memset(&msgq_attr, 0, sizeof(msgq_attr));

    void *log_pool = (void *)kmdw_ddr_reserve(DDR_LOG_BUFFER_SIZE);

    if (log_pool)
    {
        log_count = DDR_MAX_LOG_COUNT;
        msgq_attr.mq_mem = log_pool;
        msgq_attr.mq_size = DDR_LOG_BUFFER_SIZE;
        _print_string("[logger] Using DDR\n");
    }
    else
    {
        _print_string("[logger] Using HEAP\n");
        log_count = HEAP_MAX_LOG_COUNT;
    }

    log_msgq = osMessageQueueNew(log_count, MAX_LOG_LENGTH, &msgq_attr);
    if (log_msgq == NULL)
    {
        _print_string("[logger] osMessageQueueNew failed\n");
        return KMDW_STATUS_ERROR;
    }

    return KMDW_STATUS_OK;
}

char kmdw_console_getc(void)
{
    char c;
    kdrv_uart_read(handle0, (uint8_t *)&c, 1);
    return c;
}

void kmdw_console_putc(char Ch)
{
    char cc;

    if (Ch != '\0')
    {
        cc = Ch;
        kdrv_uart_write(handle0, (uint8_t *)&cc, 1);
    }

    if (Ch == '\n')
    {
        cc = '\r';
        kdrv_uart_write(handle0, (uint8_t *)&cc, 1);
    }
}

void kmdw_console_puts(char *str)
{
    char *cp;
    for (cp = str; *cp != 0; cp++)
        kmdw_console_putc(*cp);
}

int kmdw_console_echo_gets(char *buf, int len)
{
    char *cp;
    char data;
    uint32_t count;
    count = 0;
    cp = buf;

    do
    {
        kdrv_uart_get_char(handle0, &data);
        switch (data)
        {
        case RETURN_KEY:
            if (count < len)
            {
                *cp = '\0';
                kmdw_console_putc('\n');
            }
            break;
        case BACKSP_KEY:
        case DELETE_KEY:
            if ((count > 0) && (count < len))
            {
                count--;
                *(--cp) = '\0';
                kmdw_console_puts("\b \b");
            }
            break;
        default:
            if (data > 0x1F && data < 0x7F && count < len)
            {
                *cp = (char)data;
                cp++;
                count++;
                kmdw_console_putc(data);
            }
            break;
        }
    } while (data != RETURN_KEY);

    return (count);
}

__weak void kdrv_ncpu_set_scpu_debug_lvl(uint32_t lvl)
{
}

__weak void kdrv_ncpu_set_ncpu_debug_lvl(uint32_t lvl)
{
}

void kmdw_console_set_log_level_scpu(uint32_t level)
{
    scpu_debug_flags = (scpu_debug_flags & ~0x000F0000) | (((level) << 16) & 0x000F0000);
    kdrv_ncpu_set_scpu_debug_lvl(level);
}

uint32_t kmdw_console_get_log_level_scpu(void)
{
    return scpu_debug_flags;
}

void kmdw_console_set_log_level_ncpu(uint32_t level)
{
    kdrv_ncpu_set_ncpu_debug_lvl(level);
}
