//#define UFIFO_DBG

#include "cmsis_os2.h"

#include "fifoq.h"
#include "kdrv_usbd2.h"
#include "dual_fifo.h"
#include "kmdw_console.h"
#include "kdrv_power.h"

#ifdef UFIFO_DBG
#define ufifo_dbg(__format__, ...) kmdw_printf("[fifoQ USB]"__format__, ##__VA_ARGS__)
#else
#define ufifo_dbg(__format__, ...)
#endif

#define FLAG_USB_IS_CONNECTED 0x400

static dual_fifo_t gImg_dfifo = 0;
static dual_fifo_t gResult_dfifo = 0;

static uint8_t g_enp_in = 0x0;
static uint8_t g_enp_out = 0x0;

static osThreadId_t usbd_queue_rx_tid = 0;
static osThreadId_t usbd_queue_tx_tid = 0;

static uint8_t result_send_state = 0;
static uint8_t data_recv_state = 0;

extern osThreadId_t tid_usb_comm;         // from usb_com.c
#define FLAG_USB_COMM_WAIT_CONNECTED 0x20 // BIT(5) from usb_com.

// to override the usbd2 driver callback
void usbd2_event_isr_callback(kdrv_usbd2_link_status_t link_status)
{
    switch (link_status)
    {
    case USBD2_STATUS_DISCONNECTED:
        kmdw_printf("USB is disconnected\n");
        break;

    case USBD2_STATUS_CONFIGURED:

        kmdw_printf("USB is connected\n");

        osThreadFlagsSet(tid_usb_comm, FLAG_USB_COMM_WAIT_CONNECTED);
        osThreadFlagsSet(usbd_queue_tx_tid, FLAG_USB_IS_CONNECTED);
        osThreadFlagsSet(usbd_queue_rx_tid, FLAG_USB_IS_CONNECTED);

        break;
    }
}

// to override the usbd2 driver callback
bool usbd2_usr_cx_isr_callback(kdrv_usbd2_setup_packet_t *setup)
{
#define CONTROL_RESET_HARDWARE 0x1
#define CONTROL_RESET_FIFOQ 0x10
#define CONTROL_FIFOQ_INFO 0x11

    bool ret = false;

    switch (setup->bRequest)
    {
    case CONTROL_RESET_HARDWARE:
    {
        static uint8_t reset_stage = 0;

        ret = true;
        if (setup->wValue == 0)
            reset_stage = 1;
        else if (reset_stage == 1 && setup->wValue == 1)
            kdrv_power_sw_reset();
        else
            ret = false;

        break;
    }
    case CONTROL_RESET_FIFOQ:
    {
        while (1)
        {
            uint32_t buf_addr;
            osStatus_t fifo_sts = dual_fifo__dequeue_data(gImg_dfifo, &buf_addr, 0);
            if (fifo_sts == osOK)
            {
                ufifo_dbg("recycle one image data\n");
                fifo_sts = dual_fifo__put_free_buffer(gImg_dfifo, &buf_addr, 0);
                if (fifo_sts != osOK)
                {
                    ufifo_dbg("!!!!!!!! free image buffer failed\n");
                }
            }
            else
                break;
        }

        while (1)
        {
            uint32_t buf_addr;
            osStatus_t fifo_sts = dual_fifo__dequeue_data(gResult_dfifo, &buf_addr, 0);
            if (fifo_sts == osOK)
            {
                ufifo_dbg("recycle one result data\n");
                fifo_sts = dual_fifo__put_free_buffer(gResult_dfifo, &buf_addr, 0);
                if (fifo_sts != osOK)
                {
                    ufifo_dbg("!!!!!!!! free result buffer failed\n");
                }
            }
            else
                break;
        }

        if (data_recv_state == 2)
            kdrv_usbd2_terminate_endpoint(g_enp_out);

        if (result_send_state == 2)
            kdrv_usbd2_terminate_endpoint(g_enp_in);

        ret = true;
        break;
    }
    case CONTROL_FIFOQ_INFO:
    {
        kmdw_printf("\n==========================================\n");
        kmdw_printf("num of un-consumed image: %d\n", dual_fifo__num_unconsumed_data(gImg_dfifo));
        kmdw_printf("num of free-buffer image: %d\n", dual_fifo__num_free_buffer(gImg_dfifo));
        kmdw_printf("num of un-consumed result: %d\n", dual_fifo__num_unconsumed_data(gResult_dfifo));
        kmdw_printf("num of free-buffer result: %d\n", dual_fifo__num_free_buffer(gResult_dfifo));

        kmdw_printf("rx thread state : %s\n", data_recv_state == 2 ? "waiting/receiving data from host" : "dequing free buffer");
        kmdw_printf("tx thread state : %s\n", result_send_state == 2 ? "waiting/sending data to host" : "dequing result data");

        ret = true;
        break;
    }
    default:
        ret = false;
        break;
    }

    return ret;
}

static int _receive_data_via_usb(uint32_t buf_addr, uint32_t buf_size)
{
    uint32_t buf_cur_addr, wanted_size, recved_size, loop;

receive_start_over:

    wanted_size = recved_size = loop = 0;
    buf_cur_addr = buf_addr;

    // multiple reception into one contiguous buffer
    do
    {
        uint32_t len = buf_size - recved_size;

        kdrv_status_t usb_sts = kdrv_usbd2_bulk_receive(g_enp_out, (void *)buf_cur_addr, &len, 0);
        if (usb_sts != KDRV_STATUS_OK)
        {
            if (usb_sts == KDRV_STATUS_USBD_TRANSFER_TERMINATED)
            {
                if (kdrv_usbd2_get_link_status() == USBD2_STATUS_DISCONNECTED)
                {
                    ufifo_dbg("[%s] bulk receive error due to disconnection\n", __FUNCTION__);
                    return -1;
                }
                else
                {
                    ufifo_dbg("[%s] bulk receive is terminated, start-over\n", __FUNCTION__);
                    goto receive_start_over; // assume this is a sw endpoint reset
                }
            }
            else
            {
                ufifo_dbg("[%s] bulk receieve error, status %d\n", __FUNCTION__, usb_sts);
                return -2;
            }
        }

        if (loop == 0 && len < 4)
        {
            ufifo_dbg("[%s] bulk receieve error only %d bytes\n", __FUNCTION__, len);
            return -2;
        }

        if (loop++ == 0)
            wanted_size = *(uint32_t *)buf_cur_addr;

        buf_cur_addr += len;
        recved_size += len;

        //ufifo_dbg("[%s] loop %d received %d bytes\n", __FUNCTION__, loop, len);

    } while (wanted_size > recved_size);

    ufifo_dbg("[%s] total received %d bytes\n", __FUNCTION__, recved_size);

    return 0;
}

static void _rx_image_usb_thread(void *arg)
{
rx_wait_usb_connection:

    data_recv_state = 0;

    osThreadFlagsWait(FLAG_USB_IS_CONNECTED, osFlagsWaitAny, osWaitForever);

    ufifo_dbg("[recv] ready !\n");

    osStatus_t fifo_sts;

    uint32_t buf_addr;
    uint32_t buf_size = dual_fifo__buffer_size(gImg_dfifo);

    while (1)
    {
        data_recv_state = 1;

        fifo_sts = dual_fifo__get_free_buffer(gImg_dfifo, &buf_addr, osWaitForever, false);
        if (fifo_sts != osOK)
            break;

        data_recv_state = 2;

        int result = _receive_data_via_usb(buf_addr, buf_size);
        if (result == -1)
        {
            dual_fifo__put_free_buffer(gImg_dfifo, &buf_addr, osWaitForever);
            goto rx_wait_usb_connection;
        }
        else if (result == -2)
            break;

        data_recv_state = 3;

        fifo_sts = dual_fifo__enqueue_data(gImg_dfifo, &buf_addr, osWaitForever);
        if (fifo_sts != osOK)
            break;
    }

    ufifo_dbg("[recv] stopped !\n");
}

static int _send_data_via_usb(uint32_t buf_addr, uint32_t txlen)
{
    // send result to the host, blocking wait
    kdrv_status_t usb_sts = kdrv_usbd2_bulk_send(g_enp_in, (void *)buf_addr, txlen, 0);
    if (usb_sts != KDRV_STATUS_OK)
    {
        if (usb_sts == KDRV_STATUS_USBD_TRANSFER_TERMINATED)
        {
            if (kdrv_usbd2_get_link_status() == USBD2_STATUS_DISCONNECTED)
            {
                ufifo_dbg("[%s] bulk send error due to disconnection\n", __FUNCTION__);
                return -1;
            }
            else
            {
                ufifo_dbg("[%s] bulk send is terminated\n", __FUNCTION__);
            }
        }
        else
        {
            ufifo_dbg("[%s] bulk send error, status %d\n", __FUNCTION__, usb_sts);
            return -2;
        }
    }

    return 0;
}

static void _tx_result_usb_thread(void *arg)
{
tx_wait_usb_connection:

    result_send_state = 0;

    osThreadFlagsWait(FLAG_USB_IS_CONNECTED, osFlagsWaitAny, osWaitForever);

    ufifo_dbg("[send] ready !\n");

    osStatus_t fifo_sts;
    uint32_t buf_addr;

    while (1)
    {
        result_send_state = 1;

        // get result data from queue blocking wait
        fifo_sts = dual_fifo__dequeue_data(gResult_dfifo, &buf_addr, osWaitForever);
        if (fifo_sts != osOK)
            break;

        uint32_t txlen = *(uint32_t *)buf_addr;

        ufifo_dbg("[send] sending data from buf 0x%x size %d\n", buf_addr, txlen);

        result_send_state = 2;

        int result = _send_data_via_usb(buf_addr, txlen);
        if (result == -1)
        {
            dual_fifo__put_free_buffer(gResult_dfifo, &buf_addr, osWaitForever);
            goto tx_wait_usb_connection;
        }
        else if (result == -2)
            break;

        result_send_state = 3;

        // return free buf back to queue
        fifo_sts = dual_fifo__put_free_buffer(gResult_dfifo, &buf_addr, osWaitForever);
        if (fifo_sts != osOK)
            break;
    }

    ufifo_dbg("[send] stopped !\n");
}

////////////////////////////////////////////////////////////

int fifoQ_usb_init(dual_fifo_t image_dfifo, dual_fifo_t result_dfifo, uint8_t enpoint_in, uint8_t enpoint_out)
{
    usbd_queue_rx_tid = osThreadNew(_rx_image_usb_thread, NULL, NULL);
    usbd_queue_tx_tid = osThreadNew(_tx_result_usb_thread, NULL, NULL);

    if(usbd_queue_rx_tid == NULL || usbd_queue_tx_tid == NULL)
        return -1;

    g_enp_in = enpoint_in;
    g_enp_out = enpoint_out;

    gImg_dfifo = image_dfifo;
    gResult_dfifo = result_dfifo;

    return 0;
}
