#ifndef __MODEL_MGR_H__
#define __MODEL_MGR_H__

#include <stdint.h>
#include "cmsis_os2.h"
#include "ipc.h"
#include "base.h"
#include "model_type.h"
#include "common.h"
#include "kdp_flash_table.h"

/* a term means ALL_MODELS */
#define KMDW_MODEL_ALL_MODELS  -1

#define MODEL_MGR_FLASH_ADDR_MODEL_COUNT        (KDP_FLASH_FW_INFO_ADDR)
#define MODEL_MGR_FLASH_ADDR_MODEL_INFO         (MODEL_MGR_FLASH_ADDR_MODEL_COUNT + 4)

/* These crc values will ONLY change if model is changed in FLASH */
#define MODEL_INFO_SUM32_VALUE  0x00eee074

#ifndef FLASH_MODEL_HAS_CRC_CHECK
#define MODEL0_FD_SUM32_VALUE   0xf9e571d8
#define MODEL1_LM_SUM32_VALUE   0x3f96c4fc
#define MODEL3_FR_SUM32_VALUE   0xbc998f45
#endif

enum kmdw_model_rc {
    // 0 - 9 is reserved for ncpu return
    // defined in ipc.h
    // IMAGE_STATE_INACTIVE == 0
    // IMAGE_STATE_ACTIVE == 1
    // IMAGE_STATE_NPU_DONE == 2
    // IMAGE_STATE_DONE == 3
    KMDW_MODEL_RUN_RC_ABORT = 10,
    KMDW_MODEL_RUN_RC_ERROR = 11,
    KMDW_MODEL_RUN_RC_END
};

struct img_run_time_s {
    uint32_t    round_trip_time;
    uint32_t    pre_proc_time;
    uint32_t    npu_proc_time;
    uint32_t    post_proc_time;
};

struct fw_info_s {
    uint32_t                model_count;
    struct kdp_model_s      models[1];
};

struct fw_info2_s {
    uint32_t                model_dram_addr_end;
    uint32_t                model_total_size;
    uint32_t                model_checksum;
};

/* ############################
 * ##    Public Functions    ##
 * ############################ */

/**
 * @brief Init model functionality
 */
void kmdw_model_init(void);

/**
 * @brief A wrapper of load_model
 * @param model_index_p: model info index,
 *                    0-n: info_index of model to load
 *                    -1 means to load all models
 * @return always 0
 */
int32_t kmdw_model_load_model(int8_t model_info_index_p);

/**
 * @brief A wrapper of load_model_info
 * @param [in] is_model_from_ddr: if model is from ddr/host command
 */
void kmdw_model_reload_model_info(bool from_ddr);

/**
 * @brief Refresh all models
 */
void kmdw_model_refresh_models(void);

/**
 * @brief Output model_info of specified index
 * @param[in] idx_p the index of programmed models
 * @return model_info defined in ipc.h
 */
struct kdp_model_s* kmdw_model_get_model_info(int idx_p);

/**
 * @brief Output total number of models and all model IDs
 * @param[in] from_ddr: if models are in ddr or flash
 * @return total number of models and all model IDs
 */
uint32_t *kmdw_model_get_all_model_info(bool from_ddr);

/**
 * @brief Output crc value
 * @param[in] from_ddr: if models are in ddr or flash
 * @return crc value
 */
uint32_t *kmdw_model_get_crc(bool from_ddr);

/**
 * @brief Output the status of if models have been loaded
 * @return the status of if models have been loaded
 */
uint32_t kmdw_model_is_model_loaded(void);

/**
 * @brief Specify output address for model run in ncpu/npu
          !!! must be called after kmdw_model_config_model()
 * @return always 0
 */
int32_t kmdw_model_config_result(osEventFlagsId_t result_evt, uint32_t result_evt_flag);

/**
 * @brief Config model image
 * @param img_cfg image config
 * @param crop_box image crop config
 * @param pad_values image padding config
 * @param ext_param extra param
 */
void kmdw_model_config_img(struct kdp_img_cfg *img_cfg, struct kdp_crop_box_s *crop_box,
                           struct kdp_pad_value_s *pad_values, void *ext_param);

/**
 * @brief Get raw image config
 * @param idx image index
 * @return raw image config
 */
struct kdp_img_raw_s* kmdw_model_get_raw_img(int idx);

/**
 * @brief Run model
 * @param tag model tag
 * @param output model output
 * @param model_type model type
 * @param dme DME mode
 * @return kmdw_model_rc
 */
int kmdw_model_run(const char *tag, void *output, uint32_t model_type, bool dme);

/**
 * @brief Abort model execution
 */
void kmdw_model_abort(void);

/**
 * @brief Get round-trip/pre/npu/post times
 * @param img_idx raw image index
 * @param run_time a pointer of struct img_run_time_s to receive time
 * @return void
 */
void kmdw_model_get_run_time(int img_idx, struct img_run_time_s *run_time);

#ifdef EMBED_CMP_NPU
/**
 * @brief A wrapper of adding or update specific model configuration
 * @param model_type model type
 * @param cmd_len model command length
 * @param wt_len model weight length
 * @param input_len input data length
 * @param output_len output data length
 * @param setup_len setup data length
 * @param cmd_mem_addr model command memory address
 * @param wt_mem_addr model weight memory address
 * @param input_mem_addr input data memory address
 * @param output_mem_addr output data memory address
 * @param setup_mem_addr setup data memory address
 * @return int8_t  0: config ok
 *                -1: model not found
 */
int8_t kmdw_model_add_update_model(uint32_t model_type, 
                                   int cmd_len, int wt_len, int input_len, int output_len, int setup_len,
                                   uint32_t cmd_mem_addr, uint32_t wt_mem_addr, 
                                   uint32_t input_mem_addr, uint32_t output_mem_addr, uint32_t setup_mem_addr);

#endif // EMBED_CMP_NPU

#if DEBUG

/**
 * @brief Dump model debug info
 */
void kmdw_model_dump_model_info(void);

#endif // DEBUG

#endif
