/*
 * Kneron Model API Manager
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#include <stdlib.h>
#include <string.h>
#include "kdrv_ncpu.h"
#include "kmdw_model.h"
#include "kmdw_console.h"       /*for dbg_msg */
#include "kmdw_memxfer.h"       /*for flash access */
#include "kmdw_memory.h"
#include "kdp_ddr_table.h"      /*for KDP_DDR_MODEL_INFO_TEMP*/
#include "kdp_flash_table.h"    /*for KDP_DDR_MODEL_INFO_TEMP*/
#include "scu_ipc.h"            /*for NCPU triggering */
#include "kdp_crc.h"
#include "kdev_flash.h"
#include "kdrv_clock.h"         /* for kdrv_delay_us() */

#define DEBUG 0
#define ENABLE_CHKSUM_ON_MODEL 0

#define FLAG_MODELMGR_ABORT      BIT(29)  // Event flag to notify abort
#define FLAG_MODELMGR_FROM_NCPU  BIT(30)  // Event flag to know NCPU is done
#define FLAG_MODELMGR_FROM_NPU   BIT(28)  // Event flag to know NPU is done

#ifdef EMBED_CMP_NPU
  /* the following is for specific dense model wt/cmd mem modification */
  /*================================================*/
  #define WT_DATA_SIZE_BYTE          272
  
  #define CONF_QUEUE                 0
  #define GETW_QUEUE                 3
  
  #define CONF_GETW0_CMD_OFFSET      0x0038
  #define CONF_WDMA0_DST0_CMD_OFFSET 0x00f0
  
  #define ACL_NPU_GETW0              0x2e
  #define ACH_NPU_GETW0              0x2f
  #define ACL_NPU_WDMA0_DST0         0x36
  #define ACH_NPU_WDMA0_DST0         0x37
  
  #define MASK_2                     0x0003
  #define MASK_10                    0x03FF
  #define MASK_16                    0x00FFFF
  
  #define VAL_ACL(x)                 (((x)&0xffff))
  #define VAL_ACH(x)                 (((x) >> 16) & 0xffff)
  
  #define SetBitsVal(tgt, val, mask, offset) \
      ((tgt) &= ~((mask) << (offset)));      \
      ((tgt) |= (((val) & (mask)) << (offset)))
  
  /*================================================*/
#endif // EMBED_CMP_NPU

#define ALIGN_16(n)                ((n + 16 - 1) & ~(16 - 1))

extern const struct s_kdp_memxfer kdp_memxfer_module;

typedef struct {
    uint32_t n_model_count;            // model count
    struct kdp_model_s *p_model_info;  // save model info generated by compiler
    uint8_t *pn_is_model_loaded_table; // flag table to indicate if model is loaded
    uint32_t n_ddr_addr_model_end;     // DDR address of model end = user data start
    int32_t n_model_slot_index;        // scpu_to_ncpu->model_slot_index
} kmdw_model_data_t;

static kmdw_model_data_t s_model_data;

typedef struct {
    int32_t raw_img_idx;
    osEventFlagsId_t evt_caller;        // event to know/control ncpu
    uint32_t        caller_e;
    osEventFlagsId_t evt_result;        // event to know/control npu
    uint32_t        result_e;
} kmdw_img_data_t;

static kmdw_img_data_t s_img_data[IPC_IMAGE_ACTIVE_MAX];
static int32_t current_ipc_idx = 0;
static uint32_t fw_info_addr;   // Memory for reading fw_info from flash

/* ############################
 * ##    Static Functions    ##
 * ############################ */

#ifdef EMBED_CMP_NPU
static inline void udt_conf_cmd(void *cmd_addr, int reg_idx, uint16_t val16b, int queue)
{
    uint32_t val = 0x80000000;

    SetBitsVal(val, queue, MASK_2, 26);
    SetBitsVal(val, reg_idx, MASK_10, 16);
    SetBitsVal(val, val16b, MASK_16, 0);

    memcpy(cmd_addr, &val, sizeof(val));
}

static void udt_npu_model_mem(uint32_t wt_addr, uint32_t out_addr, void *cmd_addr)
{
    uint32_t *dst = (uint32_t *)((char *)cmd_addr + CONF_GETW0_CMD_OFFSET);
    udt_conf_cmd(dst,     ACL_NPU_GETW0, VAL_ACL(wt_addr), GETW_QUEUE);
    udt_conf_cmd(dst + 1, ACH_NPU_GETW0, VAL_ACH(wt_addr), GETW_QUEUE);

    dst = (uint32_t *)((char *)cmd_addr + CONF_WDMA0_DST0_CMD_OFFSET);
    udt_conf_cmd(dst,     ACL_NPU_WDMA0_DST0, VAL_ACL(out_addr), CONF_QUEUE);
    udt_conf_cmd(dst + 1, ACH_NPU_WDMA0_DST0, VAL_ACH(out_addr), CONF_QUEUE);
}
#endif // EMBED_CMP_NPU

static int flash_wait_ready(int timeout_ms)
{
    kdev_flash_status_t flash_status;
    int i;

    for (i = 0; i < timeout_ms; i++) {
        flash_status = kdev_flash_get_status();
        if (flash_status.busy == 0) break;
        kdrv_delay_us(1*1000);
    }
    if (i == timeout_ms) i = -1;  // we have timed out
    return i;
}

/**
 * @brief convert modeltype to modelInfo slot index
 * @param model_type_p: model type (defined in model_type.h)
 * @return modelInfo model index
 *         -1 means not such modeltype in flash
 */
static int8_t model_type_to_slot_index(uint32_t model_type_p)
{
    int i;
    for(i=0 ; i < s_model_data.n_model_count; i++) {
        if(s_model_data.p_model_info[i].model_type == model_type_p)
            return i;
    }

    return -1;
}

/**
 * @brief load model information generated by compiler
 * @param [in] is_model_from_ddr: if model is from ddr/host command
 * @param [in] is_reload        : is force reload
 * @return model count
 *         0 means loaded
 */
static int32_t load_model_info(bool from_ddr, bool reload)
{
    uint32_t offset;
#if ENABLE_CHKSUM_ON_MODEL
    u32 sum32;
    u8 *addr;
#endif

    if (s_model_data.p_model_info && !reload) {
        return 0;
    }

    if (reload) {
        if (reload && s_model_data.p_model_info) {
            free(s_model_data.p_model_info);
            s_model_data.p_model_info = 0;
        }
        if (reload && s_model_data.pn_is_model_loaded_table) {
            free(s_model_data.pn_is_model_loaded_table);
            s_model_data.pn_is_model_loaded_table = 0;
        }
    }

    // load model Info
    if (from_ddr) {
        uint32_t base_addr = KDP_DDR_MODEL_INFO_TEMP;
        // Use the version number for new fw_info structure. Model number is in use for dynamic model execution
        if (*(uint32_t*)(base_addr + 8) == 0) {
            return 0; //error, model_info is not ready
        }

        // get model count
        s_model_data.n_model_count = *(uint32_t*)(base_addr);
        dbg_msg("[DBG] model info: model count:%d\n", s_model_data.n_model_count);

        // get model info
        if (s_model_data.p_model_info != 0) {
            free(s_model_data.p_model_info);
            s_model_data.p_model_info = 0;
        }

        s_model_data.p_model_info = (struct kdp_model_s*)malloc(s_model_data.n_model_count * sizeof(struct kdp_model_s));
        if (s_model_data.p_model_info == 0) {
            err_msg("[ERR] insufficent memory for model information\n");
            return 0;
        }

        memcpy(s_model_data.p_model_info, (const void*)(base_addr+4),
               sizeof(struct kdp_model_s)*s_model_data.n_model_count);

        // get ddr model end addr
        offset = sizeof(struct kdp_model_s) * s_model_data.n_model_count;
        s_model_data.n_ddr_addr_model_end = *(uint32_t*)(base_addr + 4 + offset);
        dbg_msg("[DBG] modelInfo: DDR end address: 0x%x\n", s_model_data.n_ddr_addr_model_end);
    } else { // models are stored in flash
        kdp_memxfer_module.init(MEMXFER_OPS_DMA, MEMXFER_OPS_DMA);
        // get model count
        kdp_memxfer_module.flash_to_ddr((uint32_t)&(s_model_data.n_model_count), MODEL_MGR_FLASH_ADDR_MODEL_COUNT, 4);
        if (s_model_data.n_model_count == 0xFFFFFFFF) {
            info_msg("[info] model is not in flash!!\n");
            s_model_data.n_model_count = 0;
            free(s_model_data.p_model_info);
            s_model_data.p_model_info = 0;
            return 0;
        } else {
            dbg_msg("[DBG] model info: model count:%d\n", s_model_data.n_model_count);
        }

        // get model info
        if (s_model_data.p_model_info != 0) {
            free(s_model_data.p_model_info);
            s_model_data.p_model_info = 0;
        }

#ifdef EMBED_CMP_NPU
        // reserve one more room for embedding compare model
        int model_cnt = s_model_data.n_model_count + 1;
#else
        int model_cnt = s_model_data.n_model_count;
#endif
        s_model_data.p_model_info = (struct kdp_model_s*)malloc(model_cnt * sizeof(struct kdp_model_s));
        if (s_model_data.p_model_info == 0) {
            err_msg("[ERR] insufficent memory for model information\n");
            return 0;
        }
        kdp_memxfer_module.flash_to_ddr((uint32_t)(s_model_data.p_model_info),
                                        MODEL_MGR_FLASH_ADDR_MODEL_INFO,
                                        sizeof(struct kdp_model_s)*s_model_data.n_model_count);

#if ENABLE_CHKSUM_ON_MODEL
        addr = (u8 *)s_model_data.p_model_info;
        sum32 = kdp_gen_sum32(addr, sizeof(struct kdp_model_s)*s_model_data.n_model_count);
        if(MODEL_INFO_SUM32_VALUE != sum32) {
            critical_msg("[ERR] model_info wrong sum 0x%08x/correct 0x%x!!!~~~\n",
                         sum32, MODEL_INFO_SUM32_VALUE);
        }
#endif
        // get ddr model end addr
        offset = sizeof(struct kdp_model_s) * s_model_data.n_model_count;
        kdp_memxfer_module.flash_to_ddr((uint32_t)&(s_model_data.n_ddr_addr_model_end),
                                        (uint32_t)(MODEL_MGR_FLASH_ADDR_MODEL_INFO + offset), 4);
        dbg_msg("[DBG] modelInfo: DDR end address: 0x%x\n", s_model_data.n_ddr_addr_model_end);

        // load model info for query
        kdev_flash_readdata(KDP_FLASH_FW_INFO_ADDR,
                            (u8 *)KDP_DDR_MODEL_INFO_TEMP, KDP_FLASH_FW_INFO_SIZE);

        int ret = flash_wait_ready(100);  // wait for the read operation to complete
        dbg_msg("Read Flash Sector Wait Time -> %d\r\n", ret);
        if (ret < 0) {
            err_msg("[ERR] Read Flash Sector failed, ret -> %d\r\n", ret);
        }
    }

    s_model_data.pn_is_model_loaded_table =
        (uint8_t*)malloc(s_model_data.n_model_count * sizeof(uint8_t));

    // for support of dynamic model execution
    *(uint32_t*)(KDP_DDR_MODEL_INFO_TEMP + 8) = 0; //trick: we will check the work to see if model_info is uploaded

    return s_model_data.n_model_count;
}

/**
 * @brief load specific model by model info index (the order in flash)
 * @param model_index_p: model info index
 * @return always 0
 */
static int32_t load_model(uint8_t model_index_p/*starts from 0*/)
{
    uint32_t ddr_addr_model_pool; //start point = the 1st model cmd.bin
    uint32_t ddrAddr_offset;
    uint32_t flAddr;
    uint32_t nLenToLoad;
#if ENABLE_CHKSUM_ON_MODEL
    u32 sum32;
    u8 *addr;
#endif
    struct kdp_model_s *p_model;

    if(s_model_data.pn_is_model_loaded_table[model_index_p] == 1 )
        return 0; //model has been loaded
    else
        s_model_data.pn_is_model_loaded_table[model_index_p] = 1;

    //load model with (index=model_index_p) from flash to DDR
    ddr_addr_model_pool = s_model_data.p_model_info[0].cmd_mem_addr; //start point = the 1st model cmd.bin
    p_model = &(s_model_data.p_model_info[model_index_p]);

    //load cmd + weight + setup together
    ddrAddr_offset = p_model->cmd_mem_addr - ddr_addr_model_pool;
    flAddr = KDP_FLASH_ALL_MODELS_ADDR + ddrAddr_offset;
    nLenToLoad = ALIGN_16(p_model->cmd_mem_len) +
                 ALIGN_16(p_model->weight_mem_len) +
                 ALIGN_16(p_model->setup_mem_len);

    //todo: if model is from DDR/Host

    //model from flash
    kdp_memxfer_module.flash_to_ddr(p_model->cmd_mem_addr, flAddr, nLenToLoad);
#if ENABLE_CHKSUM_ON_MODEL
    dbg_msg("[INFO] modelmgr: load model[%d] at ddr[0x%x], sum32:0x%x\n",
            model_index_p, p_model->cmd_mem_addr, sum32);

    /*Calculate & compare check value */
    addr = (u8 *)p_model->cmd_mem_addr;
    sum32 = kdp_gen_sum32(addr, nLenToLoad);
#ifdef FLASH_MODEL_HAS_CRC_CHECK  // no check field is defined yet, this section won't work
    if (p_model->crc16 != crc16) {
        critical_msg("[ERR] model[%d] wrong crc 0x%x/correct 0x%x (addr:0x%p, size:%d)!!!~~~\n",
                     model_index_p, crc16, p_model->crc16, addr, nLenToLoad);
        //osDelay(1000);
    }
#else  // this is only okay for default flash image
    switch(model_index_p) {
    case 0: //FD
        if(MODEL0_FD_SUM32_VALUE != sum32) {
            critical_msg("[ERR] model[%d] wrong sum32 0x%x/correct 0x%x (addr:0x%p, size:%d)!!!~~~\n",
                         model_index_p, sum32, MODEL0_FD_SUM32_VALUE, addr, nLenToLoad);
        }
        break;
    case 1: //LM
        if(MODEL1_LM_SUM32_VALUE != sum32) {
            critical_msg("[ERR] model[%d] wrong sum32 0x%x/correct 0x%x (addr:0x%p, size:%d)!!!~~~\n",
                         model_index_p, sum32, MODEL1_LM_SUM32_VALUE, addr, nLenToLoad);
        }
        break;
    case 3: //FR
        if(MODEL3_FR_SUM32_VALUE != sum32) {
            critical_msg("[ERR] model[%d] wrong sum32 0x%x/correct 0x%x (addr:0x%p, size:%d)!!!~~~\n",
                         model_index_p, sum32, MODEL3_FR_SUM32_VALUE, addr, nLenToLoad);
        }
        break;
    }
#endif
#endif
    return 0;
}

/**
 * @brief specify model information, load model info, load model
 *        !!!will keep p_model_slot_index
 * @param [in] model_type_p: model unique ID defined by Kneron
 * @param [in] model_from_ddr: is model from ddr or host command
 * @return model_slot_index(requested by NCPU/NPU)
 */
static int32_t config_model(uint32_t model_type, bool model_from_ddr)
{
    int model_info_idx; //limitation (hard coded in flash)
    int model_idx;

    //check if model info is loaded
    load_model_info(model_from_ddr, false);

    //check if specified model type is loaded
    if( model_from_ddr == 0 ) {
        /* Special model not in DDR but in ncpu */
        if (model_type == KNERON_2D_LIVENESS_224_224_3) {
            model_idx = 3;
            model_info_idx = 4;

            goto model_common;
        }

        model_info_idx = model_type_to_slot_index(model_type);
        if(model_info_idx == -1) {
            err_msg("[ERR] model_type[%d] is not programmed in flash\n", model_type);
            return -1;
        }
        load_model(model_info_idx);

        //todo: need to remove the following hard code
        model_idx = model_info_idx;
    } else {
        model_info_idx = model_type_to_slot_index(model_type);
        if(model_info_idx == -1) {
            err_msg("[ERR] model_type[%d] is not found in DDR\n", model_type);
            return -1;
        }
        model_idx = model_info_idx;
    }

model_common:
    //keep in kdp_model_mgr static area
    s_model_data.n_model_slot_index = model_idx;

    kdrv_ncpu_set_model(s_model_data.p_model_info, model_info_idx, model_idx);
    kdrv_ncpu_set_model_active(current_ipc_idx, model_idx);

    struct kdp_img_raw_s *raw_img = kmdw_model_get_raw_img(s_img_data[current_ipc_idx].raw_img_idx);

    if (raw_img->format & IMAGE_FORMAT_PARALLEL_PROC) {
        /* Allocate parallel output buffer */
        struct scpu_to_ncpu_s* comm_out = kdrv_ncpu_get_output();
        uint32_t addr_parallel = comm_out->output_mem_addr2;
        if (addr_parallel == 0) {
            uint32_t addr1 = comm_out->models[model_idx].output_mem_addr;
            uint32_t addr2 = comm_out->models[model_idx].buf_addr;
            uint32_t len = comm_out->models[model_idx].output_mem_len;
            if (addr1 == addr2) {
                // Old memory layout, use working buffer length
                len = comm_out->models[model_idx].buf_len;
            }

            // reserve more space for larger model output (ty_416x416 need 0x40bf0)
            len = 0x50000;

            addr_parallel = kmdw_ddr_reserve(len);
            comm_out->output_mem_addr2 = addr_parallel;
            comm_out->output_mem_len2 = len;

            dbg_msg("Parallel buffer: model [%d %d], len %d, addr 0x%x for 0x%x\n",
                model_idx, model_type, len, addr_parallel,
                kdrv_ncpu_get_output()->models[model_idx].output_mem_addr);
        }
    }

    if (model_type == KNERON_FD_MBSSD_200_200_3 || model_type == KNERON_FD_MASK_MBSSD_200_200_3 || \
        model_type == KNERON_OD_MBSSD || model_type == KNERON_PD_MBSSD) {
        struct scpu_to_ncpu_s* comm_out = kdrv_ncpu_get_output();
        uint32_t len = 0x5000;
        if (comm_out->output_mem_addr3 == 0) {
            comm_out->output_mem_addr3 = kmdw_ddr_reserve(len*sizeof(uint32_t));
        }
        // clear output_mem_addr3 first uint32_t content
        *(uint32_t*) comm_out->output_mem_addr3 = 0;
    }

    return model_idx;
}

/**
 * @brief run model according to config settings
 *         !!! must be called after kapp_config_model_image()
 * @return status defined in NCPU
 */
static int32_t run_model(void)
{
    int active_idx = current_ipc_idx;
    int raw_img_idx = s_img_data[active_idx].raw_img_idx;
    struct kdp_img_raw_s *p_raw_image = kmdw_model_get_raw_img(raw_img_idx);
    uint32_t flags, wait_evt;
    uint32_t is_abort = 0;

    // Start time for ncpu/npu round trip
    p_raw_image->tick_start = osKernelGetTickCount();

    if (s_img_data[active_idx].evt_caller == NULL)
        s_img_data[active_idx].evt_caller = osEventFlagsNew(0);

    // set notify for job done
    if (s_img_data[active_idx].evt_result) {
        /* Result event already set. Let's do local event for parallel. */
        wait_evt = FLAG_MODELMGR_FROM_NPU;
    } else {
        wait_evt = FLAG_MODELMGR_FROM_NCPU;
    }

    dbg_msg("<Run-Model> wait %d[%d] evt %x\n", raw_img_idx, active_idx, s_img_data[active_idx]);

    //trigger ncpu/npu
    kdrv_ncpu_trigger_int(active_idx);

    //check abort signal
    flags = osEventFlagsWait(s_img_data[active_idx].evt_caller,
                             FLAG_MODELMGR_ABORT,
                             osFlagsWaitAll, 0);
    if( flags != osFlagsErrorResource ) {
        osEventFlagsClear(s_img_data[active_idx].evt_caller, FLAG_MODELMGR_ABORT);
        is_abort = 1;
    }

    //wait for finish of current task
    s_img_data[active_idx].caller_e = wait_evt;
    flags = osEventFlagsWait(s_img_data[active_idx].evt_caller,
                             wait_evt,
                             osFlagsNoClear, osWaitForever);
    if (flags != wait_evt)
        err_msg("[run_model] 1+ events 0x%08x (%d[%d] expected)\n", flags, wait_evt, active_idx);
    else
        dbg_msg("<Run-Model> got: %d[%d]\n", raw_img_idx, active_idx);

    osEventFlagsClear(s_img_data[active_idx].evt_caller, wait_evt);

    if( 1 == is_abort ) {
        dbg_msg("[DBG] abort after n_model_slot_index = %d\n", s_model_data.n_model_slot_index);
        return KMDW_MODEL_RUN_RC_ABORT; //abort
    }

    return kdrv_ncpu_get_input()->img_results[raw_img_idx].status;
}

static int kmdw_model_set_act_img(int idx)
{
    int active_idx;

wait4active:
    active_idx = kdrv_ncpu_set_image_active(idx);
    if (active_idx == -1) {
        // Flow control
        osDelay(5);
        goto wait4active;
    }
    s_img_data[active_idx].raw_img_idx = idx;
    current_ipc_idx = active_idx;
    return active_idx;
}

static void kmdw_model_ipc_handler(int ipc_idx, int state)
{
    int raw_img_idx;
    struct kdp_img_raw_s *p_raw_image;

    if (state == IMAGE_STATE_RECEIVING) {
        if (s_img_data[ipc_idx].evt_result) {
            dbg_msg("[kdr model] ipc_idx: %d, result_e: %d.\n", ipc_idx, s_img_data[ipc_idx].result_e);
            osEventFlagsSet(s_img_data[ipc_idx].evt_result, s_img_data[ipc_idx].result_e);
        } else {
            dbg_msg("[kdr model] ipc_idx: %d, caller_e: %x.\n", ipc_idx, s_img_data[ipc_idx].caller_e);
            osEventFlagsSet(s_img_data[ipc_idx].evt_caller, s_img_data[ipc_idx].caller_e);
        }

        // End time for ncpu/npu round trip
        raw_img_idx = s_img_data[ipc_idx].raw_img_idx;
        p_raw_image = kmdw_model_get_raw_img(raw_img_idx);
        p_raw_image->tick_end = osKernelGetTickCount();
    } else if (state == IMAGE_STATE_ACTIVE){
        dbg_msg("[kdr model] ipc_idx: %d, caller_e: %x\n", ipc_idx, s_img_data[ipc_idx].caller_e);
        osEventFlagsSet(s_img_data[ipc_idx].evt_caller, s_img_data[ipc_idx].caller_e);
    } else {
        err_msg("[ERR] wrong state: %d (ipc_idx %d)\n", state, ipc_idx);
    }
}

/* ############################
 * ##    Public Functions    ##
 * ############################ */

void kmdw_model_init(void)
{
    kdrv_ncpu_initialize(kmdw_model_ipc_handler);
}

int32_t kmdw_model_load_model(int8_t model_info_index_p)
{
    int32_t ret;
    if(s_model_data.p_model_info == 0) {
        load_model_info(false, false);
    }

    // load all models
    if (KMDW_MODEL_ALL_MODELS == model_info_index_p) {
        uint8_t i;
        for (i = 0 ; i < s_model_data.n_model_count ; i++) {
            ret = load_model(i);
        }
    } else { // load specific model
        ret = load_model(model_info_index_p);
        return ret;
    }

    return 1; //error
}

void kmdw_model_reload_model_info(bool from_ddr)
{
    load_model_info(from_ddr, true);
}

void kmdw_model_refresh_models(void) {  // reload all the models again
    uint8_t i;

    for (i = 0 ; i < s_model_data.n_model_count ; i++) {
        if (s_model_data.pn_is_model_loaded_table[i]) {  // if previously loaded
            s_model_data.pn_is_model_loaded_table[i] = 0;
            kmdw_model_load_model(i);  // reload the model again
        }
    }
}

int32_t kmdw_model_config_result(osEventFlagsId_t result_evt, uint32_t result_evt_flag)
{
    int active_idx = current_ipc_idx;

    s_img_data[active_idx].evt_result = result_evt;
    s_img_data[active_idx].result_e = result_evt_flag;

    return 0;
}

void kmdw_model_config_img(struct kdp_img_cfg *img_cfg, struct kdp_crop_box_s *crop_box,
                           struct kdp_pad_value_s *pad_values, void *ext_param)
{
    int act_img_idx = img_cfg->image_buf_active_index;
    struct kdp_img_raw_s *raw_img = kmdw_model_get_raw_img(act_img_idx);

    //keep in kdp_model_mgr static area
    int ipc_active = kmdw_model_set_act_img(act_img_idx);

    raw_img->state = IMAGE_STATE_ACTIVE;
    raw_img->seq_num = 1;

    raw_img->ref_idx = act_img_idx; //todo: need to confirm
    raw_img->input_row = img_cfg->image_row;
    raw_img->input_col = img_cfg->image_col;
    raw_img->input_channel = img_cfg->image_ch;
    raw_img->format = img_cfg->image_format;
    raw_img->image_mem_addr = img_cfg->image_mem_addr;
    raw_img->image_mem_len = img_cfg->image_mem_len;  //useless now

    if (crop_box == NULL) {
        raw_img->params_s.crop_top = 0;
        raw_img->params_s.crop_bottom = 0;
        raw_img->params_s.crop_left = 0;
        raw_img->params_s.crop_right = 0;
    } else {
        raw_img->params_s.crop_top = crop_box->top;
        raw_img->params_s.crop_bottom = crop_box->bottom;
        raw_img->params_s.crop_left = crop_box->left;
        raw_img->params_s.crop_right = crop_box->right;
    }

    if (pad_values == NULL) {
        raw_img->params_s.pad_top = 0;
        raw_img->params_s.pad_bottom = 0;
        raw_img->params_s.pad_left = 0;
        raw_img->params_s.pad_right = 0;
    } else {
        raw_img->params_s.pad_top = pad_values->pad_top;
        raw_img->params_s.pad_bottom = pad_values->pad_bottom;
        raw_img->params_s.pad_left = pad_values->pad_left;
        raw_img->params_s.pad_right = pad_values->pad_right;
    }

    if (ext_param == NULL) {
        memset(raw_img->params_s.params, 0, MAX_PARAMS_LEN*4);
    } else {
        memcpy(raw_img->params_s.params, ext_param, MAX_PARAMS_LEN*4);
    }
}

struct kdp_img_raw_s* kmdw_model_get_raw_img(int idx)
{
    struct scpu_to_ncpu_s *comm_out = kdrv_ncpu_get_output();
    return &(comm_out->raw_images[idx]);
}

int kmdw_model_run(const char *tag, void *output, uint32_t model_type, bool dme)

{
    int model_idx = config_model(model_type, dme);
    if (model_idx < 0) {
        return KMDW_MODEL_RUN_RC_ABORT;
    }

    int img_idx = s_img_data[current_ipc_idx].raw_img_idx;
    struct kdp_img_raw_s *raw_img = kmdw_model_get_raw_img(img_idx);

    raw_img->results[model_idx].result_mem_addr = (uint32_t)output;

    dbg_msg("[INFO] %s:\n", tag);
    dbg_msg(" model_idx = %d\n", model_idx);
    dbg_msg("       (row/col/ch) = %d/%d/%d\n", raw_img->input_row, raw_img->input_col, raw_img->input_channel);
    dbg_msg("             format = 0x%x\n", raw_img->format);
    dbg_msg("  crop(tp/bt/lf/rt) = %d/%d/%d/%d\n", raw_img->params_s.crop_top, raw_img->params_s.crop_bottom,
            raw_img->params_s.crop_left, raw_img->params_s.crop_right);
    dbg_msg("*ext_params(first 4)= %d/%d/%d/%d\n", raw_img->params_s.params[0], raw_img->params_s.params[1],
            raw_img->params_s.params[2], raw_img->params_s.params[3]);
    dbg_msg("            ref_idx = %d\n", raw_img->ref_idx);
    dbg_msg("         model type = %d\n", model_type);
    dbg_msg("        image  addr = 0x%x\n", raw_img->image_mem_addr);
    dbg_msg("        output addr = 0x%x\n", raw_img->results[model_idx].result_mem_addr);

    return run_model();
}

/**
 * @brief abort NCPU/NPU operations (stop sending task to ncpu/npu)
 *
 * @return n/a
 */
void kmdw_model_abort(void)
{
    int active_idx = current_ipc_idx;

    if( 0 == s_img_data[active_idx].evt_caller)
        return;

    osEventFlagsSet(s_img_data[active_idx].evt_caller, FLAG_MODELMGR_ABORT);
}

struct kdp_model_s* kmdw_model_get_model_info(int model_idx_p)
{
    if (s_model_data.p_model_info == 0) {
        return NULL;
    } else if (model_idx_p >= s_model_data.n_model_count) {
        return NULL;
    } else {
        return &(s_model_data.p_model_info[model_idx_p]);
    }
}

void kmdw_model_get_run_time(int img_idx, struct img_run_time_s *run_time)
{
    struct kdp_img_raw_s *p_raw_image;

    if (run_time == NULL)
        return;

    p_raw_image = kmdw_model_get_raw_img(img_idx);

    run_time->round_trip_time = p_raw_image->tick_end - p_raw_image->tick_start;
    run_time->pre_proc_time = p_raw_image->tick_end_pre - p_raw_image->tick_start_pre;
    run_time->npu_proc_time = p_raw_image->tick_end_npu - p_raw_image->tick_start_npu;
    run_time->post_proc_time = p_raw_image->tick_end_post - p_raw_image->tick_start_post;
}

uint32_t kmdw_model_is_model_loaded(void)
{
    uint32_t ret = 0;
    if (s_model_data.p_model_info && s_model_data.pn_is_model_loaded_table) {
        ret = 1;
    }

    return ret;
}

uint32_t *kmdw_model_get_all_model_info(bool from_ddr)
{
    uint32_t *p_model_id = NULL;
    uint8_t i;
    uint32_t model_cnt, model_id;
    uint32_t *fw_info = NULL;
    uint32_t model_info_size = sizeof(struct kdp_model_s) >> 2; // size for 1 model

    if (from_ddr) {
        if (kmdw_model_is_model_loaded()) {
            fw_info = (uint32_t *)KDP_DDR_MODEL_INFO_TEMP;
        }
    } else {
        if (fw_info_addr == 0) {
            uint32_t len = 0x1000;
            fw_info_addr = kmdw_ddr_reserve(len*sizeof(u8));
            *(uint32_t*)fw_info_addr = 0;
        }
        fw_info = (uint32_t *)fw_info_addr;
        if (fw_info) {
            kdev_flash_readdata(KDP_FLASH_FW_INFO_ADDR,
                                (u8 *)fw_info, KDP_FLASH_FW_INFO_SIZE );

            int ret = flash_wait_ready(100);  // wait for the read operation to complete
            dbg_msg("Read Flash Sector Wait Time -> %d\r\n", ret);
            if (ret < 0) {
                err_msg("[ERR] Read Flash Sector failed, ret -> %d\r\n", ret);
            }
        } else {
            err_msg("[ERR] insufficent memory for reading data from flash\n");
            fw_info = NULL;
        }
    }

    if (fw_info) {
        model_cnt = fw_info[0];  // get number of models

        if (model_cnt > 0xFF) {
            if (from_ddr)
                dbg_msg("Model is not in DDR\n");
            else
                dbg_msg("Model is not in Flash\n");
        } else {
            p_model_id = (uint32_t *)calloc(1+model_cnt, sizeof(uint32_t));

            if (p_model_id == 0) {
                err_msg("[ERR] insufficent memory for model info\n");
                p_model_id = NULL;
            } else {
                p_model_id[0] = model_cnt;
                dbg_msg("Model Count = %d\n", model_cnt);

                for (i = 0 ; i < model_cnt; i++) {
                    model_id = fw_info[1+i*model_info_size];
                    dbg_msg("Model ID %d\n", model_id);
                    if (model_id > 0xffff) {
                        free(p_model_id);
                        if (from_ddr)
                            dbg_msg("Model is not in DDR\n");
                        else
                            dbg_msg("Model is not in Flash\n");
                        p_model_id = NULL;
                        break;
                    }

                    p_model_id[i+1] = model_id;
                }
            }
        }
    }

    return p_model_id;
}

uint32_t *kmdw_model_get_crc(bool from_ddr)
{
    uint32_t *p_crc = NULL;
    uint32_t model_cnt;
    uint32_t *fw_info = NULL;
    uint32_t model_info_size = sizeof(struct kdp_model_s) >> 2; // size for 1 model

    if (from_ddr) {
        if (kmdw_model_is_model_loaded()) {
            fw_info = (uint32_t *)KDP_DDR_MODEL_INFO_TEMP;
        }
    } else {
        if (fw_info_addr == 0) {
            uint32_t len = 0x1000;
            fw_info_addr = kmdw_ddr_reserve(len*sizeof(u8));
            *(uint32_t*)fw_info_addr = 0;
        }
        fw_info = (uint32_t *)fw_info_addr;
        if (fw_info) {
            kdev_flash_readdata(KDP_FLASH_FW_INFO_ADDR,
                                (u8 *)fw_info, KDP_FLASH_FW_INFO_SIZE);

            int ret = flash_wait_ready(100);  // wait for the read operation to complete
            dbg_msg("Read Flash Sector Wait Time -> %d\r\n", ret);
            if (ret < 0) {
                err_msg("[ERR] Read Flash Sector failed, ret -> %d\r\n", ret);
            }
        } else {
            err_msg("[ERR] insufficent memory for reading data from flash\n");
            fw_info = NULL;
        }
    }

    if (fw_info) {
        model_cnt = fw_info[0];  // get number of models

        if (model_cnt > 0xFF) {
            if (from_ddr)
                dbg_msg("Model is not in DDR\n");
            else
                dbg_msg("Model is not in Flash\n");
        } else {
            p_crc = (uint32_t *)calloc(1, sizeof(uint32_t));

            if (p_crc == 0) {
                err_msg("[ERR] insufficent memory for model info\n");
                p_crc = NULL;
            } else {
                p_crc[0] = fw_info[1+model_cnt*model_info_size + 2]; // first crc
                dbg_msg("First CRC = 0x%x\n", p_crc[0]);
            }
        }
    }

    return p_crc;
}

#ifdef EMBED_CMP_NPU

int8_t kmdw_model_add_update_model(uint32_t model_type, 
                                   int cmd_len, int wt_len, int input_len, int output_len, int setup_len,
                                   uint32_t cmd_mem_addr, uint32_t wt_mem_addr, 
                                   uint32_t input_mem_addr, uint32_t output_mem_addr, uint32_t setup_mem_addr)
{
    int model_info_idx = model_type_to_slot_index(model_type);

    if (model_info_idx < 0) {
        int model_count = s_model_data.n_model_count + 1;
        s_model_data.n_model_count = model_count;
        model_info_idx = model_count - 1;

        s_model_data.p_model_info[model_info_idx].model_type = model_type;

        s_model_data.p_model_info[model_info_idx].cmd_mem_addr = cmd_mem_addr;
        s_model_data.p_model_info[model_info_idx].cmd_mem_len = cmd_len;

        s_model_data.p_model_info[model_info_idx].weight_mem_len = wt_len;
        
        s_model_data.p_model_info[model_info_idx].input_mem_addr = input_mem_addr;
        s_model_data.p_model_info[model_info_idx].input_mem_len = input_len;

        s_model_data.p_model_info[model_info_idx].output_mem_len = output_len;
        s_model_data.p_model_info[model_info_idx].buf_len = output_len;
    
        s_model_data.p_model_info[model_info_idx].setup_mem_addr = setup_mem_addr;
        s_model_data.p_model_info[model_info_idx].setup_mem_len = setup_len;

        s_model_data.pn_is_model_loaded_table[model_info_idx] = 1;
    }

    s_model_data.p_model_info[model_info_idx].weight_mem_addr = wt_mem_addr;
    s_model_data.p_model_info[model_info_idx].output_mem_addr = output_mem_addr;
    s_model_data.p_model_info[model_info_idx].buf_addr = output_mem_addr;
    
    dbg_msg("[%s] model cmd addr: 0x%x\n", __func__, s_model_data.p_model_info[model_info_idx].cmd_mem_addr);
    dbg_msg("[%s] model wt addr: 0x%x\n", __func__, s_model_data.p_model_info[model_info_idx].weight_mem_addr);
    dbg_msg("[%s] model input addr: 0x%x\n", __func__, s_model_data.p_model_info[model_info_idx].input_mem_addr);
    dbg_msg("[%s] model output addr: 0x%x\n", __func__, s_model_data.p_model_info[model_info_idx].output_mem_addr);
    dbg_msg("[%s] model buf addr: 0x%x\n", __func__, s_model_data.p_model_info[model_info_idx].buf_addr);
    dbg_msg("[%s] model setup addr: 0x%x\n", __func__, s_model_data.p_model_info[model_info_idx].setup_mem_addr);

    udt_npu_model_mem(wt_mem_addr, output_mem_addr, (void *)cmd_mem_addr);
    
    return 0;
}

#endif // EMBED_CMP_NPU

#if DEBUG

void kmdw_model_dump_model_info(void)
{
    //kdp_model_mgr_t *mgr = kdp_modelmgr();
    struct kdp_model_s *p_modelInfo = 0;
    uint8_t i;

    dbg_msg("Model info Count = %d\n", s_model_data.n_model_count);

    for (i = 0 ; i < s_model_data.n_model_count ; i++) {
        p_modelInfo = &(kmdw_model_data.p_model_info[i]);
        dbg_msg("Model(%2d) model_type(%3d)/version(%5d):\n",
                (i+1),
                p_modelInfo->model_type, p_modelInfo->model_version);

        dbg_msg("input[%x](sz:%d) -> cmd[%x](sz:%d),weight[%x](sz:%d),setup[%x](sz:%d),buf[%x](sz:%d) -> out[%x](sz:%d)\n",
                (i+1),
                p_modelInfo->input_mem_addr, p_modelInfo->input_mem_len,
                p_modelInfo->cmd_mem_addr,   p_modelInfo->cmd_mem_len,
                p_modelInfo->weight_mem_addr,p_modelInfo->weight_mem_len,
                p_modelInfo->setup_mem_addr, p_modelInfo->setup_mem_len,
                p_modelInfo->buf_addr,       p_modelInfo->buf_len,
                p_modelInfo->output_mem_addr,p_modelInfo->output_mem_len);
    }

    return;
}

#endif // DEBUG
