/**
 * Copyright (C) Kneron Inc. All right reserved
 *
 * @file kapp_db.h
 * @brief Kneron Applicaion  APIs
 */

#ifndef __KAPP_DB_H_
#define __KAPP_DB_H_

#include <stdint.h>
#include "kapp.h"
#include "common.h"

/**
 * @brief FR feature map result size in byte
 */
#define FM_SIZE_BYTE   (FR_FEAT_SIZE * 4)

/** @brief flash address of DB */
#define KAPP_DB_FLASH_ADDR   0x80000

/** @brief DB count. 1 or 2 */
#define DB_NUM  2

/** @brief default threshold for comparision.
 *  range: [0 - 1] means [strict -> loose]*/
#define KAPP_DB_DEFAULT_GUARD_THRESHOLD FID_THRESHOLD

enum kapp_db_type_valid_flag_e{
    TYPE_INVALID = 0,
    TYPE_VALID,
    TYPE_REGISTER,
};

enum kapp_db_status_code_e{
    KAPP_DB_NO_SPACE = 0x200,
    KAPP_DB_ALREADY_SAVED,
    KAPP_DB_DEL_NOT_VALID,
    KAPP_DB_NO_MATCH,
    KAPP_DB_REG_FIRST,
    KAPP_DB_USER_NOT_REG,
    KAPP_DB_DEL_FAIL,
    KAPP_DB_CONFIG_INVALID_VALUE,
    KAPP_DB_CONFIG_OUT_OF_SPACE,
    KAPP_DB_NULL_POINTER
};

typedef struct kapp_db_user_data_s {
    uint16_t user_id_in;
    uint16_t user_id_out;
    uint16_t user_idx;
    uint16_t fm_idx;
    uint16_t del_all;
} kapp_db_user_data_t;

typedef struct kapp_db_config_s {
    uint16_t    db_num;     // number of database
    uint16_t    max_uid;    // max number of user ID in each db
    uint16_t    max_fid;    // max number of feature map  
} kapp_db_config_t;

typedef union kapp_db_config_parameter_u {
    kapp_db_config_t    db_config;      // kapp_db_config_s             [6 Bytes]
    uint32_t            uint32_value;   // config uint32_t parameter    [4 Bytes]
} kapp_db_config_parameter_t;           // total                        [8 Bytes] (32-bits auto align)

/**
 * @brief compare user fd/fr result to user data in ddr
 * @param [in] fdr_addr address of fd/fr feature map data
 * @param [out] *user_id recognized user ID
 * @param [out] *score the score calculated by comparison engine. -1 means invalid
 * @parem [in] thresh_guard comparision threshold for recognition
 * @return compare result and user id where find match feature map data
 */
int32_t kapp_db_compare(
        uint32_t fdr_addr, uint16_t *user_id/*output*/, float *score/*output*/, 
        float thresh_guard);

/**
 * @brief register user fm data
 * @param [in] fdr_addr address of fd/fr feature map data
 * @param [in] user_id use id
 * @param [in] fm_idx face index
 * @return status
 */
int32_t kapp_db_register(uint32_t fdr_addr, uint16_t user_id, uint16_t fm_idx);

/**
 * @brief register user both fm data
 * @param [in] rgb_fdr_addr address of rgb fd/fr feature map data
 * @param [in] nir_fdr_addr address of nir fd/fr feature map data
 * @param [in] user_id use id
 * @param [in] fm_idx face index
 * @return status
 */
int32_t kapp_db_both_register(uint32_t rgb_fdr_addr, uint32_t nir_fdr_addr,  uint16_t user_id, uint16_t fm_idx);

/**
 * @brief Initialize database
 * @param [in] flash_db_addr_p flash address of the database
 * @return registered user count
 */
int32_t kapp_db_init(uint32_t flash_db_addr_p);

/**
 * @brief configurate database
 * @param [in] db_config configuration of the database
 * @return 0 for success, -1 for error
 */
int kapp_db_config(kapp_db_config_t *db_config);

/**
 * @brief get database configuration
 * @param [in] db_config return configuration of the database
 * @return 0 for success, -1 for error
 */
int kapp_db_get_db_config(kapp_db_config_t *db_config);

/**
 * @brief set database version
 * @param [in] db_version version of the database
 */
void kapp_db_set_db_version(uint32_t db_version);

/**
 * @brief get database version
 * @return version of the database 
 */
uint32_t kapp_db_get_db_version(void);

/**
 * @brief get current database index
 * @return current database index
 */
int32_t kapp_db_get_db_index(void);

/**
 * @brief switch between DBs if use mutiple DB setting
 * @param [in] db_index indicate which DB is activated
 */
void kapp_db_switch(uint32_t db_index);

/**
 * @brief get database meta data version
 * @return version of the database meta data 
 */
uint32_t kapp_db_get_db_meta_data_version(void);

/**
 * @brief use two DB to store 2 kinds of FM data
 */
void kapp_db_split_mode(void);

/**
 * @brief register user data to database
 * @param [in] *p_inout_p p_inout_p->user_id user id
 * @return user index in database
 */
int32_t kapp_db_add(kapp_db_user_data_t *p_inout_p);

/**
 * @brief delete user data in ddr and flash
 * @param [in] *p_inout_p p_inout_p->user_id user id, "0" means all
 * @return none
 */
int32_t kapp_db_delete(kapp_db_user_data_t *p_inout_p);

/**
 * @brief output user number and all user id data
 */
int32_t kapp_db_list(void);

/**
 * @brief output one feature map data of one user
 * @param [in] *p_inout_p p_inout_p->user_id
 *                        p_inout_p->image_idx
 */
int32_t kapp_db_upload(kapp_db_user_data_t *p_inout_p);

/**
 * @brief clear all user data not yet valid
 * @return OK or error code
 */
int32_t kapp_db_abort_reg(void);

/**
 * @brief check if a db slot is used
 * @param [in] i slot index
 * @return 1: yes  0:no
 */
int32_t kapp_db_slot_is_used(uint16_t i);

/**
 * @brief get the next available ID slot for register
 * @return slot id for register
 */
int kapp_db_get_available_ID_slot(void);

/**
 * @brief get feature map count for a user index in current db
 * @param [in] user_idx user index
 * @return count for feature map in current db
 */
int kapp_db_get_fm_cnt(int user_idx);

/**
 * @deprecated replaced by kapp_db_get_fm_cnt
 * @brief get feature map 0 count for a user index
 * @param [in] user_idx user index
 * @return count for feature map 0
 */
int kapp_db_get_fm0_cnt(int user_idx);

/**
 * @deprecated replaced by kapp_db_get_fm_cnt
 * @brief get feature map 1 count for a user index
 * @param [in] user_idx user index
 * @return count for feature map 1
 */
int kapp_db_get_fm1_cnt(int user_idx);

/**
 * @brief get feature map for a user index in current db
 * @param [in] user_idx user index
 * @return valid feature map data in current db
 */
int kapp_db_get_fm(int user_idx);

/**
 * @deprecated replaced by kapp_db_get_fm
 * @brief get feature map 0 for a user index
 * @param [in] user_idx user index
 * @return valid feature map 0 data
 */
int kapp_db_get_fm0(int user_idx);

/**
 * @deprecated replaced by kapp_db_get_fm
 * @brief get feature map 1 for a user index
 * @param [in] user_idx user index
 * @return valid feature map 1 data
 */
int kapp_db_get_fm1(int user_idx);

/**
 * @brief get slot type for a user index
 * @param [in] user_idx user index
 * @return type, 0:invalid, 1:valid, 2:register
 */
int kapp_db_get_user_id_slot_type(int user_idx);

/**
 * @brief get user ID for a user index
 * @param [in] user_idx user index
 * @return user ID (could be different from user index)
 */
int kapp_db_get_user_id(int user_idx);

/**
 * @brief get user index for a user ID
 * @param [in] user ID
 * @return user index (or -1 if failed to find user ID)
 */
int kapp_db_get_user_slot(int user_id);

/**
 * @brief Add a single FM to DB,  init the user slot if slot is not valid
 * @param [in] user indx (slot #)
 * @param [in] user id
 * @param [in] u8 pointer to FM data (only works for fm0)
 * @return 0 - success, else error
 */
int kapp_db_add_FM(int user_idx, int user_id, unsigned char *pfm_data);

/**
 * @brief Delete a single FM from DB, delete the user slot if no more FM is left
 * @param [in] user indx (slot #)
 * @param [in] FM indx (slot #)
 * @return 0 - success, else error
 */
int kapp_db_delete_FM(int user_idx, int fm_idx);

/**
 * @brief Get DDR address of a single FM in DB
 * @param [in] user indx (slot #)
 * @param [in] FM indx (slot #)
 * @return 0 - success, else error
 */
int kapp_db_query_FM(int user_idx, int fm_idx);

/**
 * @brief Get user embedding data of a single FM in DB
 * @param [in] user indx (slot #)
 * @param [in] FM indx (slot #)
 * @return user embedding data
 */
int8_t kapp_db_get_user_embed_data(uint16_t user_idx, uint16_t fm_idx);

/**
 * @brief Get wt header data of user embedding
 * @param [in] n/a
 * @return wt header data
 */
uint8_t* kapp_db_get_wt_hdr(void);

/**
 * @brief Get current user embedding data in DDR
 *        the data is created after run register or compare
 * @param [in] n/a
 * @return user embedding data
 */
int8_t kapp_db_get_current_embed_data(void);

/**
 * @brief Calculate similarity of two feature points
 * @param [in] fdr_addr address A of user feature map data
 * @param [in] fdr_addr address B of user feature map data
 * @return similarity score (smaller means more like)
 */
float kapp_db_cal_similarity( uint32_t fdr_addr_1, uint32_t fdr_addr_2);

/**
 * @brief Calculate similarity of two feature points
 * @param [out] **db_ddr_p output ddr address of DB temp
 * @param [out] **db_flash_p flash address of DB
 * @return similarity score (smaller means more like)
 */
int32_t kapp_db_query(uint32_t **db_ddr_p, uint32_t **db_flash_p);

/**
 * @brief compare user fd/fr result to user data in ddr (only normal face)
 * @param fdr_addr address of fd/fr feature map data
 * @param thresh_fid threshold value for similarity
 * @return compare result and user id where find match feature map data
 */
int32_t kapp_db_compare_normal(uint32_t fdr_addr, uint16_t *user_id, float thresh_fid);

/**
 * @brief compare user fd/fr result to itself
 * @param fdr_addr address of fd/fr feature map data
 * @param thresh_fid threshold value for comparing self
 * @return compare result
 */
int32_t kapp_db_compare_self(uint32_t fdr_addr, uint16_t user_id, float thresh_fid);

/**
 * @brief compare user fd/fr result to user data in ddr
 * @param fdr_addr address of fd/fr feature map data
 * @return compare result and user id where find match feature map data
 */
int32_t kapp_db_compare_five_faces_both(uint32_t fdr_addr, uint32_t fdr_nir_addr,
        uint16_t *user_id/*output*/, float thresh_fid, float rgb_wght);

/**
 * @brief get user information from DB
 * @return KAPP_OK
 */
int32_t kapp_db_get_user_info(
        uint16_t user_id,
        uint16_t *lp_valid_fm0,
        uint16_t *lp_valid_fm1,
        uint16_t *lp_type);

#endif /* __KAPP_DB_H_ */
