/**
 * Copyright (C) Kneron Inc. All right reserved
 *
 * @file kapp_fid.h
 * @brief Kneron Application FD/FR APIs
 */

#ifndef __KAPP_FID_H_
#define __KAPP_FID_H_

#include <stdint.h>
#include <stdbool.h>
#include "ipc.h"
#include "kapp.h"
#include "kmdw_model.h"

enum kapp_fid_status_code_e {
    KAPP_FDR_WRONG_USAGE = 0x100,    /// KAPP_FID_CODES
    KAPP_FDR_NO_FACE,                /// no face
    KAPP_FDR_BAD_POSE,               /// base face pose
    KAPP_FDR_FR_FAIL,                /// FR failed
    KAPP_LM_LOW_CONFIDENT,           /// LM low confident
    KAPP_LM_BLUR,                    /// LM blur
    KAPP_SMALL_FACE,                 /// face too small
    KAPP_RGB_Y_DARK
};

enum kapp_fid_op_e{
    KAPP_FID_OP_EXTRACT,             /// extract feature map only
    KAPP_FID_OP_INFERENCE,           /// do inference (do DB comparison)
    KAPP_FID_OP_REGISTER             /// do register (insert DB temp data)
};

typedef struct kapp_fid_inout_s {
    enum kapp_fid_op_e op;               ///operation mode
    uint16_t face_id;                       //1-5: faceId (1-5)
    uint16_t user_id;                       ///as input: user id got from Host
                                            ///as output: user id returned by KL520
    float thresh_fid[KAPP_IMG_SRC_MAX];  ///threshold value for face recognition
    bool whether_mask_mode;
} kapp_fid_inout_t;

/** @brief face detectin result
 * facedet_result_s is defined in ipc.h
 * */
typedef struct facedet_result_s kapp_face_det_res_t;

/** @brief landmark result
 * landmark_result_s is defined in ipc.h
 * */
typedef struct landmark_result_s kapp_face_lm_res_t;

/** @brief face recognition result
 * fr_result_s defined in ipc.h
 * */
typedef struct fr_result_s kapp_face_fmap_t;

typedef struct kapp_image_buffer_desc_s {
    uint32_t buf_count;         //image buffer count
    uint32_t alignment_in_bit;  //alignment size in bit
    uint32_t rgb_buf_base_addr; //rgb image buffer address
    uint32_t rgb_img_size;      //rgb image size
    uint32_t nir_buf_base_addr; //nir image buffer address
    uint32_t nir_img_size;      //nir image size
} kapp_image_buffer_desc_t;

/**
 * @brief extract face embedding
 * @param [in] img_src image source
 * @param [out] p_fd_out_p detected face bounding box
 * @param [out] p_lm_out_p detected face landmarks
 * @return OK or Err code
 */
int32_t kapp_fd_lm(int img_src, struct facedet_result_s *p_fd_out_p,
                   struct landmark_result_s *p_lm_out_p);

/**
 * @brief do face recognition for inference or register
 * @param [in,out] *p_inout_p
 *          { op;
 *            face_id; //as input
 *            user_id; //as input (or ouput) }
 *        op = KAPP_FID_OP_INFERENCE   : fd/fr + 1:n compare
 *             KAPP_FID_OP_REGISTER    : fd/fr + save to temp
 *             KAPP_FID_OP_EXTRACT     : fd/fr only
 *        mode = 1-5 : capture + fd/fr + register in DDR DB
 *        outputPtr: N/A
 * @param [out] p_fd_out_p fd result
 * @param [out] p_lm_out_p lm result
 * @param [out] p_fr_out_p fr result
 * @param [out] p_score_p score result
 * @return OK or Err code
 */
int32_t kapp_fdr(int img_src, kapp_fid_inout_t *p_inout_p,
                 struct facedet_result_s *p_fd_out_p,
                 struct landmark_result_s *p_lm_out_p,
                 struct fr_result_s *p_fr_out_p,
                 float *p_score_p);

/**
 * @brief do face recognition for inference or register with mask FR model
 * @param [in,out] *p_inout_p
 *          { op;
 *            face_id; //as input
 *            user_id; //as input (or ouput) }
 *        op = KAPP_FID_OP_INFERENCE   : fd/fr + 1:n compare
 *             KAPP_FID_OP_REGISTER    : fd/fr + save to temp
 *             KAPP_FID_OP_EXTRACT     : fd/fr only
 *        mode = 1-5 : capture + fd/fr + register in DDR DB
 *        outputPtr: N/A
 * @param [out] p_fd_out_p fd result
 * @param [out] p_lm_out_p lm result
 * @param [out] p_fr_no_mask_out_p fr result for face without mask  
 * @param [out] p_fr_mask_out_p fr result for face with mask 
 * @param [out] p_score_p score result
 * @return OK or Err code
 */
int32_t kapp_mask_fdr(int img_src, kapp_fid_inout_t *p_inout_p,
                 struct facedet_result_s *p_fd_out_p,
                 struct landmark_result_s *p_lm_out_p,
                 struct fr_result_s *p_fr_no_mask_out_p,
                 struct fr_result_s *p_fr_mask_out_p,
                 float *p_score_p);

/**
 * @brief get configured fid img cfg
 * @param [in] img_src img source index
 * @return image config
 */
struct kdp_img_cfg* kapp_get_img_cfg(int img_src);

/**
 * @brief handle fid operation
 * @param [in,out] fid input/output
 * @param [in] op operation code
 * @param [in] db_idx DB index
 * @param [in,out] p_fr_out_p fr result
 * @param [out] score DB comparision score/distance
 * @param [in] check_db_reg check on db register
 * @return OK or Err code
 */
int kapp_fid_handle_op(kapp_fid_inout_t *p_inout_p,
                       int db_idx, struct fr_result_s *p_fr_out_p,
                       float *score,
                       bool check_db_reg);
/**
 * @brief do face detection
 * @param [out] p_out_p: detected face bounding box
 * @param [in] img_cfg img config
 * @return OK or Err code
 */
int32_t kapp_fd(struct facedet_result_s *p_out_p/*out*/,
                struct kdp_img_cfg *img_cfg);


/**
 * @brief do face landmark
 * @param [out] p_out_p detected face landmarks
 * @param [in] p_fd_out_p detected face bounding box
 * @param [in] img_cfg img config
 * @return OK or Err code
 */
int32_t kapp_lm(struct landmark_result_s *p_out_p/*out*/,
                struct facedet_result_s *p_fd_out_p,
                struct kdp_img_cfg *img_cfg);



/**
 * @brief extract face embedding
 * @param [out] p_out_p extracted face embedding
 * @param [in] p_fd_out_p detected face bounding box
 * @param [in] p_lm_out_p detected face landmarks
 * @param [in] img_cfg img config
 * @param [in] whether_mask_mode whether using mask fr model
 * @return OK or Err code
 */
int32_t kapp_fr(struct fr_result_s *p_out_p/*out*/,
                struct facedet_result_s *p_fd_out_p,
                struct landmark_result_s *p_lm_out_p,
                struct kdp_img_cfg *img_cfg, bool whether_mask_mode);

/**
 * @brief init necessary data, face data base
 * @return OK
 */
int32_t kapp_fid_init(void);

/**
 * @brief (Beta) get 5 pose score for a landmark set
 * @param[out] pose : 5 pose score
 * @param[in]  landmark : 5 points landmark, 10 int value
 */
void kapp_get_pose_value(uint32_t* landmarks, float *poses);


/**
 * @brief (Beta) check if it is a bad pose for a landmark set
 * @param[in]  landmark : 5 points landmark, 10 int value
 * @return True for bad pose, False otherwise
 */
bool kapp_check_bad_pose(uint32_t *landmarks);

/**
 * @brief check head position
 */
int kapp_check_head_pose(struct landmark_result_s* landmarks,
                         uint8_t reg_idx, uint8_t force_flag,
                         uint8_t* face_reg_thresh);



#endif
