/*
 * Kneron Host Communication driver
 *
 * Copyright (C) 2018 Kneron, Inc. All rights reserved.
 *
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "os_tick.h"
#include "base.h"
#include "def_share.h"
#include "version.h"
#include "common.h"
#include "com.h"
#include "com_err.h"
#include "msg.h"
#include "kmdw_console.h"
#include "io.h"
#include "kcomm.h"
#include "power_manager.h"
#include "kmdw_ota.h"
#include "kdrv_clock.h"
#include "kdrv_power.h"
#include "kdp_system.h"

#define DEF_LOG_CATEG  "kcomm"
#define MSG_AUTH_FAIL 251  // also defined in ota.h

extern struct kdp_dme_conf_tbl_s    *p_dme_config_tbl;
extern uint32_t ack_idx;
extern uint32_t dme_raw_len;
extern volatile uint16_t active_model_index[];
extern struct fw_misc_data g_fw_misc;
static u16 SYS_sts = 0; // system status

#define U32_ENDIAN_SWAP(ver) \
    ((ver>>24)&0xff) | ((ver<<8)&0xff0000) | ((ver>>8)&0xff00) | ((ver<<24)&0xff000000);

/* ###########################
 * ##    local functions    ##
 * ########################### */

static bool valid_reset_mode(u32 mode)
{
    if ((mode == 255) || (mode == 256) ||
        (mode >> 16) == 0x1000 || (mode >> 16) == 0x2000) {
        return true;
    }
#ifdef USB_HOST
    return (mode < 2);
#else
    return (mode < 5 && mode != 2);
#endif
}

static inline int get_word_len(int byte)
{
    return (byte % 4) == 0 ? (byte/4) : (byte/4 + 1);
}

static void device_sleep(void)
{
    SYS_sts = 1;
}

static void device_wakeup(void)
{
    SYS_sts = 0;
}

/* ########################
 * ##    cmd handlers    ##
 * ######################## */

static void handle_mem_write(RspPram *rsp, u32 addr, u32 len, u32 *pdata)
{
    int wlen = get_word_len(len);

    dlog("addr [0x%x], len [%u]", addr, len);

    for (int i = 0; i < wlen; i++) {
        outw(addr + i * 4, (u32) *(pdata + i));
    }

    rsp->error = NO_ERROR;
    kcomm_send_rsp(CMD_MEM_WRITE, NULL, 0);
}

static void handle_mem_read(RspPram *rsp, u32 addr, u32 len, u32 *pdata)
{
    int wlen = get_word_len(len);

    dlog("addr [0x%x], len [%u]", addr, len);

    for (int i = 0; i < wlen; i++) {
        *(pdata + i) = inw(addr + 4 * i);
    }

    // msg byte read size is unchanged
    rsp->error = NO_ERROR;
    kcomm_send_rsp(CMD_MEM_READ, (u8*)pdata, len);
}

static void handle_ack_nack(int action, int src_id)
{
    struct kdp_dme_cfg *dme_conf = p_dme_config_tbl->dme_config;
    uint32_t dme_offset = 0;
    dme_conf += active_model_index[ack_idx];  // point to the correct model

    // Send results back to host
    if (action) {
        if ((src_id == 0) || (src_id > 0x100000)) {
            err_msg("DME ACK_NACK BAD length [%d]\n", src_id);
            kcomm_send_no_rsp();
        }
        else {
            if (src_id > 0x100000) {
                err_msg("[ACK-NACK] size error [%d]\n", src_id);
                return;
            }

            kcomm_write((u8 *) action, src_id, WMODE_DEF);
        }
        return; // Note: NACK is used to retrieve data via USB by DME
    }

    /* TODO: remove this DME cmd coupling */
    // We don't enforce Source ID check yet, since only DME is using this
    if (dme_conf->img_conf.image_format & IMAGE_FORMAT_RAW_OUTPUT){
        if (ack_idx) dme_offset = dme_raw_len;
        kcomm_write((u8*) (dme_raw_out + dme_offset), dme_out_len, WMODE_DEF);
    } else if (dme_conf->img_conf.image_format & IMAGE_FORMAT_MODEL_AGE_GENDER) {
        if (ack_idx) dme_offset = sizeof(fd_age_gender_res);
        kcomm_write((u8*) ((u32)dme_agg_out + dme_offset), sizeof(fd_age_gender_res), WMODE_DEF);
    } else {// all other DME case treat as detection
        if (ack_idx) dme_offset = sizeof(dme_res);
        kcomm_write((u8*) ((u32)dme_od_out + dme_offset), dme_out_len, WMODE_DEF);
    }
}

static void handle_mem_clear(RspPram *rsp, u32 addr, u32 len)
{
    dlog("addr [0x%x], len [%u]", addr, len);

    memset((void*)addr, 0, len);
    rsp->error = NO_ERROR;
    kcomm_send_rsp(CMD_MEM_CLR, NULL, 0);
}

static void handle_reset(RspPram *rsp, u32 mode, u32 chk_mode)
{
    dlog("mode [0x%x], chk_mode [0x%x]", mode, chk_mode);

    bool sys_reset = false;
    bool power_off = false;

    if (!valid_reset_mode(mode)) {
        rsp->error = BAD_MODE;
    } else {
        if ((mode + chk_mode) != 0xFFFFFFFF) {
            rsp->error = BAD_CONFIRMATION;
        } else {
            if (mode == 255) {
                sys_reset = true;
            } else if (mode == 256) {
                power_off = true;
            } else if ((mode >> 16) == 0x1000) {
                kmdw_console_set_log_level_scpu(mode & 0x000F);  // set scpu debug output level
                kmdw_console_set_log_level_ncpu((mode & 0x0F00)>> 8);  // set ncpu debug output level
            } else if ((mode >> 16) == 0x2000) {
                /* kai2do: this should be categorized as OTA cmd? */
                // switch active boot image, bit 0 -> scpu, bit 1 -> ncpu
                rsp->error = kmdw_ota_switch_active_partition(mode & 0x3);
            }
            rsp->error = NO_ERROR;  // not doing anything else for now
        }
    }

    rsp->param2 = 0;
    kcomm_send_rsp(CMD_RESET, NULL, 0);

    if (sys_reset) {
        NVIC_DisableIRQ(UART_FTUART010_0_IRQ);   //UART0
        NVIC_DisableIRQ(UART_FTUART010_1_IRQ);   //UART1
        NVIC_DisableIRQ(UART_FTUART010_1_1_IRQ); //UART2
        NVIC_DisableIRQ(UART_FTUART010_1_2_IRQ); //UART3
        NVIC_DisableIRQ(UART_FTUART010_1_3_IRQ); //UART4
        NVIC_DisableIRQ(NPU_NPU_IRQ);            //NPU
        kdrv_delay_us(10*1000);
        kdrv_power_sw_reset();
    } else if (power_off) {
        power_manager_shutdown();
    }
}

static void handle_sys_status(RspPram *rsp)
{
    uint32_t ncpu_version;

    g_fw_misc_present_t fw_ver_present;
    fw_ver_present.fw_raw[3] = g_fw_misc.architecture;
    fw_ver_present.fw_raw[2] = g_fw_misc.major;
    fw_ver_present.fw_raw[1] = g_fw_misc.minor;
    fw_ver_present.fw_raw[0] = g_fw_misc.update;

    rsp->param1 = fw_ver_present.fw_version;
    rsp->param2 = g_fw_misc.build;

    u16* status = (u16*)rsp->data;
    *status++ = SYS_sts;
    *status++ = APP_id;
    ncpu_version = kdp_sys_get_ncpu_version();
    *(u32*)status = U32_ENDIAN_SWAP(ncpu_version);
    status += 2;
    *(u32*) status = kdp_sys_get_ncpu_build();
    kcomm_send_rsp(CMD_SYSTEM_STATUS, rsp->data, 12);
}

static void handle_reserved(RspPram *rsp, u32 mode, u16 type, u16 size, u8 *signature)
{
    u32 error = MSG_AUTH_FAIL;
    if (type == 0)  // no signature
        error = 0;
    if (type == 1) {  // simple calculated signature
        u32 auth = *(u32*) signature;
        if (auth == 0xbc6d2d40) {
            auth = 0;
        } else {
            auth ^= U32_ENDIAN_SWAP(kdp_sys_get_ncpu_version());
            auth ^= kdp_sys_get_ncpu_build();
            auth ^= U32_ENDIAN_SWAP(*(u32*)&g_fw_misc);
            auth ^= g_fw_misc.build;
            auth ^= 0x4b353230;
        }
        if (auth == 0)
            error = 0;
    }
    rsp->param2 = 0;
    *(u32*)&rsp->data = 0;
    if (error) {
        if (mode == 1) {
            rsp->param1 = error;
            rsp->param2 = CMD_RESERVED;
            kcomm_send_rsp(CMD_ACK_NACK, NULL, 0);
            return;
        }
        rsp->param1 = error;
    } else {
        rsp->param1 = 0;
        if (mode == 1) {
            int ret = kmdw_ota_update_spl((u32)size);
            rsp->param1 = (u32) ret;
        }
        rsp->param2 = U32_ENDIAN_SWAP(kdp_sys_get_spl_version());
        *(u32*)&rsp->data = kdp_sys_get_spl_build();
    }
    kcomm_send_rsp(CMD_RESERVED, rsp->data, 4);
}

static void handle_get_kn_number(RspPram *rsp)
{
    rsp->param1 = kdp_sys_get_kn_number();
    rsp->param2 = 0;

    kcomm_send_rsp(CMD_GET_KN_NUM, NULL, 0);
}

static void handle_get_model_info(RspPram *rsp, u32 from_ddr)
{
    dbg_msg("handle_get_model_info %d\n", from_ddr);
    uint32_t *p_model_id;
    uint32_t data_size = 0;

    p_model_id = kmdw_model_get_all_model_info(from_ddr);

    if (p_model_id) {
        data_size = (1 + p_model_id[0]) * 4; // total_number + model_ids
    }

    rsp->param1 = data_size;
    rsp->param2 = 0;

    kcomm_send_rsp(CMD_GET_MODEL_INFO, NULL, 0);

    if (p_model_id) {
        kcomm_write((u8*)p_model_id, data_size, WMODE_ALT);
        free(p_model_id);
    }
}

static void handle_get_crc(RspPram *rsp, u32 from_ddr)
{
    dbg_msg("handle_get_crc %d\n", from_ddr);
    uint32_t *p_crc;
    uint32_t data_size = 0;

    p_crc = kmdw_model_get_crc(from_ddr);

    if (p_crc) {
        data_size = 4; // 4 bytes
    }

    rsp->param1 = data_size;
    rsp->param2 = 0;

    kcomm_send_rsp(CMD_GET_CRC, NULL, 0);

    if (p_crc) {
        kcomm_write((u8*)p_crc, data_size, WMODE_ALT);
        free(p_crc);
    }
}

static void handle_set_ckey(RspPram *rsp, u32 ckey)
{
    dbg_msg("handle_set_ckey 0x%08X\n", ckey);
    rsp->param1 = kdp_sys_program_key(ckey);
    rsp->param2 = 0;

    kcomm_send_rsp(CMD_SET_CKEY, NULL, 0);
}

static int handle_base_cmd(MsgHdr *hdr, CmdPram *args, RspPram *rsp, void *data)
{
    u16 cmd = hdr->cmd;
    u32 arg1 = args->param1;
    u32 arg2 = args->param2;

    switch (cmd) {
    case CMD_MEM_WRITE:
        handle_mem_write(rsp, arg1, arg2, (u32*)data);
        break;
    case CMD_MEM_READ:
        handle_mem_read(rsp, arg1, arg2, (u32*)data);
        break;
    case CMD_ACK_NACK:
        handle_ack_nack(arg1, arg2);
        break;
    case CMD_MEM_CLR:
        handle_mem_clear(rsp, arg1, arg2);
        break;
    case CMD_RESET:
        handle_reset(rsp, arg1, arg2);
        break;
    case CMD_SYSTEM_STATUS:
        handle_sys_status(rsp);
        break;
    case CMD_RESERVED:
        handle_reserved(rsp, arg1, (u16)(arg2 & 0xffff), (u16)(arg2 >> 16), (u8*)data);
        break;
    case CMD_GET_KN_NUM:
        handle_get_kn_number(rsp);
        break;
    case CMD_GET_MODEL_INFO:
        handle_get_model_info(rsp, arg1);
        break;
    case CMD_SET_CKEY:
        handle_set_ckey(rsp, arg1);
        break;
    case CMD_GET_CRC:
        handle_get_crc(rsp, arg1);
        break;
        // kai2do: how about these commands other than SFID?
    /* case CMD_QUERY_APPS:
     * case CMD_SELECT_APP:
     * case CMD_SET_MODE:
     * case CMD_SET_EVENTS:
     * case CMD_UPDATE:
     * case CMD_ABORT:
     * case CMD_SFID_RESERVE_1:
     * case CMD_SFID_START:
     *     handle_fid_start(rsp, hdr, *(float*)&arg1, *(float*)&arg2, (u8*)arg3);
     *     break; */
    default:
        return -1;
    }

    return 0;
}

static void init_cmd_handler()
{
    struct pm_s pms = {
        .nap = 0, .wakeup_nap = 0, .deep_nap = 0,
        .wakeup_deep_nap = 0, .sleep = (pm_call) device_sleep,
        .wakeup_sleep = (pm_call) device_wakeup,
        .deep_sleep = (pm_call) device_sleep,
        .wakeup_deep_sleep = (pm_call) device_wakeup,
    };

    power_manager_register(PM_DEVICE_HOST_COM, &pms);
}

/* ############################
 * ##    public functions    ##
 * ############################ */

void kcomm_enable_base_cmds(void)
{
    init_cmd_handler();
    kcomm_reg_cmd_handler(handle_base_cmd);
}
