/*
 * Kneron Host Communication driver
 *
 * Copyright (C) 2018 Kneron, Inc. All rights reserved.
 *
 */
#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include "cmsis_os2.h"
#include "base.h"
#include "kcomm.h"
#include "com.h"
#include "com_err.h"
#include "msg.h"
#include "kneron_mozart.h"
#include "kmdw_console.h"
#include "kmdw_model.h"
#include "kapp.h"
#include "kapp_fid.h"

#include "usb_com.h"
#include "kmdw_memory.h"
#include "kdrv_status.h"
#include "kdrv_uart.h"
#include "kdev_flash.h"
#include "kdrv_spif.h"
#include "kdrv_system.h"
#include "framework/event.h"
#include "os_tick.h"
#include "kdp_ddr_table.h"
#include "ipc.h"
#include "kdrv_clock.h" // for kdrv_delay_us
#include "power_manager.h"
#include "kdrv_power.h"
#include "kmdw_ota.h"

#define DEF_LOG_CATEG  "Host_com"

//rrr
#include "kmdw_camera.h"
#include "media/kdp_inference.h"
#include "kdrv_fb_mgr.h"
/* #include "sample_app/user_io.h" */
#include "kl520_include.h"

static struct facedet_result_s *p_rgb_fd_out_s = 0;
static struct landmark_result_s *p_rgb_lm_out_s = 0;
static struct fr_result_s *p_rgb_fr_out_s = 0;
static struct facedet_result_s *p_nir_fd_out_s = 0;
static struct landmark_result_s *p_nir_lm_out_s = 0;
static struct fr_result_s *p_nir_fr_out_s = 0;
static kapp_fid_inout_t *p_fid_out_s = 0;

//camera default setting
static bool cam_connect = FALSE;
/* 0x3E00[3:0]/0x3E01/0x3E02 */
static u32 m_iCurExpoTime   = 0xA840;
/* 0x3E08/0x3E09 AGC = 4.078 */
static int coarse_gain      = 0x27;
static int fine_gain        = 0x24;
static int led_state        = 1;
static int32_t rgb_ret      = 0;
static int32_t nir_ret      = 0;


extern int kdp_memxfer_ddr_to_flash(u32 dst, u32 src, size_t bytes);
//rrr
extern int sample_prepare_inference(void);

/* ###########################
 * ##    local functions    ##
 * ########################### */
void camera_capture(void)
{
    int cam_idx;
    unsigned int frame_addr_rgb, frame_addr_nir;
    int g_buf_idx[2] = {-1, -1};

    cam_idx = CAMERA_RGB_IDX;
    frame_addr_rgb = kdrv_fb_mgr_next_inf(cam_idx, &g_buf_idx[cam_idx]);
    kapp_set_img_mem_addr(cam_idx, frame_addr_rgb);

    cam_idx = CAMERA_NIR_IDX;
    frame_addr_nir = kdrv_fb_mgr_next_inf(cam_idx, &g_buf_idx[cam_idx]);
    kapp_set_img_mem_addr(cam_idx, frame_addr_nir);
}

void camera_register(void)
{
    rgb_ret = kapp_fd_lm(KAPP_IMG_SRC_RGB, p_rgb_fd_out_s, p_rgb_lm_out_s);
    nir_ret = kapp_fd_lm(KAPP_IMG_SRC_NIR, p_nir_fd_out_s, p_nir_lm_out_s);
}


static bool valid_reset_mode(u32 mode)
{
    if ((mode == 255) || (mode == 256) ||
        (mode >> 16) == 0x1000 || (mode >> 16) == 0x2000) {
        return true;
    }
#ifdef USB_HOST
    return (mode < 2);
#else
    return (mode < 5 && mode != 2);
#endif
}

static void get_io_memory(void)
{
    if(p_fid_out_s == 0)
        p_fid_out_s = (kapp_fid_inout_t*)kmdw_ddr_reserve(sizeof(kapp_fid_inout_t));
    if(p_rgb_fd_out_s == 0)
        p_rgb_fd_out_s = (struct facedet_result_s*)kmdw_ddr_reserve(sizeof(struct facedet_result_s));
    if(p_rgb_lm_out_s == 0)
        p_rgb_lm_out_s = (struct landmark_result_s*)kmdw_ddr_reserve(sizeof(struct landmark_result_s));
    if(p_rgb_fr_out_s == 0)
        p_rgb_fr_out_s = (struct fr_result_s*)kmdw_ddr_reserve(sizeof(struct fr_result_s));
    if(p_nir_fd_out_s == 0)
        p_nir_fd_out_s = (struct facedet_result_s*)kmdw_ddr_reserve(sizeof(struct facedet_result_s));
    if(p_nir_lm_out_s == 0)
        p_nir_lm_out_s = (struct landmark_result_s*)kmdw_ddr_reserve(sizeof(struct landmark_result_s));
    if(p_nir_fr_out_s == 0)
        p_nir_fr_out_s = (struct fr_result_s*)kmdw_ddr_reserve(sizeof(struct fr_result_s));
}

/* ########################
 * ##    cmd handlers    ##
 * ######################## */
static void handle_download_image_rgb(void)
{
    dlog("no input");

    if (!cam_connect) {
        sample_prepare_inference();
        kmdw_camera_set_exp_time(CAMERA_NIR_IDX, (m_iCurExpoTime&0x0000FF00)>>8, m_iCurExpoTime&0xFF);
        kmdw_camera_set_gain(CAMERA_NIR_IDX, coarse_gain, fine_gain);
    } else {
        dbg_msg("[Host_com] camera is already connected\n");
    }

    get_io_memory();

    //kl520_api_face_auto_register_both(0, 0);

    camera_capture();
    camera_register();
    camera_capture();
    camera_register();

    struct kdp_img_cfg* img_cfg = kapp_get_img_cfg(KAPP_IMG_SRC_RGB);
    dbg_msg("rgb image mem addr = %x\n", img_cfg->image_mem_addr);
    u8 * rgb_image_addr = (u8 *)(img_cfg->image_mem_addr);
    kcomm_write((u8 *) rgb_image_addr, 614400, WMODE_DEF);

    DSG("[Host_com] Received CMD_DOWNLOAD_IMAGE_RGB finish\r\n");
}
static void handle_download_image_nir(void)
{
    dlog("no input");

    struct kdp_img_cfg* img_cfg = kapp_get_img_cfg(KAPP_IMG_SRC_NIR);
    dbg_msg("nir image mem addr = %x\n", img_cfg->image_mem_addr);
    u8 * nir_image_addr = (u8 *)(img_cfg->image_mem_addr);
    kcomm_write((u8 *) nir_image_addr, 307200, WMODE_DEF);

    DSG("[Host_com] Received CMD_DOWNLOAD_IMAGE_NIR finish\r\n");
}
static void handle_set_nir_agc(RspPram *rsp, u32 p_agc, u32 p_agf)
{
            dlog("coarse gain = 0x%x, find gain = 0x%x", p_agc, p_agf);
            coarse_gain = p_agc;
            fine_gain = p_agf;
            kmdw_camera_set_gain(CAMERA_NIR_IDX, coarse_gain, fine_gain);

            rsp->param1 = NO_ERROR;
            rsp->param2 = 0;
            kcomm_send_rsp(CMD_SET_NIR_AGC, NULL, 0);
}
static void handle_set_nir_aec(RspPram *rsp, u32 p_aec)
{
            dlog("AEC = %d", p_aec);
            m_iCurExpoTime = p_aec;
            kmdw_camera_set_exp_time(CAMERA_NIR_IDX,
                    (m_iCurExpoTime&0x0000FF00)>>8, m_iCurExpoTime&0xFF);

            rsp->param1 = NO_ERROR;
            rsp->param2 = 0;
            kcomm_send_rsp(CMD_SET_NIR_AEC, NULL, 0);
}
static void handle_set_nir_led(RspPram *rsp, u32 p_led_state)
{
            dlog("led_state = %d", led_state);

            led_state = (p_led_state == 0) ? 0 : 1;
            kmdw_camera_led_switch(CAMERA_NIR_IDX, led_state);
            //nir_led_init(msgcmd_arg->len);

            rsp->param1 = NO_ERROR;
            rsp->param2 = 0;
            kcomm_send_rsp(CMD_SET_NIR_LED, NULL, 0);
}
static void handle_cam_connect(RspPram *rsp)
{
    dlog("no input");
    sample_prepare_inference();
    kmdw_camera_set_exp_time(CAMERA_NIR_IDX, (m_iCurExpoTime&0x0000FF00)>>8, m_iCurExpoTime&0xFF);
    kmdw_camera_set_gain(CAMERA_NIR_IDX, coarse_gain, fine_gain);
    cam_connect = TRUE;

    rsp->param1 = NO_ERROR;
    rsp->param2 = 0;
    kcomm_send_rsp(CMD_CAM_CONNECT, NULL, 0);
}
static void handle_nir_fd_res(RspPram *rsp)
{
    u16* status;
    rsp->param1= rgb_ret;
    rsp->param2 = nir_ret;
    status = (u16*)rsp->data;
    *status++ = p_nir_fd_out_s->xywh[0];
    *status++ = p_nir_fd_out_s->xywh[1];
    *status++ = p_nir_fd_out_s->xywh[2];
    *status++ = p_nir_fd_out_s->xywh[3];
    kcomm_send_rsp(CMD_NIR_FD_RES, rsp->data, 8);
}
static void handle_nir_lm_res(RspPram *rsp)
{
    dlog("no input");
    u16* status;
    rsp->param1 = rgb_ret;
    rsp->param2 = nir_ret;
    status = (u16*)rsp->data;
    *status++ = p_nir_lm_out_s->marks[0].x;
    *status++ = p_nir_lm_out_s->marks[0].y;
    *status++ = p_nir_lm_out_s->marks[1].x;
    *status++ = p_nir_lm_out_s->marks[1].y;
    *status++ = p_nir_lm_out_s->marks[2].x;
    *status++ = p_nir_lm_out_s->marks[2].y;
    *status++ = p_nir_lm_out_s->marks[3].x;
    *status++ = p_nir_lm_out_s->marks[3].y;
    *status++ = p_nir_lm_out_s->marks[4].x;
    *status++ = p_nir_lm_out_s->marks[4].y;
    kcomm_send_rsp(CMD_NIR_LM_RES, rsp->data, 20);
}

static void handle_nir_get_conf(RspPram *rsp)
{
    dlog("no input");
    u16* status;
    rsp->param1 = m_iCurExpoTime;
    rsp->param2 = (coarse_gain << 8) + fine_gain;
    status = (u16*)rsp->data;
    *status++ = led_state;
    err_msg("led_state %d", led_state);
    kcomm_send_rsp(CMD_NIR_GET_CONF, rsp->data, 2);
}

static void handle_reset(RspPram *rsp, u32 mode, u32 chk_mode)
{
    dlog("mode [0x%x], chk_mode [0x%x]", mode, chk_mode);

    u32 flag = 0;  // reset flag: 1->sleep, 2->reset, 3->RTC

    if (!valid_reset_mode(mode)) {
        rsp->error = BAD_MODE;
    } else {
        if ((mode + chk_mode) != 0xFFFFFFFF) {
            rsp->error = BAD_CONFIRMATION;
        } else {
            if (mode == 3) {
                flag = 1;    // go to sleep
            } else if (mode == 255) {
                flag = 2;  // reset
            } else if (mode == 256) {
                flag = 3;  // power down
            } else if ((mode >> 16) == 0x1000) {
                kmdw_console_set_log_level_scpu(mode & 0x000F);  // set scpu debug output level
                kmdw_console_set_log_level_ncpu((mode & 0x0F00)>> 8);  // set ncpu debug output level
            } else if ((mode >> 16) == 0x2000) {
                // switch active boot image, bit 0 -> scpu, bit 1 -> ncpu
                rsp->error = kmdw_ota_switch_active_partition(mode & 0x3);
            }
            rsp->error = NO_ERROR;  // not doing anything else for now
        }
    }

    rsp->param2 = 0;
    kcomm_send_rsp(CMD_RESET, NULL, 0);

    if (flag) {
        // ncpu fw needs update to support waking up from sleep
        // if (flag == 1) power_manager_sleep();
        if (flag == 2) {
            kdrv_power_sw_reset();
        } else if (flag == 3) {
            power_manager_shutdown();
        }
    }
}

static void handle_unknown_cmd(RspPram *rsp, u16 cmd)
{
    err_msg("[Host_com] got unknown cmd [0x%x]", cmd);

    rsp->error = RSP_UKNOWN_CMD;
    rsp->param2 = 0;
    kcomm_send_rsp(cmd, NULL, 0);
}

static int handle_camera_cmd(MsgHdr *hdr, CmdPram *args, RspPram *rsp, void *data)
{
    u16 cmd = hdr->cmd;
    u32 arg1 = args->param1;
    u32 arg2 = args->param2;

    switch (cmd) {
    case CMD_DOWNLOAD_IMAGE_RGB:
        handle_download_image_rgb();
        break;
    case CMD_DOWNLOAD_IMAGE_NIR:
        handle_download_image_nir();
        break;
    case CMD_SET_NIR_AGC:
        handle_set_nir_agc(rsp, arg1, arg2);
        break;
    case CMD_SET_NIR_AEC:
        handle_set_nir_aec(rsp, arg1);
        break;
    case CMD_SET_NIR_LED:
        handle_set_nir_led(rsp, arg1);
        break;
    case CMD_CAM_CONNECT:
        handle_cam_connect(rsp);
        break;
    case CMD_NIR_FD_RES:
        handle_nir_fd_res(rsp);
        break;
    case CMD_NIR_LM_RES:
        handle_nir_lm_res(rsp);
        break;
    case CMD_NIR_GET_CONF:
        handle_nir_get_conf(rsp);
        break;
    case CMD_RESET:
        handle_reset(rsp, arg1, arg2);
        break;
    default:
        handle_unknown_cmd(rsp, cmd);
        return -1;
    }

    return 0;
}

/* ############################
 * ##    public functions    ##
 * ############################ */

void kcomm_enable_camera_cmds(void)
{
    kcomm_reg_cmd_handler(handle_camera_cmd);
}
