/*
 * Kneron Host Communication driver
 *
 * Copyright (C) 2018 Kneron, Inc. All rights reserved.
 *
 */
#include <stdio.h>
#include <string.h>
#include "base.h"
#include "def_share.h"
#include "com.h"
#include "com_err.h"
#include "msg.h"
#include "kmdw_console.h"
#include "kcomm.h"
#include "kapp.h"
#include "kapp_fid.h"
#include "kapp_db.h"
#include "kmdw_memory.h"
#include "common.h"

#define DEF_LOG_CATEG  "kcomm"

enum FID_DB_OP_MODE {
    DB_IMPORT   = 1,
    DB_EXPORT   = 2,
    DB_QUERY    = 3,
    DB_ADD_FM   = 4,
    DB_DEL_FM   = 5,
    DB_QUERY_FM = 6,
    DB_OP_MAX   = 7
};

enum FID_DB_CONFIG_MODE {
    DB_SET_CONFIG = 1,
    DB_GET_CONFIG,
    DB_GET_DB_INDEX,
    DB_SWITCH_DB_INDEX,
    DB_SET_VERSION,
    DB_GET_VERSION,
    DB_GET_META_DATA_VERSION
};

enum OUTPUT_ITEM {
    OUTPUT_ITEM_RGB_FD = 0,
    OUTPUT_ITEM_RGB_LM,
    OUTPUT_ITEM_RGB_FR,
    OUTPUT_ITEM_RGB_LV,
    OUTPUT_ITEM_RGB_SCORE,
    OUTPUT_ITEM_MAX=15
};
#define M_OUTPUT_ITEM(x) (1u<<x)
#define OUTPUT_ITEM_2D   ((1u<<OUTPUT_ITEM_RGB_FD) | \
                          (1u<<OUTPUT_ITEM_RGB_LM) | \
                          (1u<<OUTPUT_ITEM_RGB_FR) | \
                          (1u<<OUTPUT_ITEM_RGB_LV) | \
                          (1u<<OUTPUT_ITEM_RGB_SCORE))

static u16 SFID_user;   // SFID user id
static u16 SFID_index;  // SFID image index
static kapp_image_buffer_desc_t SFID_buf_desc;

static struct facedet_result_s  *rgb_fd_out = 0;
static struct landmark_result_s *rgb_lm_out = 0;
static struct fr_result_s       *rgb_fr_out = 0;
static float rgb_score_out_s = -1.0;  //-1 means not set

static struct facedet_result_s  *nir_fd_out = 0;
static struct landmark_result_s *nir_lm_out = 0;
static struct fr_result_s       *nir_fr_out = 0;
static kapp_fid_inout_t         *p_fid_out_s = 0;

static uint32_t fm = 0;

extern int kdp_memxfer_ddr_to_flash(u32 dst, u32 src, size_t bytes);

/* ###########################
 * ##    local functions    ##
 * ########################### */

static int config_img(int src_img_idx, struct kdp_img_cfg *img_cfg, kapp_image_buffer_desc_t *buf_desc)
{
    int fmt = img_cfg->image_format;
    int pixel_size = kapp_get_pixel_size(fmt);
    int ch = kapp_get_chnl_size(fmt);
    int img_buf_idx;

    if (pixel_size == 0) {
        return -1;
    }

    img_cfg->image_mem_len = img_cfg->image_col * img_cfg->image_row * pixel_size;

    buf_desc->buf_count = KAPP_IMG_SRC_MAX;
    buf_desc->alignment_in_bit  = 32;

    if (src_img_idx == KAPP_IMG_SRC_RGB) {
        buf_desc->rgb_buf_base_addr = img_cfg->image_mem_addr;
        buf_desc->rgb_img_size = img_cfg->image_mem_len;
        img_buf_idx = 0;
    } else if (src_img_idx == KAPP_IMG_SRC_NIR) {
        buf_desc->nir_buf_base_addr = img_cfg->image_mem_addr;
        buf_desc->nir_img_size = img_cfg->image_mem_len;
        img_buf_idx = 1;
    }

    kapp_config_image(src_img_idx, img_cfg->image_col, img_cfg->image_row, ch,
                         fmt, img_cfg->image_mem_addr, img_buf_idx);

    return 0;
}

static int fid_ret_to_err(int ret)
{
    switch (ret) {
    case KAPP_DB_NO_MATCH:
        return MSG_APP_DB_NO_MATCH;
    case KAPP_DB_ALREADY_SAVED:
        return MSG_APP_UID_EXIST;
    case KAPP_DB_REG_FIRST:
        return MSG_APP_IDX_FIRST;
    case KAPP_DB_USER_NOT_REG:
        return MSG_APP_IDX_MISSING;
    case KAPP_DB_NO_SPACE:
        return MSG_APP_DB_FAIL;
    case KAPP_FDR_WRONG_USAGE:
        return MSG_APP_BAD_INDEX;
    case KAPP_DB_DEL_NOT_VALID:
        return MSG_APP_DB_NO_USER;
    default:
        return MSG_APP_BAD_IMAGE;
    }
}

static bool valid_app_id(int id)
{
    return ((id > APP_UNKNOWN) && (id < APP_ID_MAX));
}

static u16* set_img_cfg(struct kdp_img_cfg *cfg, u16 *data)
{
    u16 col = *data++;
    u16 row = *data++;
    u32 *pword = (u32*)data;
    u32 fmt = *pword++;

    cfg->image_col = col;
    cfg->image_row = row;
    cfg->image_format = fmt;

    dlog("img col [%u], row [%u], fmt [0x%x]", col, row, fmt);

    return (u16*) pword; // final pointer pos
}

static void output_results(u16 *output_buf, void *out_result, u32 count)
{
    u32 *result = out_result;
    while (count) {
        *output_buf++ = (u16) *result++;
        count--;
    }
}

/* This function checks the output data and if valid, will append them to the message
** input:  output flags to check, pointer to start of output data in message buffer
** output: number of bytes added to the message buffer
*/
static u32 fdlmfrlv_outputs(u32 *check_flag, u16 *pfdfr_out, u32 lv_flag)
{
    u32 out_size = 0;
    u32 out_flag = *check_flag;

    if (*(u32*)rgb_fd_out) { // FD valid?
        if (out_flag & M_OUTPUT_ITEM(OUTPUT_ITEM_RGB_FD)) {  // do we export FD?
            output_results(pfdfr_out, &rgb_fd_out->xywh[0], (FD_RES_LENGTH / 2)); // u16 output, 4 members
            out_size += FD_RES_LENGTH;
            pfdfr_out += (FD_RES_LENGTH / 2);
        }
    } else {
        //unset FD, LM, FR, LV, score
        out_flag &= ~(M_OUTPUT_ITEM(OUTPUT_ITEM_RGB_FD) |
                      M_OUTPUT_ITEM(OUTPUT_ITEM_RGB_LM) |
                      M_OUTPUT_ITEM(OUTPUT_ITEM_RGB_FR) |
                      M_OUTPUT_ITEM(OUTPUT_ITEM_RGB_LV) |
                      M_OUTPUT_ITEM(OUTPUT_ITEM_RGB_SCORE)
                      );
    }

    if (out_flag & M_OUTPUT_ITEM(OUTPUT_ITEM_RGB_LM)) {  // export LM?
        if (*(u32*)rgb_fd_out) { // if FD is good, assume LM is also good
            output_results(pfdfr_out, &rgb_lm_out->marks[0].x, (LM_RES_LENGTH / 2)); // 10 members
            out_size += LM_RES_LENGTH;
            pfdfr_out += (LM_RES_LENGTH / 2);
        }
    }

    if (out_flag & M_OUTPUT_ITEM(OUTPUT_ITEM_RGB_FR)) {  // export FR?
        const int fr_res_node_size = FR_RES_LENGTH/FR_RES_COUNT;
        u8 *feature_map_set[FR_RES_COUNT];
        int32_t origin_db_idx = kapp_db_get_db_index();

#ifdef EMBED_CMP_NPU
        if (!(rgb_fr_out->feature_map_fixed[0] == 0 &&
              rgb_fr_out->feature_map_fixed[1] == 0 &&
              rgb_fr_out->feature_map_fixed[2] == 0)) { // FR valid? (invalid when 3 contiguous data are 0)

            feature_map_set[0] = (u8*)rgb_fr_out->feature_map_fixed;
#ifdef MASK_FDR
            feature_map_set[1] = (u8*)nir_fr_out->feature_map_fixed;

            for (int db_idx=0 ;db_idx < FR_RES_COUNT; db_idx++)
            {
                uint8_t* wt_hdr = kapp_db_get_wt_hdr(); // wt hdr has 16 bytes
                memcpy((u8*)pfdfr_out, wt_hdr, 16);
                pfdfr_out += (16 / 2);

                memcpy((u8*)pfdfr_out, feature_map_set[db_idx], FR_FEAT_SIZE);
                pfdfr_out += (FR_FEAT_SIZE / 2);

                kapp_db_switch(db_idx);
                uint8_t embed_data = kapp_db_get_current_embed_data(); // user embedding data
                *pfdfr_out = embed_data;
                pfdfr_out += (2 / 2);

                out_size += fr_res_node_size;
            }
#else
            uint8_t* wt_hdr = kapp_db_get_wt_hdr(); // wt hdr has 16 bytes
            memcpy((u8*)pfdfr_out, wt_hdr, 16);
            pfdfr_out += (16 / 2);

            memcpy((u8*)pfdfr_out, feature_map_set[0], FR_FEAT_SIZE);
            pfdfr_out += (FR_FEAT_SIZE / 2);

            uint8_t embed_data = kapp_db_get_current_embed_data(); // user embedding data
            *pfdfr_out = embed_data;
            pfdfr_out += (2 / 2);

            out_size += fr_res_node_size;
#endif
        }
#else
        if (!(rgb_fr_out->feature_map[0] == 0 &&
              rgb_fr_out->feature_map[1] == 0 &&
              rgb_fr_out->feature_map[2] == 0)) { // FR valid? (invalid when 3 contiguous data are 0)
            feature_map_set[0] = (u8*)rgb_fr_out->feature_map;
#ifdef MASK_FDR
            feature_map_set[1] = (u8*)nir_fr_out->feature_map;
#endif
            for (int db_idx = 0 ;db_idx < FR_RES_COUNT; db_idx++) {
                memcpy((u8*)pfdfr_out, feature_map_set[db_idx], fr_res_node_size);
                //dbg_msg("FR result: %1.4f, %1.4f, %1.4f, ...,%1.4f, %1.4f\n", 
                //        rgb_fr_out->feature_map[0], rgb_fr_out->feature_map[1], rgb_fr_out->feature_map[2], 
                //        rgb_fr_out->feature_map[FR_FEAT_SIZE-2], rgb_fr_out->feature_map[FR_FEAT_SIZE-1]);
                out_size += fr_res_node_size;
                pfdfr_out += (fr_res_node_size/ 2);
            }
        }
#endif  /*EMBED_CMP_NPU*/
        else {
            //unset FR, score
            out_flag &= ~(M_OUTPUT_ITEM(OUTPUT_ITEM_RGB_FR) |
                          M_OUTPUT_ITEM(OUTPUT_ITEM_RGB_SCORE)
                         );
            //out_flag &= 0x0fffb;
    }

        kapp_db_switch(origin_db_idx);
    }
    if (out_flag & M_OUTPUT_ITEM(OUTPUT_ITEM_RGB_LV)) {  // export LV?
        if (lv_flag & 0x80) { // LV valid?
            *(u32*)pfdfr_out = lv_flag & 1;  // store the results
            out_size += LV_RES_LENGTH;
            pfdfr_out += (LV_RES_LENGTH / 2);
        }
        else
            out_flag &= ~(M_OUTPUT_ITEM(OUTPUT_ITEM_RGB_LV));
    }

    if (out_flag & M_OUTPUT_ITEM(OUTPUT_ITEM_RGB_SCORE)) {  // export SCORE?
        if ( -1 != rgb_score_out_s ) { // score valid?
            //dbg_msg("score = %1.4f\n", rgb_score_out_s);
            memcpy((u8*)pfdfr_out, (u8*)(&rgb_score_out_s), SCORE_RES_LENGTH);
            out_size += SCORE_RES_LENGTH;
            pfdfr_out += (SCORE_RES_LENGTH / 2);
        }
        else
            out_flag &= ~(M_OUTPUT_ITEM(OUTPUT_ITEM_RGB_SCORE));
    }

    //----------------------------------------------------------------------------

    *check_flag = out_flag;
    return out_size;
}

static void _get_io_memory(void)
{
    if(p_fid_out_s == 0)
        p_fid_out_s = (kapp_fid_inout_t*)kmdw_ddr_reserve(sizeof(kapp_fid_inout_t));
    if(rgb_fd_out == 0)
        rgb_fd_out = (struct facedet_result_s*)kmdw_ddr_reserve(sizeof(struct facedet_result_s));
    if(rgb_lm_out == 0)
        rgb_lm_out = (struct landmark_result_s*)kmdw_ddr_reserve(sizeof(struct landmark_result_s));
    if(rgb_fr_out == 0)
        rgb_fr_out = (struct fr_result_s*)kmdw_ddr_reserve(sizeof(struct fr_result_s));

    if(nir_fd_out == 0)
        nir_fd_out = (struct facedet_result_s*)kmdw_ddr_reserve(sizeof(struct facedet_result_s));
    if(nir_lm_out == 0)
        nir_lm_out = (struct landmark_result_s*)kmdw_ddr_reserve(sizeof(struct landmark_result_s));
    if(nir_fr_out == 0)
        nir_fr_out = (struct fr_result_s*)kmdw_ddr_reserve(sizeof(struct fr_result_s));
}

/* ########################
 * ##    cmd handlers    ##
 * ######################## */

static void handle_fid_start(RspPram *rsp, MsgHdr *hdr,
                             float thresh_val_1, float thresh_val_2, u8 *data)
{
    init_img_cfg(); // allocate memory only when needed, to workaround for KDP2 FIFOQ

    int ret = 0;
    int app_type = APP_SFID;
   

    dbg_msg("app_type [%d], thresh_val_1 [%.3f], thresh_val_2 [%.3f], data [0x%x]",
          app_type, thresh_val_1, thresh_val_2, data);

    kapp_set_dme_mode(false);

    if (SFID_user) {
        ret = kapp_db_abort_reg();
        if (ret != KAPP_OK) {
            err_msg("[kcomm] Start Abort Old Session FAILED.\r\n");
        }
    }

    info_msg("[kcomm] Start FD/FR/DB.\r\n");
    SFID_user = 0; // switch to NORMAL FD/FR/DB MODE
    SFID_index = 0;

    // init global memory
    _get_io_memory();

#ifdef MASK_FDR
    thresh_val_2 = thresh_val_1; // currently share same threshold for fid, fid_mask
    p_fid_out_s->thresh_fid[NO_MASK_FDR_DB_INDEX] = (thresh_val_1 > EPSILON_FLT)? thresh_val_1 : FID_THRESHOLD;
    p_fid_out_s->thresh_fid[MASK_FDR_DB_INDEX] = (thresh_val_2 > EPSILON_FLT)? thresh_val_2 : FID_MASK_THRESHOLD;
#else
    p_fid_out_s->thresh_fid[KAPP_IMG_SRC_RGB] = (thresh_val_1 > EPSILON_FLT)? thresh_val_1 : FID_THRESHOLD;
#endif

    // config image settings
    if (hdr->msg_len > 8) {
        u16 *prd = set_img_cfg(&rgb_img_config, (u16*)data);
    }

    ret = config_img(KAPP_IMG_SRC_RGB, &rgb_img_config, &SFID_buf_desc);

    if (ret != 0) {
        rsp->error = MSG_CONFIG_ERROR;
        rsp->param2 = 0; // image size set to 0
    } else {
        dbg_msg("[kcomm] FID START, buf count [%u], alignment [%u], "
                "rgb addr [0x%x], size [%u], nir addr [0x%x], size [%u]\r\n",
                SFID_buf_desc.buf_count, SFID_buf_desc.alignment_in_bit,
                SFID_buf_desc.rgb_buf_base_addr, SFID_buf_desc.rgb_img_size,
                SFID_buf_desc.nir_buf_base_addr, SFID_buf_desc.nir_img_size);

        APP_id = app_type;

        // start continuous inference & DB match
        rsp->error = NO_ERROR;
        rsp->param2 = SFID_buf_desc.rgb_img_size;
    }

    kcomm_send_rsp(CMD_SFID_START, NULL, 0);
}

static void handle_fid_new_user(RspPram *rsp, u32 uid, u32 idx)
{
    dlog("uid [%u], idx [%u]", uid, idx);

    if ((uid == 0) || !valid_app_id(APP_id)) {
        rsp->error = (uid == 0) ? MSG_SFID_USR_ZERO : MSG_CONFIG_ERROR;
        rsp->param2 = 0;
        kcomm_send_rsp(CMD_SFID_NEW_USER, NULL, 0);
        return;
    }

    if (SFID_user == 0) {  // switch to ADD USER MODE
        SFID_user = uid;
        SFID_index = idx;
    } else {
        if (SFID_user == uid) {
            SFID_index = idx;
        } else {
            if (idx == 1) {
                // abort any previous NEW USER session first
                int ret = kapp_db_abort_reg();
                if (ret == KAPP_OK) {
                    SFID_user = uid;
                    SFID_index = 1;
                } else { // shouldn't happen
                    err_msg("[kcomm] Add user Abort Old Session %d Failed -> %d\r\n", uid, ret);
                    // don't know how to handle this, leave to image cmd
                    SFID_user = uid; // force the case anyway
                }
            } else { // we don't send error code out anymore
                SFID_user = uid;  // let image cmd deal with the problem
                SFID_index = idx;
            }
        }
    }

    info_msg("[kcomm] Add New User %d Indx -> %d\r\n", SFID_user, SFID_index);

    rsp->error = NO_ERROR;
    rsp->param2 = 0;
    kcomm_send_rsp(CMD_SFID_NEW_USER, NULL, -1);  // use zero payload response
}

static void handle_send_img(RspPram *rsp, u32 img_size, u32 out_flag)
{
    dlog("out_flag [0x%x], img_size [%u]", out_flag, img_size);

    kapp_db_config_t db_config;
    int ret_get_db_config = kapp_db_get_db_config(&db_config);

    u32 out_size, lv_flag = 0;  // lv_flag (byte) bit 7-> result_valid, bit 0-> result
    u16 *pfdfr_out;

    out_flag &= OUTPUT_ITEM_2D;  // ---- ---Z rmdS VRMD 
                                 //-> nirfr.nirlm.nirfd.rgbScore rgbLV.FR.LM.FD
    pfdfr_out = (u16*)rsp->data;  // point to output data field
    *pfdfr_out++ = 0;  // initialize the response msg data fields
    *pfdfr_out++ = 0;  // now pointing to the optional results field

    if (APP_id != APP_SFID) {
        rsp->error = MSG_CONFIG_ERROR;
        rsp->param2 = CMD_SFID_SEND_IMAGE;
        kcomm_send_rsp(CMD_ACK_NACK, NULL, 0);
        return;
    }

    struct facedet_result_s *fd_out = rgb_fd_out;
    struct landmark_result_s *lm_out = rgb_lm_out;
    struct fr_result_s *fr_out = rgb_fr_out;
#ifdef MASK_FDR
    struct fr_result_s *fr_out_2 = nir_fr_out; //reserved if model has 2 output node
#endif
    kapp_fid_inout_t *fdr_out = p_fid_out_s;
    rgb_score_out_s = -1;

    if (SFID_index == 0) {
        fdr_out->op = KAPP_FID_OP_INFERENCE;
        fdr_out->face_id = 0; //inference mode
    } else {
        fdr_out->op = KAPP_FID_OP_REGISTER;
        fdr_out->face_id = SFID_index; //reg mode, fm index
    }

    fdr_out->user_id = SFID_user;
   
    // get image
    int ret = kcomm_read(SFID_buf_desc.rgb_buf_base_addr, img_size);  // specify how many bytes
    if (ret == img_size) { // got image
        // input range check first
        if ((SFID_user != 0) && (SFID_index != 0)) {
            if (0 != ret_get_db_config) {
                rsp->error = MSG_CONFIG_ERROR;
                rsp->param2 = 0;
                kcomm_send_rsp(CMD_SFID_SEND_IMAGE, rsp->data, 4);
                return;
            }
            
            if ((SFID_user > db_config.max_uid) || (SFID_index > db_config.max_fid)) {
                // bad user id or image index
                rsp->error = MSG_APP_BAD_INDEX;
                rsp->param2 = 0;
                kcomm_send_rsp(CMD_SFID_SEND_IMAGE, rsp->data, 4);
                return;
            }
        }

        // now send command to AppMgr
        *(u32*)fd_out = 0;  // initialize the results, FD and FR only
        *(u32*)fr_out = 0;  // fdr() does not perform LV right now
#ifdef MASK_FDR
        *(u32*)fr_out_2 = 0;  // fdr() does not perform LV right now
        ret = kapp_mask_fdr(KAPP_IMG_SRC_RGB/*img config*/, fdr_out, fd_out, lm_out, 
                            fr_out, fr_out_2, &rgb_score_out_s);
#else
        ret = kapp_fdr(KAPP_IMG_SRC_RGB, fdr_out, fd_out, lm_out, fr_out, &rgb_score_out_s);
#endif
        if (ret == KAPP_OK) {
            u16* mode = (u16*) &rsp->param2;
            rsp->error = NO_ERROR; // got user_id or feature map stored
            // we have u16 mode and u16 user_id/fm_indx in a u32 field (Little Endian)
            if (SFID_user) {
                *mode++ = 1;
                *mode = SFID_index;
            } else {
                *mode++ = 0;
                *mode = fdr_out->user_id;
            }
        } else { // error conditions
            rsp->error = fid_ret_to_err(ret);
            rsp->param2 = 0;
        }
        out_size = fdlmfrlv_outputs(&out_flag, pfdfr_out, lv_flag);
        pfdfr_out = (u16*)&rsp->data;
        *pfdfr_out = out_flag;  // update output flag

        kcomm_send_rsp(CMD_SFID_SEND_IMAGE, rsp->data, 4 + out_size);
    } else { // no image
        rsp->error = MSG_APP_BAD_IMAGE;
        rsp->param2 = 0;
        if (ret < 0) {
            kcomm_send_no_rsp();  // if tx/rx timed out, abort the msg reply
        } else {
            kcomm_send_rsp(CMD_SFID_SEND_IMAGE, rsp->data, 4);
        }
    }
}

static void handle_sfid_add_db(RspPram *rsp)
{
    dlog("");

    if (!valid_app_id(APP_id)) {
        rsp->error = MSG_CONFIG_ERROR;  // not in SFID mode
        rsp->param2 = 0;
        kcomm_send_rsp(CMD_SFID_ADD_DB, NULL, 0);
        return;
    }

    kapp_db_user_data_t data;
    data.user_id_in = SFID_user;

    int ret = kapp_db_add(&data);
    if (ret == KAPP_OK) {  // success
        rsp->error = NO_ERROR;
    } else { // error conditions
        switch (ret) {
        case KAPP_DB_ALREADY_SAVED:
            rsp->error = MSG_APP_UID_EXIST;
            break;
        case KAPP_DB_REG_FIRST:
            rsp->error = MSG_APP_IDX_FIRST;
            break;
        case KAPP_DB_USER_NOT_REG:
            rsp->error = MSG_APP_IDX_MISSING;
            break;
        default:
            rsp->error = MSG_APP_DB_FAIL;
            break;
        }
    }
    rsp->param2 = 0;
    kcomm_send_rsp(CMD_SFID_ADD_DB, NULL, 0);
}

static void handle_sfid_del_db(RspPram *rsp, u32 uid)
{
    dlog("uid [%u]", uid);

    kapp_db_user_data_t data;

    if (!valid_app_id(APP_id)) {
        rsp->error = MSG_CONFIG_ERROR;  // not in SFID mode
        rsp->param2 = 0;
        kcomm_send_rsp(CMD_SFID_DELETE_DB, NULL, 0);
        return;
    }

    info_msg("[kcomm] DELETE USER %d\n", uid);

    data.user_id_in = uid;
    data.del_all = (uid == ALL_USER_UID) ? TRUE : FALSE;

    int ret = kapp_db_delete(&data);
    if (ret == KAPP_OK) {
        rsp->error = NO_ERROR;
    } else {
        if (ret == KAPP_DB_DEL_FAIL) {
            rsp->error = MSG_APP_DB_NO_USER;
        } else {
            rsp->error = MSG_APP_DB_FAIL;
        }
    }
    rsp->param2 = 0;
    kcomm_send_rsp(CMD_SFID_DELETE_DB, NULL, 0);
}

static void handle_sfid_edit_db(RspPram *rsp, u32 mode, u8 *param)
{
    dlog("mode [%u], param [0x%x]", mode, param);

    u32 *db_ddr_addr, *db_flash_addr;
    u32 db_size = kapp_db_query(&db_ddr_addr, &db_flash_addr);  // get DB size & location
    bool app_id_ok = valid_app_id(APP_id);
    u16 *puid_out = (u16*)rsp->data;  // point to output data field    
    int i = 0;
    int ret = 0;

    kapp_db_config_t db_config;
    int ret_get_db_config = kapp_db_get_db_config(&db_config);

    if (mode == DB_IMPORT) {
        if (!app_id_ok) {
            rsp->error = MSG_CONFIG_ERROR;  // image processing not set up yet
            rsp->param2 = APP_id;  // add source id
            kcomm_send_rsp(CMD_SFID_EDIT_DB, NULL, 0);
            return;
        }

        ret = kcomm_read((u32)db_ddr_addr, db_size >= (*(u32*)param) ? *(u32*)param: db_size);  // get DB binary
        if (ret < 0) {
            kcomm_send_no_rsp();  // if tx/rx timed out, abort the msg processing
            kcomm_send_rsp(CMD_SFID_EDIT_DB, NULL, 0);               
            kapp_db_init((u32)db_flash_addr); // reload database from flash, addr needed if db was flushed
            return;
        }
        rsp->param2 = 0;
        if (ret == (db_size >= (*(u32*)param) ? *(u32*)param: db_size)) {
            kdp_memxfer_ddr_to_flash((u32)db_flash_addr, (u32) db_ddr_addr, (ret == (db_size >= (*(u32*)param) ? *(u32*)param: db_size)));  // update db in flash
            rsp->error = NO_ERROR;
        } else {
            rsp->error = MSG_DATA_ERROR;
            kapp_db_init((u32)db_flash_addr);  // need to reload database
        }
        rsp->param2 = (db_size >= (*(u32*)param) ? *(u32*)param: db_size); // add source id
        kcomm_send_rsp(CMD_SFID_EDIT_DB, NULL, 0);
        return;
    } else if ((mode < DB_OP_MAX) && !app_id_ok) {  // for sub command 2-5
        rsp->error = MSG_CONFIG_ERROR;  // image processing not set up yet
        rsp->param2 = 0;
    } else if (mode == DB_EXPORT) { // export DB
        rsp->error = 0;
        if ((0 == *(u32*)param) || (db_size <=  (*(u32*)param))){
            rsp->param2 = db_size;  // set size of DB for export
            kcomm_send_rsp(CMD_SFID_EDIT_DB, NULL, 0);  // send reply now
            kcomm_write((u8*)db_ddr_addr, db_size, WMODE_ALT);  // use alternate data mode
        }
        else
        {
            rsp->param2 = *(u32*)param;  // set size of DB for export
            kcomm_send_rsp(CMD_SFID_EDIT_DB, NULL, 0);  // send reply now
            kcomm_write((u8*)db_ddr_addr, *(u32*)param, WMODE_ALT);  // use alternate data mode            
        }
        return;
    } else if (mode == DB_QUERY) { // send DB user list
        rsp->error = 0;
        i = 0;

        if (0 == ret_get_db_config) {
            for (ret = 0; ret < db_config.max_uid; ret++) { // get DB user info
                if (kapp_db_get_user_id_slot_type(ret) == TYPE_VALID) {
                    *puid_out = kapp_db_get_user_id(ret);
                    puid_out++;
                    i++;
                }
            }
        }
        if (0 == i) {
            rsp->error = MSG_APP_DB_NO_USER;
            kcomm_send_rsp(CMD_SFID_EDIT_DB, NULL, 0); 
        } else {           
            rsp->param2 = i;  // user id count
            kcomm_send_rsp(CMD_SFID_EDIT_DB, rsp->data, i*2);  // send reply now
        }
        
        return;
    } else if (mode == DB_ADD_FM) { // add single FM in DB
        i = *(u16*)param;  // get user id of request
        int uidx = kapp_db_get_user_slot(i);  // get DB slot
        if (uidx < 0) {  // user id not in DB
            uidx = kapp_db_get_available_ID_slot();

            if ((0 != ret_get_db_config) || (uidx >= db_config.max_uid)) {
                rsp->error = MSG_DB_ADD_FM_FAIL;
                rsp->param2 = MSG_SFID_NO_MEM;
                kcomm_send_rsp(CMD_SFID_EDIT_DB, rsp->data, 4);
                return;
            }
        }

#ifdef EMBED_CMP_NPU
        int data_size = FR_RES_LENGTH; // wt_hdr + user_embed
#else
        int data_size = FR_RES_LENGTH;
#endif

        if (fm == 0) {
            fm = kmdw_ddr_reserve(data_size);
        }
        ret = kcomm_read(fm, data_size);
        if (ret == data_size) {
            ret = kapp_db_add_FM(uidx, i, (u8*)fm);
            rsp->param2 = 0;
            if (ret > 0) { // add FM successful
                rsp->error = 0;
                rsp->param2 = ret;
            } else {
                rsp->error = MSG_DB_ADD_FM_FAIL;  // DB error occurred
            }
            if (ret == -1) {
                rsp->param2 = MSG_APP_UID_EXIST;  // user id mismatch
            } else if (ret == -2) {
                rsp->param2 = MSG_APP_BAD_INDEX;
            }
            
            kcomm_send_rsp(CMD_SFID_EDIT_DB, rsp->data, 4);
            return;
        } else {
            rsp->error = MSG_DB_ADD_FM_FAIL;
            rsp->param2 = 0;
            if (ret < 0) {
                kcomm_send_no_rsp();  // if tx/rx timed out, abort the msg reply
            } else {
                kcomm_send_rsp(CMD_SFID_EDIT_DB, rsp->data, 4);
            }
            return;
        }
    } else if (mode == DB_DEL_FM) { // delete single FM in DB (not support mask fdr mode)
#ifdef MASK_FDR 
        rsp->error = MSG_DB_DEL_FM_FAIL;
        rsp->param2 = MSG_APP_BAD_MODE;
#else
        i = *(u16*)param;  // get user id of request
        int uidx = kapp_db_get_user_slot(i);  // get DB slot
        if (uidx < 0) {  // user id not in DB
            rsp->error = MSG_DB_DEL_FM_FAIL;
            rsp->param2 = MSG_APP_UID_EXIST;
        } else {
            i = *(u16*)(&param[2]); // get FM index of request
            ret = kapp_db_get_fm0(uidx); // get FM index of DB
            bool fm_exist = (ret & (1 << (i - 1)));
            if (!fm_exist) {
                rsp->error = MSG_DB_DEL_FM_FAIL;
                rsp->param2 = MSG_APP_BAD_INDEX;
                rsp->param2 += ret << 16;  // show current index count
            } else {  // delete FM
                if (kapp_db_delete_FM(uidx, i) == 0) {
                    rsp->error = 0;
                } else {
                    rsp->error = MSG_DB_DEL_FM_FAIL;  // DB error occurred
                }
                rsp->param2 = 0;
            }
        }
#endif
    } else if (mode == DB_QUERY_FM) {
        i = *(u16*)param;                           // get user id of request
        int uidx = kapp_db_get_user_slot(i);        // get DB slot
        if (uidx < 0) {                             // user id not in DB
            rsp->error = MSG_DB_QUERY_FM_FAIL;
            rsp->param2 = MSG_APP_DB_NO_USER;
        } else {
#ifdef MASK_FDR
            kapp_db_config_t db_config;
            int origin_db_idx = kapp_db_get_db_index();

            kapp_db_get_db_config(&db_config);

            i = *(u16*)(&param[2]);                 // get FM index of request
            bool fm_exist;

            // check fm exist in 2 db
            for (int db_idx = 0; db_idx < db_config.db_num; db_idx++) {
                kapp_db_switch(db_idx);

                ret = kapp_db_get_fm(uidx);         // get FM index of DB
                fm_exist = (ret & (1 << (i - 1)));
                if (!fm_exist) {
                    rsp->error = MSG_DB_QUERY_FM_FAIL;
                    rsp->param2 = MSG_APP_BAD_INDEX;
                    rsp->param2 += ret << 16;       // show current index count

                    kcomm_send_rsp(CMD_SFID_EDIT_DB, NULL, 0);
                    return;
                }
            }

            // send 2 db fm and return
            uint32_t db_num = (uint32_t)db_config.db_num;   // In Mask FDR mode send No Mask and Mask db feature maps
            uint32_t padding_data;

#ifdef EMBED_CMP_NPU
            int data_size = FR_FEAT_SIZE + 16 + 2; // wt_hdr
#else
            int data_size = (FR_FEAT_SIZE * 4);
#endif

            rsp->error = 0;
            rsp->param2 = data_size;

            kcomm_send_rsp(CMD_SFID_EDIT_DB, (uint8_t*)(&db_num), sizeof(uint32_t));
            for (int db_idx = 0; db_idx < db_config.db_num; db_idx++) {
                kapp_db_switch(db_idx);

                uint32_t fm_ddr_addr = kapp_db_query_FM(uidx, i);

#ifdef EMBED_CMP_NPU
                padding_data = (uint32_t)kapp_db_get_user_embed_data(uidx, i);  // embedding data
#else
                padding_data = 0x80;                                            // dummy data (for check compare mode SCPU/NCPU)
#endif

                kcomm_write((u8*)fm_ddr_addr, data_size, WMODE_ALT);    // use alternate data mode
                osDelay(5);                                             // workaround: wait data send
                kcomm_write((u8*)(&padding_data), sizeof(uint32_t), WMODE_ALT);
            }

            kapp_db_switch(origin_db_idx);

            return;
#else
            i = *(u16*)(&param[2]);                 // get FM index of request
            ret = kapp_db_get_fm(uidx);             // get FM index of DB
            bool fm_exist = (ret & (1 << (i - 1)));
            if (!fm_exist) {
                rsp->error = MSG_DB_QUERY_FM_FAIL;
                rsp->param2 = MSG_APP_BAD_INDEX;
                rsp->param2 += ret << 16;           // show current index count
            } else {
                // query FM
                uint32_t fm_ddr_addr = kapp_db_query_FM(uidx, i);
                uint32_t db_num = 1;                // In default mode only send one db feature map
                uint32_t padding_data;
                
#ifdef EMBED_CMP_NPU
                int data_size = FR_FEAT_SIZE + 16 + 2;  // wt_hdr
                padding_data = (uint32_t)kapp_db_get_user_embed_data(uidx, i);
#else
                int data_size = FR_RES_LENGTH;
                padding_data = 0x80;                    // user embedding data won't be negative in new compare method
#endif
                rsp->error = 0;
                rsp->param2 = data_size;
                
                kcomm_send_rsp(CMD_SFID_EDIT_DB, (uint8_t*)(&db_num), sizeof(uint32_t));

                kcomm_write((u8*)fm_ddr_addr, data_size, WMODE_ALT);    // use alternate data mode
                osDelay(5);                                             // workaround: wait data send
                kcomm_write((u8*)(&padding_data), sizeof(uint32_t), WMODE_ALT);

                return;
            }
#endif
        }
    } else { // unknown DB action
        rsp->error = MSG_DB_BAD_PARAM;
        rsp->param2 = 0;
    }

    kcomm_send_rsp(CMD_SFID_EDIT_DB, NULL, 0);
}

static void handle_sfid_db_config(RspPram *rsp, uint32_t mode, kapp_db_config_parameter_t *db_config_parameter)
{
    int32_t ret = 0;
    uint32_t uint32_t_value = 0;

    kapp_db_config_parameter_t ret_db_config_parameter;
    memset(&ret_db_config_parameter, 0, sizeof(kapp_db_config_parameter_t));

    if (!valid_app_id(APP_id)) {
        ret = MSG_CONFIG_ERROR;
        goto FUNC_OUT;
    }

    switch (mode)
    {
    case DB_SET_CONFIG:
        info_msg("[kcomm] Set Config DB Num %d Max UID %d Max FID %d\r\n", db_config_parameter->db_config.db_num, db_config_parameter->db_config.max_uid, db_config_parameter->db_config.max_fid);
        
        ret = kapp_db_config(&(db_config_parameter->db_config));
        break;
    case DB_GET_CONFIG:
        ret = kapp_db_get_db_config(&(ret_db_config_parameter.db_config));

        info_msg("[kcomm] Get Config DB Num %d Max UID %d Max FID %d\r\n", ret_db_config_parameter.db_config.db_num, ret_db_config_parameter.db_config.max_uid, ret_db_config_parameter.db_config.max_fid);
        
        if (0 == ret){
            rsp->error = NO_ERROR;
            rsp->param2 = sizeof(kapp_db_config_parameter_t);   // set size of DB configuration
            kcomm_send_rsp(CMD_SFID_DB_CONFIG, rsp->data, 4);   // send reply now
            kcomm_write((u8*)&ret_db_config_parameter,
                        sizeof(kapp_db_config_parameter_t),
                        WMODE_ALT);                             // use alternate data mode
            return;
        }
        break;
    case DB_GET_DB_INDEX:
        uint32_t_value =  kapp_db_get_db_index();

        info_msg("[kcomm] Get Cureent DB Index %d\r\n", (int32_t)uint32_t_value);

        rsp->error = NO_ERROR;
        rsp->param2 = uint32_t_value;
        kcomm_send_rsp(CMD_SFID_DB_CONFIG, rsp->data, 4);       // send reply now
        return;
    case DB_SWITCH_DB_INDEX:
        uint32_t_value = db_config_parameter->uint32_value;

        if (0 != kapp_db_get_db_config(&(ret_db_config_parameter.db_config))) {
            ret = MSG_APP_DB_FAIL;
            break;
        }

        if (ret_db_config_parameter.db_config.db_num <= uint32_t_value) {
            ret = MSG_CONFIG_ERROR;
            break;
        }

        kapp_db_switch(uint32_t_value);

        info_msg("[kcomm] Switch Cureent DB Index to %d\r\n", uint32_t_value);
        break;
    case DB_SET_VERSION:
        uint32_t_value = db_config_parameter->uint32_value;

        kapp_db_set_db_version(uint32_t_value);

        info_msg("[kcomm] Set DB Version to %u\r\n", uint32_t_value);
        break;
    case DB_GET_VERSION:
        uint32_t_value = kapp_db_get_db_version();

        info_msg("[kcomm] Get DB Vresion %u\r\n", uint32_t_value);

        rsp->error = NO_ERROR;
        rsp->param2 = uint32_t_value;
        kcomm_send_rsp(CMD_SFID_DB_CONFIG, rsp->data, 4);
        return;
    case DB_GET_META_DATA_VERSION:
        uint32_t_value = kapp_db_get_db_meta_data_version();

        info_msg("[kcomm] Get DB Meta Data Vresion %u\r\n", uint32_t_value);

        rsp->error = NO_ERROR;
        rsp->param2 = uint32_t_value;
        kcomm_send_rsp(CMD_SFID_DB_CONFIG, rsp->data, 4);
        return;
    default:
        ret = MSG_CONFIG_ERROR;
        break;
    }

FUNC_OUT:
    if (0 < ret) {
        rsp->error = (MSG_CONFIG_ERROR == ret) ? MSG_CONFIG_ERROR : MSG_DB_BAD_PARAM;
        rsp->param2 = ret;
    } else {
        rsp->error = NO_ERROR;
        rsp->param2 = 0;
    }

    info_msg("[kcomm] Send Config DB rsp -> error %d, param2 %d\r\n", rsp->error, rsp->param2);

    kcomm_send_rsp(CMD_SFID_DB_CONFIG, rsp->data, 4);
}

static int handle_fid_cmd(MsgHdr *hdr, CmdPram *args, RspPram *rsp, void *data)
{
    u16 cmd = hdr->cmd;
    u32 arg1 = args->param1;
    u32 arg2 = args->param2;
    void *arg3 = args->data;

    switch (cmd) {
    /* kai2do: why so many cmds here? */
    case CMD_QUERY_APPS:
    case CMD_SELECT_APP:
    case CMD_SET_MODE:
    case CMD_SET_EVENTS:
    case CMD_UPDATE:
    case CMD_ABORT:
    case CMD_SFID_START:
        handle_fid_start(rsp, hdr, *(float*)&arg1, *(float*)&arg2, (u8*)arg3);
        break;
    case CMD_SFID_NEW_USER:
        handle_fid_new_user(rsp, arg1, arg2);
        break;
    case CMD_SFID_SEND_IMAGE:
        handle_send_img(rsp, arg1, arg2);
        break;
    case CMD_SFID_ADD_DB:
        handle_sfid_add_db(rsp);
        break;
    case CMD_SFID_DELETE_DB:
        handle_sfid_del_db(rsp, arg1);
        break;
    case CMD_SFID_EDIT_DB:
        handle_sfid_edit_db(rsp, arg1, (u8*)arg3);
        break;
    case CMD_SFID_DB_CONFIG:
        handle_sfid_db_config(rsp, arg1, (kapp_db_config_parameter_t *)&args->param2);
        break;
    default:
        return -1;
    }

    return 0;
}

static void init_cmd_handler()
{
}

/* ############################
 * ##    public functions    ##
 * ############################ */

void kcomm_enable_fid_cmds(void)
{
    init_cmd_handler();
    kcomm_reg_cmd_handler(handle_fid_cmd);
}
