/*
 * Kneron USB Communication driver
 *
 * Copyright (C) 2019 Kneron, Inc. All rights reserved.
 *
 */

#include <string.h>
#include "cmsis_os2.h"
#include "base.h"
#include "kneron_mozart.h"
#include "kdrv_scu_ext.h"
#include "base.h"
#include "os_tick.h"
#include "usb_com.h"
#include "kmdw_console.h"
#include "kdrv_usbd2.h"
#include "msg.h"
#include "kdp_system.h"

// USB I/O control
#define USBD_ENDPOINT_CMD_OUT 0x02
#define USBD_ENDPOINT_CMD_IN 0x81
#define USBD_ENDPOINT_OTHER_OUT 0x03
#define USBD_ENDPOINT_OTHER_IN 0x84

#define FLAG_USB_COMM_WAIT_CONNECTED BIT(5)
#define FLAG_USB_COMM_WAIT_CMD_COMPLETE BIT(6)

uint32_t trace_w, trace_r, get_result_wait;

osThreadId_t tid_usb_comm;
static osThreadId_t tid_to_notify;
static uint32_t flags_to_notify;
volatile int usb_mode = 0;

kdrv_usbd2_endpoint_descriptor_t enp_cmd_bulk_out =
    {
        .bLength = 0x07,                           // 7 bytes
        .bDescriptorType = 0x05,                   // Endpoint Descriptor
        .bEndpointAddress = USBD_ENDPOINT_CMD_OUT, //
        .bmAttributes = 0x02,                      // TransferType = Bulk
        .wMaxPacketSize = 0x0200,                  // max 512 bytes
        .bInterval = 0x00,                         // never NAKs
};

kdrv_usbd2_endpoint_descriptor_t enp_cmd_bulk_in =
    {
        .bLength = 0x07,                          // 7 bytes
        .bDescriptorType = 0x05,                  // Endpoint Descriptor
        .bEndpointAddress = USBD_ENDPOINT_CMD_IN, //
        .bmAttributes = 0x02,                     // TransferType = Bulk
        .wMaxPacketSize = 0x0200,                 // max 512 bytes
        .bInterval = 0x00,                        // never NAKs
};

kdrv_usbd2_endpoint_descriptor_t enp_aux_bulk_in =
    {
        .bLength = 0x07,                            // 7 bytes
        .bDescriptorType = 0x05,                    // Endpoint Descriptor
        .bEndpointAddress = USBD_ENDPOINT_OTHER_IN, //
        .bmAttributes = 0x02,                       // TransferType = Interrupt
        .wMaxPacketSize = 0x0200,                   // max 1024 bytes
        .bInterval = 0x64,                          // interval = 100 * frame period (125us) = 12 ms
};

kdrv_usbd2_endpoint_descriptor_t enp_aux_bulk_out =
    {
        .bLength = 0x07,                             // 7 bytes
        .bDescriptorType = 0x05,                     // Endpoint Descriptor
        .bEndpointAddress = USBD_ENDPOINT_OTHER_OUT, //
        .bmAttributes = 0x02,                        // TransferType = Bulk
        .wMaxPacketSize = 0x0200,                    // max 512 bytes
        .bInterval = 0x00,                           // never NAKs
};

kdrv_usbd2_interface_descriptor_t intf_desc =
    {
        .bLength = 0x9,          // 9 bytes
        .bDescriptorType = 0x04, // Inteface Descriptor
        .bInterfaceNumber = 0x0, // Interface Number
        .bAlternateSetting = 0x0,
        .bNumEndpoints = 0x4,      // 4 endpoints
        .bInterfaceClass = 0xFF,   // Vendor specific
        .bInterfaceSubClass = 0x0, // 3rd party uses 0x01
        .bInterfaceProtocol = 0x0, // 3rd party uses 0x50
        .iInterface = 0x0,         // No String Descriptor
        .endpoint[0] = &enp_cmd_bulk_out,
        .endpoint[1] = &enp_cmd_bulk_in,
        .endpoint[2] = &enp_aux_bulk_in,
        .endpoint[3] = &enp_aux_bulk_out,
};

kdrv_usbd2_config_descriptor_t confg_desc =
    {
        .bLength = 0x09,                                                        // 9 bytes
        .bDescriptorType = 0x02,                                                // Type: Configuration Descriptor
        .wTotalLength = (9 + 9 + 4 * sizeof(kdrv_usbd2_endpoint_descriptor_t)), // stotal bytes including config/interface/endpoint descriptors
        .bNumInterfaces = 0x1,                                                  // Number of interfaces
        .bConfigurationValue = 0x1,                                             // Configuration number
        .iConfiguration = 0x0,                                                  // No String Descriptor
        .bmAttributes = 0xC0,                                                   // Self-powered, no Remote wakeup
        .MaxPower = 0x0,                                                        // 0 syould be ok for self-powered device
        .interface = &intf_desc,
};

kdrv_usbd2_device_descriptor_t dev_desc =
    {
        .bLength = 0x12,         // 18 bytes
        .bDescriptorType = 0x01, // Type : Device Descriptor
        .bcdUSB = 0x200,         // USB 2.0
        .bDeviceClass = 0x00,    // Device class, 0x0: defined by the interface descriptors
        .bDeviceSubClass = 0x00, // Device sub-class
        .bDeviceProtocol = 0x00, // Device protocol
        .bMaxPacketSize0 = 0x40, // Max EP0 packet size: 64 bytes
        .idVendor = 0x3231,      // Vendor ID
        .idProduct = 0x0100,     // Product ID
        .bcdDevice = 0x0001,     // Device release number
        .iManufacturer = 0x01,   // Manufacture string index, FIXME
        .iProduct = 0x02,        // Product string index, FIXME
        .iSerialNumber = 0x3,    // Serial number string index
        .bNumConfigurations = 1, // Number of configurations, FIXME
        .config = &confg_desc,   // configuration descriptor
};

// Device Qualifier Descriptor, plz see USB 2.0 SPEC
kdrv_usbd2_device_qualifier_descriptor_t dev_qual_desc =
    {
        .bLength = 0xA,
        .bDescriptorType = 0x06,
        .bcdUSB = 0x200,
        .bDeviceClass = 0x0,
        .bDeviceSubClass = 0x0,
        .bDeviceProtocol = 0x0,
        .bMaxPacketSize0 = 0x40,
        .bNumConfigurations = 0x1,
        .bReserved = 0x0,
};

#define USB_MANUFACTURER_STRING {'K',0,'n',0,'e',0,'r',0,'o',0,'n',0}
#define USB_PRODUCT_STRING {'K',0,'n',0,'e',0,'r',0,'o',0,'n',0,' ',0,'K',0,'L',0,'5',0,'2',0,'0',0}
#define USB_SERIAL_STRING {'0',0,'0',0,'0',0,'0',0,'0',0,'0',0,'0',0,'0',0}

kdrv_usbd2_prd_string_descriptor_t str_desc_mun =
{
    .bLength = (2 + 12),
    .bDescriptorType = 0x03,
    .bString = USB_MANUFACTURER_STRING,
};

kdrv_usbd2_prd_string_descriptor_t str_desc_prd =
{
    .bLength = (2 + 24),
    .bDescriptorType = 0x03,
    .bString = USB_PRODUCT_STRING,
};

kdrv_usbd2_prd_string_descriptor_t str_desc_serial =
{
    .bLength = (2 + 16),
    .bDescriptorType = 0x03,
    .bString = USB_SERIAL_STRING,
};

kdrv_usbd2_string_descriptor_t str_desc =
{
    .bLength = 4,
    .bDescriptorType = 0x03,
    .bLanguageID = 0x0409,
    .desc[0] = &str_desc_mun,
    .desc[1] = &str_desc_prd,
    .desc[2] = &str_desc_serial,
};

/* ############################
 * ##    static functions    ##
 * ############################ */

__weak void usbd2_event_isr_callback(kdrv_usbd2_link_status_t link_status)
{
    switch (link_status)
    {
    case USBD2_STATUS_DISCONNECTED:
        break;

    case USBD2_STATUS_CONFIGURED:
        osThreadFlagsSet(tid_usb_comm, FLAG_USB_COMM_WAIT_CONNECTED);
        break;
    }
}

// control transfer callback
__weak bool usbd2_usr_cx_isr_callback(kdrv_usbd2_setup_packet_t *setup)
{
    return true;
}

static void usb_init_dev(void)
{
    // retrieve real serial number here from efuse
    // then conver it to hex string format
    uint32_t uid = kdp_sys_get_kn_number();
    int32_t sidx = 0;
    for (int i = 7; i >= 0; i--)
    {
        uint32_t hex = (uid >> i * 4) & 0xF;
        str_desc_serial.bString[sidx] = (hex < 10) ? '0' + hex : 'A' + (hex - 10);
        sidx += 2;
    }

    kdrv_usbd2_initialize(&dev_desc, &dev_qual_desc, &str_desc, usbd2_event_isr_callback, usbd2_usr_cx_isr_callback);
    kdrv_usbd2_set_enable(true);
}

#define DEBUG_LOCK_UP 0
static void usb_comm_thread(void *argument)
{
    osThreadFlagsWait(FLAG_USB_COMM_WAIT_CONNECTED, osFlagsWaitAny, osWaitForever);

    while (1)
    {
        // receiving a new command msg ...

        uint32_t msg_rbuf_len = MSG_DATA_BUF_MAX + sizeof(MsgHdr) + sizeof(RspPram) + 4;

        kdrv_status_t status = kdrv_usbd2_bulk_receive(USBD_ENDPOINT_CMD_OUT, (uint32_t *)msg_rbuf, &msg_rbuf_len, 0);
        if (status != KDRV_STATUS_OK)
        {
            if(status == KDRV_STATUS_USBD_TRANSFER_TERMINATED)
                osDelay(10); // FIXME ! endpoint is being reset ??
            else
                err_msg("[usb_com] kdrv_usbd2_bulk_receive err %d\n", status);

            continue;
        }

        osThreadFlagsSet(tid_to_notify, flags_to_notify);
        osThreadFlagsWait(FLAG_USB_COMM_WAIT_CMD_COMPLETE, osFlagsWaitAny, osWaitForever);
    }
}

/* ############################
 * ##    public functions    ##
 * ############################ */

void usb_init()
{
    tid_usb_comm = osThreadNew(usb_comm_thread, NULL, NULL);
    usb_init_dev();
}

void usb_idle(void)
{
    usb_mode = 1;
}

void usb_com_register_notification(osThreadId_t tid, uint32_t flags)
{
    tid_to_notify = tid;
    flags_to_notify = flags;
}

/* send usb msg out
*  mode = 0 -> regular response message, send immediately
*  mode = 1 -> ack packet, send immediately
*  mode = 2 -> FR data, send only after response message is done
*/

int usb_com_write(uint8_t *BufferPtr, uint32_t BufferLen, uint32_t flags, uint32_t mode)
{
    kdrv_status_t status = kdrv_usbd2_bulk_send(USBD_ENDPOINT_CMD_IN, (uint32_t *)BufferPtr, BufferLen, 0);
    if (status != KDRV_STATUS_OK)
    {
        err_msg("[kcomm] Error in usb_com write [%d]\r\n", status);
        return -1; // error
    }

    if (mode == 0)
        osThreadFlagsSet(tid_usb_comm, FLAG_USB_COMM_WAIT_CMD_COMPLETE);

    return 0;
}

int usb_com_read(uint8_t *BufferPtr, uint32_t BufferLen, uint32_t flags, uint32_t mode)
{
    kdrv_status_t status = kdrv_usbd2_bulk_receive(USBD_ENDPOINT_CMD_OUT, (uint32_t *)BufferPtr, &BufferLen, 0);
    if (status != KDRV_STATUS_OK)
    {
        err_msg("[kcomm] Error in usb_com read [%d]\n", status);
        return -1;
    }
    trace_r = 23;

    if (mode == 0)
        osThreadFlagsSet(tid_usb_comm, FLAG_USB_COMM_WAIT_CMD_COMPLETE);

    return (int)BufferLen;
}
