/* --------------------------------------------------------------------------
 * Copyright (c) 2013-2016 ARM Limited. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *      Name:    main.c
 *      Purpose: RTX for Kneron
 *
 *---------------------------------------------------------------------------*/

#include <string.h>
#include "kdrv_ddr.h"
#include "kdrv_system.h"
#include "kmdw_memory.h"
#include "kdp_ddr_table.h"
#include "kmdw_console.h"
#include "kmdw_camera.h"
#include "board_kl520.h"
#include "pinmux.h"
#include "kl520_include.h"
#include "uvc_utils.h"
#include "kmdw_display.h"

extern int sample_uvc_ctl_test(void);
osThreadId_t tid_console;
#define FLAGS_UVC_CAMERA_INIT_DONE_EVT         BIT4

static int display_inited;

typedef struct myCamera_ctx {
    bool init;
    enum camera_state state;
    uint32_t width;
    uint32_t height;
    uint32_t pixelformat;
}  uvc_camera_exp_ctx;

uvc_camera_exp_ctx cam_settings[IMGSRC_NUM] = {
    {false, CAMERA_STATE_IDLE, 0, 0, 0},
};
typedef uint32_t (*console_cmd_func)(void);
struct console_cmd_info
{
    char *desc;
    console_cmd_func func;
};

static void kapp_display_init(void)
{
    struct video_input_params params;

    if (display_inited == 0) {
        params.input_fmt = V2K_PIX_FMT_YCBCR;
        params.input_xres = LCD_WIDTH;
        params.input_yres = LCD_HEIGHT;
        kmdw_video_renderer_open(&params);
        kmdw_video_renderer_set_camera(0);
        kmdw_display_set_pen_rgb565(BLACK, 4);
        kmdw_video_renderer_buffer_initialize(&params);
        kmdw_video_renderer_start();
        display_inited = 1;
    }
}

static void kapp_display_exit(void)
{
    if (display_inited == 1) {
        kmdw_video_renderer_stop();
        display_inited = 0;
    }
}

static uint32_t open_uvc_camera()
{
    uint32_t ret;
    uint8_t cam_idx = 0;
    struct cam_capability cap;
    struct cam_format fmt;

    char fmtstr[8];
    memset(&cap, 0, sizeof(cap));
    memset(&fmt, 0, sizeof(fmt));

    if (0 != (ret = kmdw_camera_open(cam_idx)))
        return ret;

    if (0 != (ret = kmdw_camera_get_device_info(cam_idx, &cap)))
        return ret;

    fmt.width = cam_settings[cam_idx].width;
    fmt.height = cam_settings[cam_idx].height;
    fmt.pixelformat = cam_settings[cam_idx].pixelformat;
//    kmdw_printf("[%s] cam_idx:%d, width:%d, height:%d,  pixelformat: %d\n", __func__, cam_idx, fmt.width, fmt.height, fmt.pixelformat);
    if (0 != (ret = kmdw_camera_set_frame_format(cam_idx, (struct cam_format *)&fmt)))
        return ret;

    if (0 != (ret = kmdw_camera_get_frame_format(cam_idx, &fmt)))
        return ret;

    memset(fmtstr, 0, 8);
    memcpy(fmtstr, &fmt.pixelformat, 4);

    if (0 != (ret = kmdw_camera_buffer_init(cam_idx)))
        return ret;

    cam_settings[cam_idx].state = CAMERA_STATE_INITED;
    return 0;
}

static uint32_t uvc_camera_setting(uint8_t cam_idx, unsigned int width, unsigned int height, unsigned int pixelformat)
{
    if(true == cam_settings[cam_idx].init) {
        kmdw_printf("[%s] camera index %d is already initialized\n", __func__, cam_idx);
        return 1;
    }
    cam_settings[cam_idx].init = true;
    cam_settings[cam_idx].width = width;
    cam_settings[cam_idx].height = height;
    cam_settings[cam_idx].pixelformat = pixelformat;
    return 0;
}

static uint32_t start_uvc_camera()
{
    uint32_t ret;
    uint8_t cam_idx = 0;
    if (0 != (ret = kmdw_camera_start(cam_idx))) {
        return ret;
    }

    kapp_display_init();

    cam_settings[cam_idx].state = CAMERA_STATE_RUNNING;
    return 0;
}
static uint32_t stop_uvc_camera()
{
    uint32_t ret;
    uint8_t cam_idx = 0;

    kapp_display_exit();

    if(0 != (ret = kmdw_camera_stop(cam_idx)))
        return ret;

    cam_settings[cam_idx].state = CAMERA_STATE_INITED;
    return 0;
}

static uint32_t close_uvc_camera()
{
    uint32_t ret;
    uint8_t cam_idx=0;
    if(0 != (ret = kmdw_camera_close(cam_idx)))
        return ret;

    cam_settings[cam_idx].state = CAMERA_STATE_IDLE;
    return ret;
}

static uint32_t uvc_camera_ioctl_test(void)
{
    sample_uvc_ctl_test();
    return 0;
}

static uint32_t sample_app_quit(void)
{
    DSG("Oh! My God, Bye bye !!");
    return 0;
}
struct console_cmd_info app_idle_mode_array[] = {
    {"Open UVC Camera",             open_uvc_camera  },
    {"Start UVC Camera",            start_uvc_camera  },
    {"Stop UVC Camera",             stop_uvc_camera  },
    {"Close UVC Camera",            close_uvc_camera  },
    {"Test UVC Camera Control Command",  uvc_camera_ioctl_test },
    {"Quit",                        sample_app_quit  },
};


void uvc_test_console(void *arg)
{
    int id;
    int i = 0;
    uint32_t idle_cmd_size = ARRAY_SIZE(app_idle_mode_array);
    char buf[64];
    console_cmd_func cmd_func;

    kmdw_camera_init();
    uvc_camera_setting(0, YCBCR422_IMG_SOURCE_W, YCBCR422_IMG_SOURCE_H, V2K_PIX_FMT_YCBCR);
    osThreadFlagsWait(FLAGS_UVC_CAMERA_INIT_DONE_EVT, osFlagsWaitAll, osWaitForever);

    while(1)
    {
        id = 0;
         DSG("\n === IDLE Mode === ");
        for (i = 0; i < idle_cmd_size; ++i)
        { 
            sprintf(buf, "(%2d) %s", i + 1, app_idle_mode_array[i].desc);
            DSG("%s", buf);
        }
        DSG_NOLF(" command >> ");
        kmdw_console_echo_gets(buf, sizeof(buf));

        id = atoi(strtok(buf, " \r\n\t"));
        if (id > 0 && id <= idle_cmd_size)
        {
            cmd_func = app_idle_mode_array[id - 1].func;
            if (cmd_func) {
                cmd_func();    
            }
        }
        else {
            err_msg("Invalid command: %d", id);
            continue;
        }
    }
}

void console_entry(void)
{
    tid_console = osThreadNew(uvc_test_console, NULL, NULL);
    osThreadSetPriority (tid_console, osPriorityBelowNormal);
}

int main(void)
{
    SystemCoreClockUpdate();
    osKernelInitialize();

    kdrv_system_init();
    kdrv_system_init_ncpu();
    kdrv_ddr_system_init(DDR_INIT_ALL);
    pinmux_init();
    kmdw_ddr_init(KDP_DDR_HEAP_HEAD_FOR_MALLOC+1, KDP_DDR_MODEL_RESERVED_END);
    kdrv_uart_initialize();
    kmdw_console_init(UART0_DEV);
    kmdw_console_set_log_level_scpu(LOG_PROFILE);

    kmdw_display_initialize();	

    console_entry();
    if (osKernelGetState() == osKernelReady)
    {
        osKernelStart();
    }

    while (1)
    {
    }
}
