/* -------------------------------------------------------------------------- 
 * Copyright (c) 2013-2016 ARM Limited. All rights reserved.
 *
 * SPDX-License-Identifier: Apache-2.0
 *
 * Licensed under the Apache License, Version 2.0 (the License); you may
 * not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *      Name:    main.c
 *      Purpose: RTX for Kneron
 *
 *---------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include "kdrv_uart.h"

#include "kneron_mozart.h"                // Device header
#include "kneron_mozart_ext.h"
#include "cmsis_os2.h"                    // ARM::CMSIS:RTOS2:Keil RTX5
#include "Driver_Common.h"
#include "base.h"
#include "io.h"
#include "string.h"
#include "kdrv_ddr.h"
#include "kmdw_model.h"

#ifdef KL520
#include "kdrv_scu_ext.h"
#include "bootloader.h"
#include "kdrv_power.h"
#include "kdrv_system.h"
#include "kdrv_clock.h"
#include "kdrv_scu.h"
#include "power_manager.h"
#include "system_ncpu.h"
#endif

#include "kmdw_memory.h"     //for ddr_malloc
#include "kdp_ddr_table.h"  //for default ddr table
#include "version.h"
#include "kmdw_console.h"
#include "pinmux.h"
#include "kmdw_camera.h"
#include "sample_app_wifi.h"
#include "media/kdp_command_adapter.h"
#include "kmdw_display.h"

#include "kdrv_i2c.h"
#include "kdrv_uart.h"

/**
 * @info(), Get kernel information
 */
extern uint32_t kdp_sys_get_key_status(void);

int cpu_type = 0;
extern u32 bootup_status;
extern u32 warm_boot;

static void info(void)
{
    uint32_t rom_table;
	
    rom_table = inw(CPU_ID_STS_ADDR);
    if (CPU_ID_SCPU == rom_table) {
        DSG("=== [I am SCPU] firmware version: SDK %d.%d.%d.%d build.%d ===", 
                                                       g_fw_misc.architecture,
                                                       g_fw_misc.major,
                                                       g_fw_misc.minor,
                                                       g_fw_misc.update,
                                                       g_fw_misc.build);
    }
    else if (CPU_ID_NCPU == rom_table) {
        DSG("=== [Who am I? It must be something wrong. ===");
    }  
    
    DSG("=== This is host mode ===");

    if (warm_boot) {
        DSG("  -> Bootup Status: 0x%x (warm boot) <-", bootup_status);
    } else {
        DSG("  -> Bootup Status: 0x%x <-", bootup_status);
    }

    if (bootup_status & SCU_REG_BTUP_STS_HWR)
        DSG("                  : Hardware Reset");
    if (bootup_status & SCU_REG_BTUP_STS_WDR)
        DSG("                  : Watchdog Reset");
    if (bootup_status & SCU_REG_BTUP_STS_SMR)
        DSG("                  : Dormant Reset");
    if (bootup_status & SCU_REG_BTUP_STS_PMR)
        DSG("                  : Poweroff Reset");
    if (bootup_status & SCU_REG_BTUP_STS_PWRBTN_STS)
        DSG("                  : Power Button Reset");
    if (bootup_status & SCU_REG_BTUP_STS_RTC_BTUPTS)
        DSG("                  : RTC Alarm Reset");
    if (bootup_status & SCU_REG_BTUP_STS_PMR1)
        DSG("                  : Power Mode 1 Reset");
    if (bootup_status & SCU_REG_BTUP_STS_PMR2)
        DSG("                  : Power Mode 2 Reset");
}

/**
 * @brief main, main dispatch function
 */
int main(void) 
{
    uint32_t ncpu_sum32 = 0;
    uint32_t keystatus;

    kdrv_system_init();
    kdrv_system_init_ncpu();
    kdrv_ddr_system_init(DDR_INIT_ALL);

    pinmux_init();
    kmdw_ddr_init(KDP_DDR_HEAP_HEAD_FOR_MALLOC+1, KDP_DDR_MODEL_RESERVED_END);
    kdrv_uart_initialize();
    kmdw_console_init(UART0_DEV);   // for log
    kdrv_i2c_initialize(KDRV_I2C_CTRL_0, KDRV_I2C_SPEED_400K);
    /* Only need to check crc after loading ncpu fw */
    /* load & launch ncpu fw from flash, skip if ncpu uses debugger loaded via JTAG */
    ncpu_sum32 = load_ncpu_fw(-1); 
    if(ncpu_sum32 != 0) {
        /* scpu image has been checked on the SPL */
        DSG("\n !!!! WARNING !!!! NCPU FW image check failure...(wrong sum32 0x%08x)\n", ncpu_sum32);
    }
    kdrv_delay_us(1000); 
    keystatus = kdp_sys_get_key_status();
    //DSG("\n#########################################");
    if(keystatus == 0)
    {
        DSG("## With key  ##");
    }
    else if(keystatus == 1)
    {
        DSG("## Without key  ##");
    }
    else if(keystatus == 0xff)
    {
        DSG("## Fail to read eFuse status  ##");
    }
   // DSG("#########################################\n");
    info();
    SystemCoreClockUpdate();            // System Initialization
    osKernelInitialize();               // Initialize CMSIS-RTOS
    power_manager_init();

    kmdw_model_init();
    kmdw_camera_init();
    //kmdw_display_initialize();        // to reduce code size, 
                                        // remove minor part from this example

    sample_doorlock_entry();

    //application is triggered in host_com.c
    if (osKernelGetState() == osKernelReady) {
        osKernelStart();
    }

    while(1) {
    }
}
