'''
 @file bin_gen.py
 @brief Generate flash final bin file from other seperate bin files.
     start addr offset     size
     0x00000000   0       8KB        boot_spl.bin
     0x00002000           88KB        fw_scpu.bin
     0x00018000           64KB        fw_ncpu.bin
     0x00028000            4KB        boot_cfg0.bin
     0x00029000           96KB        reserved
     0x00041000           88KB        fw_scpu1.bin
     0x00057000           64KB        fw_ncpu1.bin
     0x00067000            4KB        boot_cfg1.bin
     0x00068000           96KB        reserved
     0x00080000  512K     512KB       user feature map database
     0x00300000 3076K     232B        fw_info.bin
     0x00301000           3576800     all_models.bin (MAX size = 28MB)
'''

import argparse
import io
import os
import code
import sys
import pdb
import numpy as np
import platform
import shutil

SPL_CPU_ONLY=0
BIN_DIR="./flash_bin"
OUTPUT_FW_FILE="flash_image.bin"

MAX_MODEL_SIZE=0x1C00000

flash_bin_info = [
                  [0x00000000, 8*1024,  'boot_spl.bin'],
                  [0x00002000, 88*1024, 'fw_scpu.bin'],
                  [0x00018000, 64*1024, 'fw_ncpu.bin'],
                  [0x00028000, 4*1024,  'boot_cfg0.bin'],
                  [0x00029000, 96*1024, 'fw_reserved.bin'],
                  [0x00041000, 88*1024, 'fw_scpu1.bin'],
                  [0x00057000, 64*1024, 'fw_ncpu1.bin'],
                  [0x00067000, 4*1024,  'boot_cfg1.bin'],
                  [0x00068000, 96*1024, 'fw_reserved.bin'],
                  [0x00080000, 512*1024,'fid_map.bin'],
                  [0x00300000, 236,     'fw_info.bin'],
                  [0x00301000, 16244624, 'all_models.bin']]

def file_read_binary(src):
     ''' Load depth file '''
     with open(src, "rb") as f:
         data = f.read()
         return data

def write_data_to_binary_file(ofile, data):
    with io.open(ofile, 'wb') as fo:
        fo.write(bytearray(data))

def bin_gen(bin_info, ofile):
    model_size = 0
    total_bin_size = 0
    platform_os = platform.system()

    if platform_os == 'Windows':
        cmd = r'..\nef_utility\nef_utility_win.exe -X flash_bin\models_520.nef'
        nef_file = r'flash_bin\models_520.nef'
        fw_info_path = r'output\fw_info.bin'
        all_models_path = r'output\all_models.bin'
        fw_info_path_new = r'flash_bin\fw_info.bin'
        all_models_path_new = r'flash_bin\all_models.bin'
    elif platform_os == 'Linux':
        cmd = r'../nef_utility/nef_utility_linux -X flash_bin/models_520.nef'
        nef_file = r'flash_bin/models_520.nef'
        fw_info_path = r'output/fw_info.bin'
        all_models_path = r'output/all_models.bin'
        fw_info_path_new = r'flash_bin/fw_info.bin'
        all_models_path_new = r'flash_bin/all_models.bin'

    if SPL_CPU_ONLY == 0:
        if os.path.isfile(nef_file):
            os.system(cmd)
            shutil.move(fw_info_path, fw_info_path_new)
            shutil.move(all_models_path, all_models_path_new)
        else:
            print('models_520.nef file does not exist in flash_bin folder ')
            sys.exit(-1)

        model_file = BIN_DIR + "/" + bin_info[-1][2]
        if os.path.isfile(model_file):
            model_size = os.path.getsize(model_file)
            #print('model size = %d' %model_size)
            if(model_size >= MAX_MODEL_SIZE):
                print('[ERROR] model file size is too big')
                sys.exit(-1)
        else:
            print('model file is not available')
            sys.exit(-1)
        total_bin_size = bin_info[-1][0] + model_size
    else:
        total_bin_size = bin_info[-1][0] + bin_info[-1][1]

    bin_data = np.zeros(total_bin_size, dtype=np.uint8)
    bin_data.fill(0xff)

    for bin_file in bin_info:
        size_in_kb = int(bin_file[1]/1024)
        print('{:>15s}  =>  address: 0x{:08X}, size: 0X{:08X} ( ~{:>6d}K)' \
                .format(bin_file[2], bin_file[0], bin_file[1], size_in_kb), end='')
        bin_start_addr = bin_file[0]
        bin_file = BIN_DIR + "/" + bin_file[2]

        if os.path.isfile(bin_file) == 0:
            print('  not existed, skip!')
        else:
            print('')
            rdata = file_read_binary(bin_file)
            rdata = list(rdata)
            rdata = np.array(rdata)
            bin_data[bin_start_addr:bin_start_addr + len(rdata)] = rdata

    write_data_to_binary_file(ofile, bin_data)
    print('==================')

    if SPL_CPU_ONLY == 0 and os.path.isfile(nef_file):
        os.rmdir("output")
        os.remove(fw_info_path_new)
        os.remove(all_models_path_new)

    if os.path.isfile(OUTPUT_FW_FILE):
        full_path_name=os.path.abspath(OUTPUT_FW_FILE)
        print(' {:s} is generated'.format(full_path_name))
        return True
    else:
        print('Error!!')
        return False

if __name__ == '__main__':
    ### please make sure image directory and model path area correct
    print('')
    print('Generating flash binary file...')
    print('')
    print(' --- Bin file information list ---')

    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--CPU_ONLY', action='store_true', help='SPL/SCPU/NCPU only', default=False)

    args = parser.parse_args()
    if (args.CPU_ONLY):
        SPL_CPU_ONLY = 1

    if SPL_CPU_ONLY == 1:
        flash_bin_info = flash_bin_info[:3]

    ret = bin_gen(flash_bin_info, OUTPUT_FW_FILE)
    if ret:
        print("\n[done]")

